@@ -1,6 +1,6 @@
 ;;;; batch-driver.scm - Driver procedure for the compiler
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -37,45 +37,46 @@
 
 
 #{compiler
-  build-information compiler-arguments process-command-line dump-nodes
+  compiler-arguments process-command-line dump-nodes dump-undefined-globals
   default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
-  installation-home compiler-cleanup-hook
-  foreign-type-table-size file-io-only
-  unit-name insert-timer-checks used-units
+  compiler-cleanup-hook check-global-exports
+  foreign-type-table-size file-io-only parse-easy-ffi
+  unit-name insert-timer-checks used-units inline-max-size
   debugging perform-lambda-lifting! disable-stack-overflow-checking no-c-syntax-checks
-  register-ffi-macro ffi-include-path-list
+  register-ffi-macro ffi-include-path-list ffi-dont-include
   foreign-declarations emit-trace-info block-compilation analysis-database-size line-number-database-size
-  target-heap-size target-stack-size try-harder default-installation-home target-heap-growth target-heap-shrinkage
+  target-heap-size target-stack-size try-harder target-heap-growth target-heap-shrinkage
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
-  target-initial-heap-size split-level default-user-pass-2
+  target-initial-heap-size split-level postponed-initforms
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
   broken-constant-nodes inline-substitutions-enabled enable-sharp-greater-read-syntax
   emit-profile profile-lambda-list profile-lambda-index profile-info-vector-name
-  direct-call-ids foreign-type-table first-analysis optimizers-used
+  direct-call-ids foreign-type-table first-analysis emit-closure-info emit-line-info
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
   reorganize-recursive-bindings substitution-table simplify-named-call emit-unsafe-marker
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
   transform-direct-lambdas! source-filename compressed-literals literal-compression-threshold
-  debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
+  debugging-chicken bomb check-signature posq stringify symbolify flonum? build-lambda-list
   string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
   put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
   build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node expression-has-side-effects?
-  simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
+  simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list external-protos-first
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database dump-exported-globals
-  copyright compiler-features default-declarations units-used-by-default words-per-flonum default-debugging-declarations
+  default-declarations units-used-by-default words-per-flonum default-debugging-declarations
   default-profiling-declarations default-optimization-passes compressed-literals-initializer
+  inline-max-size file-requirements
   foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   chop-separator chop-extension display-real-name-table display-line-number-database explicit-use-flag
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
-  export-list module-export-list compute-module-exports
-  foreign-argument-conversion foreign-result-conversion parse-script-header}
+  export-list do-lambda-lifting
+  foreign-argument-conversion foreign-result-conversion}
 
 
 (include "tweaks")
@@ -88,7 +89,10 @@
   (define (option-arg p)
     (if (null? (cdr p))
 	(quit "missing argument to `-~A' option" (car p))
-	(cadr p) ) )
+	(let ([arg (cadr p)])
+	  (if (symbol? arg)
+	      (quit "invalid argument to `~A' option" arg)
+	      arg) ) ) )
   (initialize-compiler)
   (set! explicit-use-flag (memq 'explicit-use options))
   (let ([initforms `((##core#declare
@@ -117,31 +121,26 @@
                                      (string->number oname))
                                     (else oname))))]
                           [else 1])]
-	[home (chop-separator (or (getenv "CHICKEN_HOME") default-installation-home ""))]
 	[ipath (map chop-separator (string-split (or (getenv "CHICKEN_INCLUDE_PATH") "") ";"))]
-	[strict (memq 'strict options)]
-	[strict-srfi0 (memq 'strict-srfi-0 options)]
-	[strict-letrec (memq 'strict-letrec options)]
-	[strictr (memq 'strict-reader options)]
 	[opasses default-optimization-passes]
 	[time0 #f]
 	[time-breakdown #f]
 	[forms '()]
 	[cleanup-forms '(((##sys#implicit-exit-handler)))]
-	[profile (memq 'profile options)]
+	[profile (or (memq 'profile options) (memq 'accumulate-profile options) (memq 'profile-name options))]
+	[profile-name (or (and-let* ((pn (memq 'profile-name options))) (cadr pn)) default-profile-name)]
 	[hsize (memq 'heap-size options)]
 	[hisize (memq 'heap-initial-size options)]
 	[hgrowth (memq 'heap-growth options)]
 	[hshrink (memq 'heap-shrinkage options)]
 	[kwstyle (memq 'keyword-style options)]
 	[lcthreshold (memq 'compress-literals options)]
-	[hygienic (or (memq 'hygienic options) (memq 'syntax options))]
-	[hygienic-at-rt (or (memq 'hygienic-at-run-time options) (memq 'syntax-at-run-time options))]
 	[uunit (memq 'unit options)]
 	[a-only (memq 'analyze-only options)]
 	[dynamic (memq 'dynamic options)]
 	[dumpnodes #f]
 	[ffi-mode (memq 'ffi options)]
+	[ffi-parse-mode (memq 'ffi-parse options)]
 	[ofilelist #f]
 	[quiet (memq 'quiet options)]
 	[ssize (or (memq 'nursery options) (memq 'stack-size options))] )
@@ -212,42 +211,20 @@
 
     (define (end-time pass)
       (when time-breakdown
-	(printf "milliseconds needed for ~a: ~s~%" pass (- (cputime) time0)) ) )
-
-    (define (install-macros)
-      (fluid-let ([##sys#read-line-counter ##sys#read-line-counter])
-	(##syncase#install-macro-package (or strict strict-srfi0) (not strict) (or strict strict-srfi0 strict-letrec)) ) )
-
-    (define (strictness s s0)
-      (cond [s
-	     (set! strict #t)
-	     (when verbose (display "Strict mode\n"))
-	     (set! ##sys#strict-mode #t)
-	     (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-0 #:srfi-16 #:srfi-9 #:srfi-26)
-	     (register-feature! #:strict)
-	     (##sys#undefine-non-standard-macros '()) ]
-	    [s0 
-	     (set! strict-srfi0 #t)
-	     (when verbose (display "Strict mode with SRFI-0\n"))
-	     (set! ##sys#strict-mode #t)
-	     (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-16 #:srfi-9 #:srfi-26)
-	     (register-feature! #:strict)
-	     (##sys#undefine-non-standard-macros '(cond-expand)) ] )
-      (if (or s s0 strictr) 
-	  (begin
-	    (strict-reader #t)
-	    (keyword-style #:none)
-	    (case-sensitive #f) )
-	  (enable-sharp-greater-read-syntax) ) )
+	(printf "milliseconds needed for ~a: \t~s~%" pass (- (cputime) time0)) ) )
 
     (define (read-form in)
-      (if ffi-mode
+      (if (or ffi-mode ffi-parse-mode)
 	  (let ([s (read-string #f in)])
 	    (if (string=? s "")
-		(end-of-file)
-		`(declare 
-		   (foreign-declare ,s)
-		   (foreign-parse ,s) ) ) )
+		#!eof
+		(cond [ffi-mode
+		       `(declare 
+			  (foreign-declare ,s)
+			  (foreign-parse ,s) ) ]
+		      [else
+		       `(declare
+			  (foreign-parse ,s) ) ] ) ) )
 	  (##sys#read in infohook) ) )
 
     (when (or (< partitions 1) (not outfile))
@@ -262,22 +239,36 @@
        (collect-options 'debug) ) )
     (set! try-harder #t)
     (set! dumpnodes (memq '|D| debugging-chicken))
+    (when (memq 'lambda-lift options) (set! do-lambda-lifting #t))
     (when (memq 't debugging-chicken) (##sys#start-timer))
     (when (memq 'b debugging-chicken) (set! time-breakdown #t))
+    (when (memq 'raw options)
+      (set! explicit-use-flag #t)
+      (set! cleanup-forms '())
+      (set! initforms '()) )
+    (when (memq 'no-lambda-info options)
+      (set! emit-closure-info #f) )
     (when (memq 'no-warnings options) 
       (when verbose (printf "Warnings are disabled~%~!"))
-      (set! warnings-enabled #f)
       (set! ##sys#warnings-enabled #f) )
     (when (memq 'optimize-leaf-routines options) (set! optimize-leaf-routines #t))
     (when (memq 'unsafe options) 
       (set! unsafe #t)
-      (##match#set-error-control #:unspecified) )
-    (when (memq 'unsafe-libraries options)
+      (##match#set-error-control #:fail) )
+    (when (and dynamic (memq 'unsafe-libraries options))
       (set! emit-unsafe-marker #t) )
     (when (memq 'disable-interrupts options) (set! insert-timer-checks #f))
     (when (memq 'fixnum-arithmetic options) (set! number-type 'fixnum))
     (when (memq 'block options) (set! block-compilation #t))
     (when (memq 'disable-c-syntax-checks options) (set! no-c-syntax-checks #t))
+    (when (memq 'emit-external-prototypes-first options) (set! external-protos-first #t))
+    (when (memq 'inline options) (set! inline-max-size default-inline-max-size))
+    (when (memq 'track-scheme options) (set! emit-line-info #t))
+    (and-let* ([inlimit (memq 'inline-limit options)])
+      (set! inline-max-size 
+	(let ([arg (option-arg inlimit)])
+	  (or (string->number arg)
+	      (quit "invalid argument to `-inline-limit' option: `~A'" arg) ) ) ) )
     (when (memq 'case-insensitive options) 
       (when verbose (printf "Identifiers and symbols are case insensitive~%~!"))
       (register-feature! 'case-insensitive)
@@ -304,18 +295,17 @@
     (set! ##sys#include-pathnames
       (append (map chop-separator (collect-options 'include-path))
 	      ##sys#include-pathnames
-	      ipath
-	      (if home (list home) '()) ) )
+	      ipath) )
     (set! ofilelist (outfile->list outfile partitions))
     (when (and outfile filename (any (cut string=? <> filename) ofilelist))
       (quit "source- and output-filename are the same") )
 
     ;; Handle feature options:
     (for-each register-feature! (collect-options 'feature))
-    (for-each unregister-feature! (collect-options 'no-feature))
 
     ;; Handle FFI defines and include-paths:
     (for-each register-ffi-macro (collect-options 'ffi-define))
+    (when (memq 'ffi-no-include options) (set! ffi-dont-include #t))
     (set! ffi-include-path-list (append (collect-options 'ffi-include-path) ffi-include-path-list))
 
     ;; Load extensions:
@@ -324,36 +314,35 @@
       (when verbose
 	(printf "Loading compiler extensions...~%~!")
 	(load-verbose #t) )
-      (for-each (lambda (f) (load (##sys#resolve-include-filename f #t))) extends) )
+      (for-each (lambda (f) (load (##sys#resolve-include-filename f #f #t))) extends) )
     (set! ##sys#features (delete #:compiler-extension ##sys#features eq?))
 
     (set! ##sys#features (cons '#:compiling ##sys#features))
-    ;; Install highlevel macros:
-    (if hygienic
-	(begin
-	  (when verbose 
-	    (printf "Using hygienic macros~%~!") )
-	  (install-macros) )
-	(begin
-	  (set! ##sys#features (cons #:match (cons #:match-support ##sys#features)))
-	  (##sys#provide 'match) 
-	  (when (or strict strict-srfi0 strict-letrec) (##sys#enable-strict-letrec) ) ) )
+    (set! ##sys#features (cons #:match (cons #:match-support ##sys#features)))
+    (##sys#provide 'match) 
 
-    ;; Loads required extensions:
-    (let ([ids (map string->symbol (collect-options 'require-for-syntax))])
-      (for-each
-       (lambda (id)
-	 (when verbose (printf "Requiring ~A ...~%" id))
-	 (##sys#require id) )
-       ids)
-      (let ([rs (##sys#lookup-runtime-requirements ids)])
-	(when (pair? rs)
-	  (set! initforms
-	    (append initforms (map (lambda (r) `(##sys#require ',r)) rs)) ) ) ) )
+    ;; Insert postponed initforms:
+    (set! initforms (append initforms postponed-initforms))
+
+    ;; Handle `-extension' options:
+    (when (memq 'extension options)
+      (set! initforms 
+	(append 
+	 initforms
+	 `((define-extension 
+	     ,(string->symbol
+	       (cond (outfile (pathname-file outfile))
+		     (filename (pathname-file filename))
+		     (else (quit "no filename available for `-extension' option")) ) ) ) ) ) ) )
+
+    ;; Append required extensions to initforms:
+    (let ([ids (map string->symbol (collect-options 'require-extension))])
+      (set! initforms
+	(append initforms (map (lambda (r) `(##core#require-extension ',r)) ids)) ) )
 
     (when (memq 'run-time-macros options)
       (set! ##sys#enable-runtime-macros #t) )
-    (strictness strict strict-srfi0)
+    (enable-sharp-greater-read-syntax)
     (set! target-heap-size
       (if hsize
 	  (arg-val (option-arg hsize))
@@ -372,7 +361,7 @@
     (set! emit-trace-info (not (memq 'no-trace options)))
     (set! disable-stack-overflow-checking (memq 'disable-stack-overflow-checks options))
     (when (memq 'm debugging-chicken) (set-gc-report! #t))
-    (when (memq 'usual-integrations options)
+    (unless (memq 'no-usual-integrations options)
       (set! standard-bindings default-standard-bindings)
       (set! extended-bindings default-extended-bindings) )
     (when verbose
@@ -381,15 +370,23 @@
 		  "stacktrace"
 		  "none") ) )
     (when profile
-      (set! emit-profile #t)
-      (set! initforms (append initforms default-profiling-declarations))
-      (when verbose
-	(printf "Generating profile~%~!" emit-profile) ) )
+      (let ([acc (eq? 'accumulate-profile (car profile))])
+	(set! emit-profile #t)
+	(set! initforms
+	  (append
+	   initforms
+	   default-profiling-declarations
+	   (if acc
+	       '((set! ##sys#profile-append-mode #t))
+	       '() ) ) )
+	(when verbose
+	  (printf "Generating ~aprofile~%~!" (if acc "accumulated " "") emit-profile) ) ) )
 
     (cond ((memq 'version options)
 	   (print-version #t)
 	   (newline) )
-	  ((or (memq 'help options) (memq '-help options)) (print-usage))
+	  ((or (memq 'help options) (memq '-help options) (memq 'h options) (memq '-h options))
+	   (print-usage))
 	  ((not filename)
 	   (unless quiet
 	     (print-version #t)
@@ -402,10 +399,8 @@
 	   (set! source-filename filename)
 	   (debugging 'r "options" options)
 	   (debugging 'r "debugging options" debugging-chicken)
-	   (debugging 'r "home directory" home)
 	   (debugging 'r "target heap size" target-heap-size)
 	   (debugging 'r "target stack size" target-stack-size)
-	   (debugging 'r "compiler features" compiler-features)
 
 	   ;; Read toplevel expressions:
 	   (set! ##sys#line-number-database (make-vector line-number-database-size '()))
@@ -428,53 +423,7 @@
 				     (reverse forms)
 				     (map string->expr postlude) ) ) )
 			(let* ([f (car files)]
-			       [in (check-and-open-input-file f)]
-			       [c0 (peek-char in)] )
-			  (set! ##sys#read-line-counter 1)
-
-			  ;; Check for script header:
-			  (cond [ffi-mode]
-				[(char=? #\# c0)
-				 (read-char in)
-				 (if (not (char=? #\! (peek-char in)))
-				     (if (string=? "-" f)
-					 (quit "source input from pipe may not begin with `#' character if not a script") 
-					 (begin
-					   (close-input-port in)
-					   (set! in (check-and-open-input-file f)) ) )
-				     (begin
-				       (read-char in)
-				       (let ([header (open-output-string)]
-					     [cleanup #t] )
-					 (do ([c (read-char in) (read-char in)])
-					     ((char=? c #\newline)
-					      (set! ##sys#read-line-counter (add1 ##sys#read-line-counter)) )
-					   (if (eof-object? c)
-					       (quit "unexpected end of file - incorrect header")
-					       (write-char c header) ) )
-					 (when verbose (printf "Compiling UNIX script~%~!"))
-					 (case (parse-script-header (get-output-string header) in)
-					   [(script)
-					    (set! cleanup #f) ]
-					   [(chicken)]
-					   [(chicken-hygienic) 
-					    (install-macros) ]
-					   [(r5rs)
-					    (install-macros)
-					    (strictness #t #f) ]
-					   [(srfi0)
-					    (install-macros)
-					    (strictness #t #t) ]
-					   [(#f) (warning "unrecognized script header - might be corrupt")]
-					   [else (strictness #t #f)] ) ; r4rs & ieee
-					 (when cleanup
-					   (set! cleanup-forms
-					     (cons '(##sys#script-main '0 '0 main)
-						   cleanup-forms) ) ) ) ) ) ]
-				[(char=? #\@ c0)
-				 (read-line in)
-				 (when verbose (printf "Compiling Windows batch file~%~!")) ] )
-
+			       [in (check-and-open-input-file f)] )
 			  (do ([x (read-form in) (read-form in)])
 			      ((eof-object? x) 
 			       (close-checked-input-file in f) )
@@ -497,15 +446,11 @@
 		  [exps (append
 			 (map (lambda (ic) `(set! ,(cdr ic) ',(car ic))) immutable-constants)
 			 (map (lambda (n) `(##core#callunit ,n)) used-units)
-			 (if hygienic-at-rt
-			     '((##core#callunit "syntax_case")
-			       (##syncase#install-macro-package '#t '#t) )
-			     '() )
 			 (if emit-profile
 			     `((set! ,profile-info-vector-name 
 				 (##sys#register-profile-info
 				  ',plen
-				  ',(if unit-name #f "PROFILE") ) ) )
+				  ',(if unit-name #f profile-name))))
 			     '() )
 			 (map (lambda (pl)
 				`(##core#inline 
@@ -526,27 +471,23 @@
 					      (lambda () ,is) ) )
 			       (list is) ) )
 			 exps0
-			 (if hygienic
-			     (compute-module-exports module-export-list export-list)
-			     '() )
 			 (if (and (not unit-name) (not dynamic))
 			     cleanup-forms
-			     '() ) ) ] )
+			     '() )
+			 '((##core#undefined))) ] )
 
    	     (when (debugging '|N| "real name table:")
 	       (display-real-name-table) )
 	     (when (debugging 'n "line number database:")
 	       (display-line-number-database) )
 
-	     (when hygienic-at-rt
-	       (set! used-units (cons "syntax_case" used-units))
-	       (when unit-name
-		 (warning "enabling hygienic macros at runtime will apply to all programs using this unit") ) )
-
 	     (when (and block-compilation unit-name)
-	       (warning "compilation of library unit in block-mode - globals may not be accessible outside this unit"
+	       (warning "compilation of library unit `~a' in block-mode - globals may not be accessible outside this unit"
 			unit-name) )
 
+	     (when (and unit-name dynamic)
+	       (warning "library unit `~a' compiled in dynamic mode" unit-name) )
+
 	     (set! ##sys#line-number-database line-number-database-2)
 	     (set! line-number-database-2 #f)
 
@@ -566,7 +507,9 @@
 			    'lambda '(())
 			    (list (build-node-graph
 				   (canonicalize-begin-body exps) ) ) ) ] 
-		    [proc (or (user-pass-2) (and optimizers-used default-user-pass-2))] )
+		    [proc (user-pass-2)] )
+	       (when (debugging 'M "; requirements:")
+		 (pretty-print (hash-table->alist file-requirements)))
 	       (when proc
 		 (when verbose (printf "Secondary user pass...~%"))
 		 (begin-time)
@@ -575,14 +518,12 @@
 		   (print-db "analysis (u)" '|0| db 0)
 		   (end-time "pre-analysis (u)")
 		   (begin-time)
-		   (proc node0 
-			 (lambda (k p) (cut get db <> <>))
-			 (lambda (k p x) (cut put! db <> <>)) )
+		   (proc node0)
 		   (end-time "secondary user pass")
 		   (print-node "secondary user pass" '|U| node0) )
 		 (set! first-analysis #t) )
 
-	       (when (memq 'lambda-lift options)
+	       (when do-lambda-lifting
 		 (begin-time)
 		 (set! first-analysis #f)
 		 (let ([db (analyze-expression node0)])
@@ -610,6 +551,10 @@
 
 		   (begin-time)
 		   (let ([db (analyze-expression node2)])
+		     (when first-analysis
+		       (check-global-exports db)
+		       (when (memq 'u debugging-chicken)
+			 (dump-undefined-globals db)) )
 		     (set! first-analysis #f)
 		     (end-time "analysis")
 		     (print-db "analysis" '|4| db i)
@@ -646,7 +591,7 @@
 			    (let ([node3 (perform-closure-conversion node2 db)])
 			      (end-time "closure conversion")
 			      (print-db "final-analysis" '|8| db i)
-			      (when (debugging 'e "exported toplevel variables:")
+			      (when (memq 'e debugging-chicken)
 				(dump-exported-globals db) )
 			      (let ([upap (user-post-analysis-pass)])
 				(when upap 
@@ -683,23 +628,3 @@
                                 (compiler-cleanup-hook)
                                 (when verbose 
                                   (printf "compilation finished.~%~!") ) ) ) ] ) ) ) ) ) ) ) ) ) )
-
-
-;;; Parse script header:
-
-(define (parse-script-header line in)
-  (let ([len (string-length line)])
-    (define (cmp-suffix str)
-      (let ([index (- len (string-length str))])
-	(and (> index 0)
-	     (string=? str (substring line index len))
-	     (memq (string-ref line (sub1 index)) '(#\space #\/)) ) ) )
-    (cond [(cmp-suffix "-script") 'script]
-	  [(cmp-suffix "-script-meta")
-	   (read-line in)
-	   'script]
-	  [(cmp-suffix "scheme-r4rs") 'r4rs]
-	  [(cmp-suffix "scheme-r5rs") 'r5rs]
-	  [(cmp-suffix "scheme-srfi-0") 'srfi0]
-	  [(cmp-suffix "scheme-ieee-1178-1990") 'ieee]
-	  [else #f] ) ) )
