@@ -1,6 +1,6 @@
 ;;;; c-platform.scm - Platform specific parameters and definitions
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -37,7 +37,7 @@
 
 
 #{compiler
-  build-information compiler-arguments process-command-line
+  compiler-arguments process-command-line
   default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
@@ -56,7 +56,7 @@
   reorganize-recursive-bindings substitution-table simplify-named-call find-inlining-candidates perform-inlining!
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
   transform-direct-lambdas! decompose-lambda-list rewrite
-  debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
+  debugging-chicken bomb check-signature posq stringify symbolify flonum? build-lambda-list
   string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
   put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
@@ -64,7 +64,7 @@
   simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database
-  copyright compiler-features default-declarations default-debugging-declarations units-used-by-default words-per-flonum
+  default-declarations default-debugging-declarations units-used-by-default words-per-flonum
   parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   target-include-file default-profiling-declarations
@@ -78,21 +78,16 @@
 
 ;;; Parameters:
 
-(define copyright "(c)2000-2004 Felix L. Winkelmann")
-
 (define default-optimization-passes 3)
 
-(define compiler-features 
-  '(target-has-switch foreign-interface apply-simplifications) )
-
 (define default-declarations
   '((always-bound
      ##sys#standard-input ##sys#standard-output ##sys#standard-error)
     (bound-to-procedure
-     ##sys#for-each ##sys#map ##sys#print
+     ##sys#for-each ##sys#map ##sys#print ##sys#setter
      ##sys#setslot ##sys#dynamic-wind ##sys#call-with-values ##sys#match-error
      ##sys#start-timer ##sys#stop-timer ##sys#gcd ##sys#lcm ##sys#make-promise ##sys#structure? ##sys#slot 
-     ##sys#allocate-vector ##sys#list->vector
+     ##sys#allocate-vector ##sys#list->vector ##sys#block-ref ##sys#block-set!
      ##sys#list ##sys#cons ##sys#append ##sys#vector ##sys#foreign-char-argument ##sys#foreign-fixnum-argument
      ##sys#foreign-flonum-argument ##sys#error ##sys#peek-c-string ##sys#peek-nonnull-c-string 
      ##sys#peek-and-free-c-string ##sys#peek-and-free-nonnull-c-string
@@ -126,16 +121,18 @@
 (define target-include-file "chicken.h")
 
 (define valid-compiler-options
-  '(-help help version verbose explicit-use quiet no-trace no-warnings unsafe block assume-self-calls
-    check-syntax hygienic to-stdout usual-integrations case-insensitive emit-debug-info profile
-    fixnum-arithmetic disable-interrupts optimize-leaf-routines debug-calls syntax syntax-at-run-time
-    debug-loops strict strict-srfi-0 strict-reader lambda-lift run-time-macros
-    disable-stack-overflow-checks disable-c-syntax-checks unsafe-libraries
-    hygienic-at-run-time analyze-only strict-letrec dynamic ffi) )
+  '(-help h help version verbose explicit-use quiet no-trace no-warnings unsafe block
+    check-syntax to-stdout no-usual-integrations case-insensitive no-lambda-info profile inline
+    fixnum-arithmetic disable-interrupts optimize-leaf-routines
+    lambda-lift run-time-macros tag-pointers accumulate-profile
+    disable-stack-overflow-checks disable-c-syntax-checks unsafe-libraries raw 
+    emit-external-prototypes-first track-scheme
+    analyze-only dynamic ffi ffi-custom ffi-parse ffi-no-include extension) )
 
 (define valid-compiler-options-with-argument
-  '(debug output-file include-path heap-size stack-size unit uses keyword-style require-for-syntax
-    prelude postlude prologue epilogue nursery extend feature no-feature compress-literals split-level
+  '(debug output-file include-path heap-size stack-size unit uses keyword-style require-extension 
+	  inline-limit profile-name
+    prelude postlude prologue epilogue nursery extend feature compress-literals split-level
     heap-growth heap-shrinkage heap-initial-size ffi-define ffi-include-path split) )
 
 
@@ -163,31 +160,38 @@
 
 (define default-extended-bindings
   '(bitwise-and bitwise-ior bitwise-xor bitwise-not add1 sub1 fx+ fx- fx* fx/ fxmod
-    fx= fx> fx< fx>= fx<= fixnum? fxneg fxmax fxmin identity
-    arithmetic-shift signum void flush-output thread-specific thread-specific-set!
+    fx= fx> fx< fx>= fx<= fixnum? fxneg fxmax fxmin identity fp+ fp- fp* fp/ fpmin fpmax fpneg
+    fp> fp< fp= fp>= fp<= atom? fxand fxnot fxior fxxor fxshr fxshl
+    arithmetic-shift void flush-output thread-specific thread-specific-set!
     not-pair? null-list? print print* error cpu-time proper-list? call/cc
     u8vector->byte-vector s8vector->byte-vector u16vector->byte-vector s16vector->byte-vector 
     u32vector->byte-vector 
     s32vector->byte-vector byte-vector-length block-ref block-set! number-of-slots
     f32vector->byte-vector f64vector->byte-vector byte-vector-ref byte-vector-set!
+    hash-table-ref
     first second third fourth make-record-instance
-    locative-ref locative-set! locative->object locative?
+    u8vector-length s8vector-length u16vector-length s16vector-length u32vector-length s32vector-length
+    f32vector-length f64vector-length setter
+    u8vector-ref s8vector-ref u16vector-ref s16vector-ref u32vector-ref s32vector-ref
+    f32vector-ref f64vector-ref
+    u8vector-set! s8vector-set! u16vector-set! s16vector-set! u32vector-set! s32vector-set!
+    locative-ref locative-set! locative->object locative? global-ref
     null-pointer? pointer->object) )
 
 (define internal-bindings
-  '(##sys#slot ##sys#setslot
+  '(##sys#slot ##sys#setslot ##sys#block-ref ##sys#block-set!
     ##sys#call-with-current-continuation ##sys#size ##sys#byte ##sys#setbyte
     ##sys#pointer? ##sys#generic-structure? ##sys#structure? ##sys#check-structure
     ##sys#check-exact ##sys#check-number ##sys#check-list ##sys#check-pair ##sys#check-string ##sys#check-symbol 
     ##sys#check-char ##sys#check-vector ##sys#check-byte-vector ##sys#list
     ##sys#call-with-values ##sys#fits-in-int? ##sys#fits-in-unsigned-int? ##sys#flonum-in-fixnum-range? 
     ##sys#fudge ##sys#immediate? ##sys#direct-return
-    ##sys#make-structure ##sys#apply
-    ##sys#bytevector? ##sys#make-vector
+    ##sys#make-structure ##sys#apply ##sys#apply-values ##sys#continuation-graft
+    ##sys#bytevector? ##sys#make-vector ##sys#setter
     ##sys#foreign-char-argument ##sys#foreign-fixnum-argument ##sys#foreign-flonum-argument
     ##sys#foreign-block-argument ##sys#foreign-number-vector-argument
     ##sys#foreign-string-argument ##sys#foreign-pointer-argument 
-    ##sys#foreign-integer-argument ##sys#foreign-unsigned-integer-argument
+    ##sys#foreign-integer-argument ##sys#foreign-unsigned-integer-argument ##sys#double->number
     ##sys#peek-fixnum ##sys#setislot ##sys#poke-integer ##sys#permanent? ##sys#values ##sys#poke-double
     ##sys#intern-symbol ##sys#make-symbol ##sys#null-pointer? ##sys#peek-byte) )
 
@@ -209,9 +213,13 @@
 (define non-foldable-extended-bindings
   '(##sys#slot ##sys#setslot ##sys#call-with-current-continuation ##sys#fudge flush-output print void
     u8vector->byte-vector s8vector->byte-vector u16vector->byte-vector s16vector->byte-vector u32vector->byte-vector 
-    s32vector->byte-vector ##sys#make-structure print* ##sys#make-vector ##sys#apply ##sys#setislot
+    s32vector->byte-vector ##sys#make-structure print* ##sys#make-vector ##sys#apply ##sys#setislot ##sys#block-ref
     f32vector->byte-vector f64vector->byte-vector ##sys#byte ##sys#setbyte byte-vector-ref byte-vector-set!
-    ##sys#intern-symbol ##sys#make-symbol make-record-instance error cpu-time) )
+    u8vector-length s8vector-length u16vector-length s16vector-length u32vector-length s32vector-length
+    f32vector-length f64vector-length ##sys#apply-values ##sys#setter setter
+    u8vector-ref s8vector-ref u16vector-ref s16vector-ref
+    u8vector-set! s8vector-set! u16vector-set! s16vector-set! u32vector-set! s32vector-set!
+    ##sys#intern-symbol ##sys#make-symbol make-record-instance error cpu-time ##sys#block-set!) )
 
 (define foldable-extended-bindings
   (lset-difference
@@ -356,7 +364,7 @@
  'eqv? 8
  (lambda (db classargs cont callargs)
    ;; (eqv? <var> <var>) -> (quote #t)
-   ;; (eqv? ...) -> (##core#inline "C_eqp" ...) [fixnum-mode, or one argument is a constant and not a flonum]
+   ;; (eqv? ...) -> (##core#inline "C_eqp" ...) [one argument is a constant and not a flonum]
    (and (= (length callargs) 2)
 	(let ([arg1 (first callargs)]
 	      [arg2 (second callargs)] )
@@ -364,8 +372,7 @@
 		   (eq? '##core#variable (node-class arg2))
 		   (equal? (node-parameters arg1) (node-parameters arg2))
 		   (make-node '##core#call '(#t) (list cont (qnode #t))) )
-	      (and (or (eq? number-type 'fixnum)
-		       (and (eq? 'quote (node-class arg1))
+	      (and (or (and (eq? 'quote (node-class arg1))
 			    (not (flonum? (first (node-parameters arg1)))) )
 		       (and (eq? 'quote (node-class arg2))
 			    (not (flonum? (first (node-parameters arg2)))) ) )
@@ -403,17 +410,32 @@
   (define (rewrite-apply db classargs cont callargs)
     ;; (apply <fn> <x1> ... '(<y1> ...)) -> (<fn> <x1> ... '<y1> ...)
     ;; (apply ...) -> ((##core#proc "C_apply") ...)
-    (let ([lastarg (last callargs)])
-      (if (eq? 'quote (node-class lastarg))
-	  (make-node
-	   '##core#call '(#f)
-	   (cons* (first callargs)
-		  cont 
-		  (append (cdr (butlast callargs)) (map qnode (first (node-parameters lastarg)))) ) )
-	  (make-node
-	   '##core#call '(#t)
-	   (cons* (make-node '##core#proc '("C_apply" #t) '())
-		  cont callargs) ) ) ) )
+    ;; (apply values <lst>) -> ((##core#proc "C_apply_values") lst)
+    ;; (apply ##sys#values <lst>) -> ((##core#proc "C_apply_values") lst)
+    (and (pair? callargs)
+	 (let ([lastarg (last callargs)]
+	       [proc (car callargs)] )
+	   (if (eq? 'quote (node-class lastarg))
+	       (make-node
+		'##core#call '(#f)
+		(cons* (first callargs)
+		       cont 
+		       (append (cdr (butlast callargs)) (map qnode (first (node-parameters lastarg)))) ) )
+	       (or (and (eq? '##core#variable (node-class proc))
+			(= 2 (length callargs))
+			(let ([name (car (node-parameters proc))])
+			  (and (memq name '(values ##sys#values))
+			       (or (get db name 'standard-binding)
+				   (get db name 'extended-binding) )
+			       (make-node
+				'##core#call '(#t)
+				(list (make-node '##core#proc '("C_apply_values" #t) '())
+				      cont
+				      (cadr callargs) ) ) ) ) ) 
+		   (make-node
+		    '##core#call '(#t)
+		    (cons* (make-node '##core#proc '("C_apply" #t) '())
+			   cont callargs) ) ) ) ) ) )
   (rewrite 'apply 8 rewrite-apply)
   (rewrite '##sys#apply 8 rewrite-apply) )
 
@@ -470,6 +492,7 @@
 (rewrite '##sys#call-with-values 13 "C_call_with_values" #t)
 (rewrite 'cpu-time 13 "C_cpu_time" #t)
 (rewrite 'locative-ref 13 "C_locative_ref" #t)
+(rewrite '##sys#continuation-graft 13 "C_continuation_graft" #t)
 
 (rewrite 'caar 2 1 "C_u_i_caar" #f #f)
 (rewrite 'cdar 2 1 "C_u_i_cdar" #f #f)
@@ -518,6 +541,7 @@
 (rewrite 'symbol? 2 1 "C_i_symbolp" #t #f)
 (rewrite 'vector? 2 1 "C_i_vectorp" #t #f)
 (rewrite 'pair? 2 1 "C_i_pairp" #t "C_notvemptyp")
+(rewrite 'atom? 2 1 "C_i_atomp" #t #f)
 (rewrite 'procedure? 2 1 "C_i_closurep" #t #f)
 (rewrite 'port? 2 1 "C_i_portp" #t #f)
 (rewrite 'boolean? 2 1 "C_booleanp" #t #f)
@@ -550,8 +574,9 @@
 (rewrite 'char>=? 2 2 "C_fixnum_greater_or_equal_p" #t #f)
 (rewrite 'char<=? 2 2 "C_fixnum_less_or_equal_p" #t #f)
 (rewrite '##sys#slot 2 2 "C_slot" #t #f)		; consider as safe, the primitive is unsafe anyway.
+(rewrite '##sys#block-ref 2 2 "C_i_block_ref" #t #f) ; must be safe for pattern matcher
 (rewrite '##sys#size 2 1 "C_block_size" #t #f)
-(rewrite 'bitwise-not 2 1 "C_fixnum_not" #t #f)
+(rewrite 'fxnot 2 1 "C_fixnum_not" #t #f)
 (rewrite 'fx* 2 2 "C_fixnum_times" #t #f)
 (rewrite 'fx/ 2 2 "C_fixnum_divide" #f #f)
 (rewrite 'fxmod 2 2 "C_fixnum_modulo" #f #f)
@@ -560,8 +585,15 @@
 (rewrite 'fx< 2 2 "C_fixnum_lessp" #t #f)
 (rewrite 'fx>= 2 2 "C_fixnum_greater_or_equal_p" #t #f)
 (rewrite 'fx<= 2 2 "C_fixnum_less_or_equal_p" #t #f)
+(rewrite 'fp= 2 2 "C_flonum_equalp" #t #f)
+(rewrite 'fp> 2 2 "C_flonum_greaterp" #t #f)
+(rewrite 'fp< 2 2 "C_flonum_lessp" #t #f)
+(rewrite 'fp>= 2 2 "C_flonum_greater_or_equal_p" #t #f)
+(rewrite 'fp<= 2 2 "C_flonum_less_or_equal_p" #t #f)
 (rewrite 'fxmax 2 2 "C_i_fixnum_max" #t #f)
 (rewrite 'fxmin 2 2 "C_i_fixnum_min" #t #f)
+(rewrite 'fpmax 2 2 "C_i_flonum_max" #t #f)
+(rewrite 'fpmin 2 2 "C_i_flonum_min" #t #f)
 (rewrite 'char-numeric? 2 1 "C_u_i_char_numericp" #t #f)
 (rewrite 'char-alphabetic? 2 1 "C_u_i_char_alphabeticp" #t #f)
 (rewrite 'char-whitespace? 2 1 "C_u_i_char_whitespacep" #t #f)
@@ -572,7 +604,7 @@
 (rewrite 'list-tail 2 2 "C_i_list_tail" #t #f)
 (rewrite '##sys#structure? 2 2 "C_i_structurep" #t #f)
 (rewrite '##sys#bytevector? 2 2 "C_bytevectorp" #t #f)
-(rewrite 'block-ref 2 2 "C_slot" #f #f)
+(rewrite 'block-ref 2 2 "C_slot" #f #f)	; ok to be unsafe, lolevel is anyway
 (rewrite 'number-of-slots 2 1 "C_block_size" #f #f)
 
 (rewrite 'assv 14 'fixnum 2 "C_i_assq" "C_u_i_assq")
@@ -592,6 +624,29 @@
 (rewrite 'abs 14 'fixnum 1 "C_fixnum_abs" "C_fixnum_abs")
 (rewrite 'abs 16 1 "C_a_i_abs" #t words-per-flonum)
 
+(rewrite 'bitwise-xor 21 0 "C_fixnum_xor" "C_fixnum_xor" "C_a_i_bitwise_xor" words-per-flonum)
+(rewrite 'bitwise-and 21 -1 "C_fixnum_and" "C_u_fixnum_and" "C_a_i_bitwise_and" words-per-flonum)
+(rewrite 'bitwise-ior 21 0 "C_fixnum_or" "C_u_fixnum_or" "C_a_i_bitwise_ior" words-per-flonum)
+
+(rewrite 'bitwise-not 22 1 "C_a_i_bitwise_not" #t words-per-flonum "C_fixnum_not")
+
+(rewrite 'fp+ 16 2 "C_a_i_flonum_plus" #t words-per-flonum)
+(rewrite 'fp- 16 2 "C_a_i_flonum_difference" #t words-per-flonum)
+(rewrite 'fp* 16 2 "C_a_i_flonum_times" #t words-per-flonum)
+(rewrite 'fp/ 16 2 "C_a_i_flonum_quotient" #t words-per-flonum)
+(rewrite 'fpneg 16 1 "C_a_i_flonum_negate" #t words-per-flonum)
+
+(rewrite 'exp 16 1 "C_a_i_exp" #t words-per-flonum)
+(rewrite 'sin 16 1 "C_a_i_sin" #t words-per-flonum)
+(rewrite 'cos 16 1 "C_a_i_cos" #t words-per-flonum)
+(rewrite 'tan 16 1 "C_a_i_tan" #t words-per-flonum)
+(rewrite 'log 16 1 "C_a_i_log" #t words-per-flonum)
+(rewrite 'asin 16 1 "C_a_i_asin" #t words-per-flonum)
+(rewrite 'acos 16 1 "C_a_i_acos" #t words-per-flonum)
+(rewrite 'atan 16 1 "C_a_i_atan" #t words-per-flonum)
+(rewrite 'sqrt 16 1 "C_a_i_sqrt" #t words-per-flonum)
+(rewrite 'atan 16 2 "C_a_i_atan2" #t words-per-flonum)
+
 (rewrite 'zero? 5 "C_eqp" 0 'fixnum)
 (rewrite 'zero? 2 1 "C_i_zerop" #t #f)
 (rewrite 'zero? 2 1 "C_u_i_zerop" #f #f)
@@ -640,9 +695,11 @@
 (rewrite '>= 9 "C_fixnum_greater_or_equal_p" "C_flonum_greater_or_equal_p" #t #f)
 (rewrite '<= 9 "C_fixnum_less_or_equal_p" "C_flonum_less_or_equal_p" #t #f)
 
+(rewrite 'setter 11 1 '##sys#setter #t)
 (rewrite 'for-each 11 2 '##sys#for-each #t)
 (rewrite 'map 11 2 '##sys#map #t)
 (rewrite 'block-set! 11 3 '##sys#setslot #t)
+(rewrite '##sys#block-set! 11 3 '##sys#setslot #f)
 (rewrite 'make-record-instance 11 #f '##sys#make-structure #f)
 (rewrite 'substring 11 3 '##sys#substring #f)
 (rewrite 'string-append 11 2 '##sys#string-append #f)
@@ -690,9 +747,7 @@
 (rewrite '##sys#round 13 "C_flonum_round" #t)
 (rewrite '##sys#allocate-vector 13 "C_allocate_vector" #t)
 (rewrite '##sys#ensure-heap-reserve 13 "C_ensure_heap_reserve" #t)
-(rewrite '##sys#call-host 13 "C_call_host" #t)
-(rewrite '##sys#host-data 13 "C_host_data" #t)
-(rewrite '##sys#set-host-data! 13 "C_set_host_data" #t)
+(rewrite 'return-to-host 13 "C_return_to_host" #t)
 (rewrite '##sys#context-switch 13 "C_context_switch" #t)
 (rewrite '##sys#intern-symbol 13 "C_string_to_symbol" #t)
 (rewrite '##sys#make-symbol 13 "C_make_symbol" #t)
@@ -726,37 +781,42 @@
  (lambda (db classargs cont callargs)
    ;; (##sys#setslot <x> <y> <immediate>) -> (##core#inline "C_i_set_i_slot" <x> <y> <i>)
    ;; (##sys#setslot <x> <y> <z>) -> (##core#inline "C_i_setslot" <x> <y> <z>)
-   (make-node 
-    '##core#call '(#t)
-    (list cont
-	  (make-node
-	   '##core#inline
-	   (let ([val (third callargs)])
-	     (if (and (eq? 'quote (node-class val))
-		      (immediate? (first (node-parameters val))) ) 
-		 '("C_i_set_i_slot")
-		 '("C_i_setslot") ) )
-	   callargs) ) ) ) )
+   (and (= (length callargs) 3)
+	(make-node 
+	 '##core#call '(#t)
+	 (list cont
+	       (make-node
+		'##core#inline
+		(let ([val (third callargs)])
+		  (if (and (eq? 'quote (node-class val))
+			   (immediate? (first (node-parameters val))) ) 
+		      '("C_i_set_i_slot")
+		      '("C_i_setslot") ) )
+		callargs) ) ) ) ) )
 
 (rewrite 'fx+ 17 2 "C_fixnum_plus" "C_u_fixnum_plus")
 (rewrite 'fx- 17 2 "C_fixnum_difference" "C_u_fixnum_difference")
+(rewrite 'fxshl 17 2 "C_fixnum_shift_left")
+(rewrite 'fxshr 17 2 "C_fixnum_shift_right")
 (rewrite 'fxneg 17 1 "C_fixnum_negate" "C_u_fixnum_negate")
-(rewrite 'bitwise-xor 19 0 "C_fixnum_xor" "C_fixnum_xor" #t)
-(rewrite 'bitwise-and 19 -1 "C_fixnum_and" "C_u_fixnum_and" #t)
-(rewrite 'bitwise-ior 19 0 "C_fixnum_or" "C_u_fixnum_or" #t)
+(rewrite 'fxxor 17 2 "C_fixnum_xor" "C_fixnum_xor")
+(rewrite 'fxand 17 2 "C_fixnum_and" "C_u_fixnum_and")
+(rewrite 'fxior 17 2 "C_fixnum_or" "C_u_fixnum_or")
 
 (rewrite
  'arithmetic-shift 8
  (lambda (db classargs cont callargs)
    ;; (arithmetic-shift <x> <-int>) -> (##core#inline "C_fixnum_shift_right" <x> -<int>)
    ;; (arithmetic-shift <x> <+int>) -> (##core#inline "C_fixnum_shift_left" <x> <int>)
-   ;; _ -> (##core#inline "C_i_arithmetic_shift" <x> <y>)
+   ;; _ -> (##core#inline "C_a_i_arithmetic_shift" <x> <y>)
+   ;; not in fixnum-mode: _ -> (##core#inline_allocate ("C_a_i_arithmetic_shift" words-per-flonum) <x> <y>)
    (and (= 2 (length callargs))
 	(let ([val (second callargs)])
 	  (make-node
 	   '##core#call '(#t)
 	   (list cont
 		 (or (and-let* ([(eq? 'quote (node-class val))]
+				[(eq? number-type 'fixnum)]
 				[n (first (node-parameters val))]
 				[(fixnum? n)] )
 		       (if (negative? n)
@@ -766,7 +826,10 @@
 			   (make-node
 			    '##core#inline '("C_fixnum_shift_left")
 			    (list (first callargs) val) ) ) )
-		     (make-node '##core#inline '("C_i_arithmetic_shift") callargs) ) ) ) ) ) ) )
+		     (if (eq? number-type 'fixnum)
+			 (make-node '##core#inline '("C_i_fixnum_arithmetic_shift") callargs)
+			 (make-node '##core#inline_allocate (list "C_a_i_arithmetic_shift" words-per-flonum) 
+				    callargs) ) ) ) ) ) ) ) )
 
 (rewrite '##sys#byte 17 2 "C_subbyte")
 (rewrite '##sys#setbyte 17 3 "C_setbyte")
@@ -776,6 +839,7 @@
 (rewrite '##sys#setislot 17 3 "C_i_set_i_slot")
 (rewrite '##sys#poke-integer 17 3 "C_poke_integer")
 (rewrite '##sys#poke-double 17 3 "C_poke_double")
+(rewrite '##sys#double->number 17 1 "C_double_to_number")
 (rewrite 'string=? 17 2 "C_i_string_equal_p" "C_u_i_string_equal_p")
 (rewrite 'string-ci=? 17 2 "C_i_string_ci_equal_p")
 (rewrite '##sys#fudge 17 1 "C_fudge")
@@ -803,6 +867,27 @@
 (rewrite 'byte-vector-set! 2 3 "C_setbyte" #f #f)
 (rewrite 'byte-vector-length 2 1 "C_block_size" #f #f)
 
+(rewrite 'u8vector-ref 2 2 "C_u_i_u8vector_ref" #f #f)
+(rewrite 's8vector-ref 2 2 "C_u_i_s8vector_ref" #f #f)
+(rewrite 'u16vector-ref 2 2 "C_u_i_u16vector_ref" #f #f)
+(rewrite 's16vector-ref 2 2 "C_u_i_s16vector_ref" #f #f)
+
+(rewrite 'u8vector-set! 2 3 "C_u_i_u8vector_set" #f #f)
+(rewrite 's8vector-set! 2 3 "C_u_i_s8vector_set" #f #f)
+(rewrite 'u16vector-set! 2 3 "C_u_i_u16vector_set" #f #f)
+(rewrite 's16vector-set! 2 3 "C_u_i_s16vector_set" #f #f)
+(rewrite 'u32vector-set! 2 3 "C_u_i_u32vector_set" #f #f)
+(rewrite 's32vector-set! 2 3 "C_u_i_s32vector_set" #f #f)
+
+(rewrite 'u8vector-length 2 1 "C_u_i_8vector_length" #f #f)
+(rewrite 's8vector-length 2 1 "C_u_i_8vector_length" #f #f)
+(rewrite 'u16vector-length 2 1 "C_u_i_16vector_length" #f #f)
+(rewrite 's16vector-length 2 1 "C_u_i_16vector_length" #f #f)
+(rewrite 'u32vector-length 2 1 "C_u_i_32vector_length" #f #f)
+(rewrite 's32vector-length 2 1 "C_u_i_32vector_length" #f #f)
+(rewrite 'f32vector-length 2 1 "C_u_i_32vector_length" #f #f)
+(rewrite 'f64vector-length 2 1 "C_u_i_64vector_length" #f #f)
+
 (rewrite 'not-pair? 17 1 "C_i_not_pair_p")
 (rewrite 'null-list? 17 1 "C_i_null_list_p" "C_i_nullp")
 
@@ -858,7 +943,7 @@
 		    (decompose-lambda-list 
 		     llist
 		     (lambda (vars argc rest)
-		       (and (<= 1 argc 2)
+		       (and (= argc 2)
 			    (let ([var (or rest (second llist))])
 			      (and (not (get db var 'references))
 				   (not (get db var 'assigned)) 
@@ -869,3 +954,48 @@
 				      (list val cont (qnode #f)) ) ) ) ) ) ) ) ) ) ) ) ) )
   (rewrite 'call-with-current-continuation 8 rewrite-call/cc)
   (rewrite 'call/cc 8 rewrite-call/cc) )
+
+(declare (hide setter-map))
+
+(define setter-map
+  '((car . set-car!)
+    (cdr . set-cdr!)
+    (hash-table-ref . hash-table-set!)
+    (block-ref . block-set!)
+    (byte-vector-ref . byte-vector-set!)
+    (locative-ref . locative-set!)
+    (u8vector-ref . u8vector-set!)
+    (s8vector-ref . s8vector-set!)
+    (u16vector-ref . u16vector-set!)
+    (s16vector-ref . s16vector-set!)
+    (u32vector-ref . u32vector-set!)
+    (s32vector-ref . s32vector-set!)
+    (f32vector-ref . f32vector-set!)
+    (f64vector-ref . f64vector-set!)
+    (pointer-u8-ref . pointer-u8-set!)
+    (pointer-s8-ref . pointer-s8-set!)
+    (pointer-u16-ref . pointer-u16-set!)
+    (pointer-s16-ref . pointer-s16-set!)
+    (pointer-u32-ref . pointer-u32-set!)
+    (pointer-s32-ref . pointer-s32-set!)
+    (pointer-f32-ref . pointer-f32-set!)
+    (pointer-f64-ref . pointer-f64-set!)
+    (string-ref . string-set!)
+    (global-ref . global-set!)
+    (vector-ref . vector-set!) ) )
+
+(rewrite
+ '##sys#setter 8
+ (lambda (db classargs cont callargs)
+   ;; (setter <known-getter>) -> <known-setter>
+   (and (= 1 (length callargs))
+	(let ((arg (car callargs)))
+	  (and (eq? '##core#variable (node-class arg))
+	       (let ((sym (car (node-parameters arg))))
+		 (and (or (get db sym 'standard-binding)
+			  (get db sym 'extended-binding))
+		      (and-let* ((a (assq sym setter-map)))
+			(make-node
+			 '##core#call '(#t)
+			 (list cont (varnode (cdr a))) ) ) ) ) ) ) ) ) )
+			       
