@@ -1,6 +1,6 @@
 ;;;; chicken-more-macros.scm - More syntax extensions
 ;
-; Copyright (c) 2000-2002, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -33,12 +33,7 @@
 ; Germany
 
 
-(##sys#provide 'moremacros)
-
-(if (memq #:hygienic-macros ##sys#features)
-    (begin
-      (load (##sys#resolve-include-filename "chicken-highlevel-macros"))
-      (##sys#abort-load) ) )
+(##sys#provide 'chicken-more-macros)
 
 
 ;;; Non-standard macros:
@@ -52,6 +47,7 @@
      (##sys#check-syntax 'define-record name 'symbol)
      (##sys#check-syntax 'define-record slots '#(symbol 0))
      (let ([prefix (symbol->string name)]
+	   [setters (memq #:record-setters ##sys#features)]
 	   [nsprefix (##sys#qualified-symbol-prefix name)] )
        `(begin
 	  (define ,(##sys#string->qualified-symbol nsprefix (string-append "make-" prefix))
@@ -61,43 +57,27 @@
 	  ,@(let mapslots ((slots slots) (i 1))
 	      (if (eq? slots '())
 		  slots
-		  (let ((slotname (symbol->string (##sys#slot slots 0))))
+		  (let* ((slotname (symbol->string (##sys#slot slots 0)))
+			 (setr (##sys#string->qualified-symbol nsprefix (string-append prefix "-" slotname "-set!")))
+			 (getr (##sys#string->qualified-symbol nsprefix (string-append prefix "-" slotname)) ) )
 		    (cons
 		     `(begin
-			(define ,(##sys#string->qualified-symbol nsprefix (string-append prefix "-" slotname))
-			  (lambda (x) 
-			    (##sys#check-structure x ',name)
-			    (##sys#slot x ,i) ) )
-			(define ,(##sys#string->qualified-symbol nsprefix (string-append prefix "-" slotname "-set!"))
+			(define ,setr
 			  (lambda (x val)
 			    (##sys#check-structure x ',name)
-			    (##sys#setslot x ,i val) ) ) ) 
+			    (##sys#block-set! x ,i val) ) )
+			(define ,getr
+			  ,(if setters
+			       `(getter-with-setter
+				 (lambda (x) 
+				   (##sys#check-structure x ',name)
+				   (##sys#block-ref x ,i) )
+				 ,setr)
+			       `(lambda (x)
+				  (##sys#check-structure x ',name)
+				  (##sys#block-ref x ,i) ) ) ) )
 		     (mapslots (##sys#slot slots 1) (fx+ i 1)) ) ) ) ) ) ) ) ) )
 
-(##sys#register-macro
- 'define-macro
- (lambda (head . body)
-   (let ([name head]
-	 [llist #f]
-	 [proc #f] )
-     (cond [(symbol? head)
-	    (cond [(and (pair? body) (symbol? (car body)))
-		   (set! proc (car body)) ]
-		  [else
-		   (##sys#check-syntax 'define-macro body '((lambda lambda-list . #(_ 1))))
-		   (set! llist (cadar body))
-		   (set! proc (cddar body)) ] ) ]
-	   [else
-	    (##sys#check-syntax 'define-macro head '(symbol . lambda-list))
-	    (##sys#check-syntax 'define-macro body '#(_ 1)) 
-	    (set! name (car head))
-	    (set! llist (cdr head))
-	    (set! proc body) ] )
-     `(,(if ##sys#enable-runtime-macros '##core#elaborationtimetoo '##core#elaborationtimeonly)
-       ,(cond [(symbol? proc) `(##sys#copy-macro ',proc ',name)]
-	      [(symbol? llist) `(##sys#register-macro-2 ',name (lambda (,llist) ,@proc))]
-	      [else `(##sys#register-macro ',name (lambda ,llist ,@proc))] ) ) ) ) )
-
 (##sys#register-macro
  'receive
  (lambda (vars . rest)
@@ -135,7 +115,7 @@
        [reverse reverse] )
    (lambda (filename)
      `(begin
-	,@(with-input-from-file (##sys#resolve-include-filename filename)
+	,@(with-input-from-file (##sys#resolve-include-filename filename #t)
 	    (lambda ()
 	      (do ([x (read) (read)]
 		   [xs '() (cons x xs)] )
@@ -206,8 +186,8 @@
 	   (begin
 	     (case (##sys#slot ss 0)
 	       [(eval) (set! e #t)]
-	       [(load) (set! l #t)]
-	       [(compile) (set! c #t)]
+	       [(load run-time) (set! l #t)]
+	       [(compile compile-time) (set! c #t)]
 	       [else (##sys#error "invalid situation specifier" (##sys#slot ss 0))] )
 	     (loop (##sys#slot ss 1)) ) ) )
      (if (memq '#:compiling ##sys#features)
@@ -266,22 +246,46 @@
 
 (##sys#register-macro-2
  'let-values
- (lambda (form)
-   (##sys#check-syntax 'let-values form '(#(_ 0) . #(_ 1)))
-   (let* ([vbindings (car form)]
-	  [body (cdr form)]
-	  [llists (map (lambda (x) (car x)) vbindings)]
-	  [vars (apply ##sys#append llists)]
-	  [aliases (map (lambda (v) (cons v (gensym v))) vars)] 
-	  [lookup (lambda (v) (cdr (assq v aliases)))]
-	  [llists2 (map (lambda (llist) (map lookup llist)) llists)] )
-     (let fold ([llists llists]
-		[exps (map (lambda (x) (cadr x)) vbindings)]
-		[llists2 llists2] )
-       (if (null? llists)
-	   `(let ,(map (lambda (v) (##sys#list v (lookup v))) vars) ,@body)
-	   `(##sys#call-with-values (lambda () ,(car exps))
-	      (lambda ,(car llists2) ,(fold (cdr llists) (cdr exps) (cdr llists2))) ) ) ) ) ) )
+ (letrec ((append* (lambda (il l)
+		      (if (not (pair? il))
+			  (cons il l)
+			  (cons (car il)
+				(append* (cdr il) l)))))
+	    (map* (lambda (proc l)
+		    (cond ((null? l) '())
+			  ((not (pair? l)) (proc l))
+			  (else (cons (proc (car l)) (map* proc (cdr l))))))))
+   (lambda (form)
+     (##sys#check-syntax 'let-values form '(#(_ 0) . #(_ 1)))
+     (let* ([vbindings (car form)]
+	    [body (cdr form)]
+	    [llists (map car vbindings)]
+	    [vars (let loop ((llists llists) (acc '()))
+		    (if (null? llists)
+			acc
+			(let* ((llist (car llists))
+			       (new-acc
+				(cond ((list? llist) (append llist acc))
+				      ((pair? llist) (append* llist acc))
+				      (else (cons llist acc)))))
+			  (loop (cdr llists) new-acc))))]
+	    [aliases (map (lambda (v) (cons v (gensym v))) vars)]
+	    [lookup (lambda (v) (cdr (assq v aliases)))]
+	    [llists2 (let loop ((llists llists) (acc '()))
+		       (if (null? llists)
+			   (reverse acc)
+			   (let* ((llist (car llists))
+				  (new-acc
+				   (cond ((not (pair? llist)) (cons (lookup llist) acc))
+					 (else (cons (map* lookup llist) acc)))))
+			     (loop (cdr llists) new-acc))))])
+       (let fold ([llists llists]
+		  [exps (map (lambda (x) (cadr x)) vbindings)]
+		  [llists2 llists2] )
+	 (if (null? llists)
+	     `(let ,(map (lambda (v) (##sys#list v (lookup v))) vars) ,@body)
+	     `(##sys#call-with-values (lambda () ,(car exps))
+				      (lambda ,(car llists2) ,(fold (cdr llists) (cdr exps) (cdr llists2))) ) ) ) ) ) ) )
 
 (##sys#register-macro-2
  'let*-values
@@ -292,8 +296,8 @@
      (let fold ([vbindings vbindings])
        (if (null? vbindings)
 	   `(let () ,@body)
-	   `(##sys#call-with-values (lambda () ,(cadar vbindings))
-	      (lambda ,(caar vbindings) ,(fold (cdr vbindings))) ) ) ) ) ) )
+	   `(let-values (,(car vbindings))
+	      ,(fold (cdr vbindings))) ) ) ) ) )
 
 (##sys#register-macro-2 
  'letrec-values
@@ -328,8 +332,13 @@
                     [val (if (pair? head)
                            `(lambda ,(cdr head) ,@(cdr xs))
                            (cadr xs) ) ] )
+	       (when (or (not (pair? val)) (not (eq? 'lambda (car val))))
+		 (syntax-error 'define-inline "invalid substitution form - must be lambda"
+			       name) )
                (list (list 'quote name) val) ) ) ] )
-  (##sys#register-macro-2 'define-inline (lambda (form) `(##core#define-inline ,@(quotify-proc form 'define-inline)))) )
+  (##sys#register-macro-2 
+   'define-inline
+   (lambda (form) `(##core#define-inline ,@(quotify-proc form 'define-inline)))) )
 
 (##sys#register-macro-2
  'define-constant
@@ -536,7 +545,8 @@
   (define (make-if-tree vars defaulters body-proc rest rename)
     (let recur ((vars vars) (defaulters defaulters) (non-defaults '()))
       (if (null? vars)
-	  `(if (null? ,rest) (,body-proc . ,(reverse non-defaults))
+	  `(if (##core#check (null? ,rest))
+	       (,body-proc . ,(reverse non-defaults))
 	       (##sys#error (##core#immutable '"too many optional arguments") ,rest))
 	  (let ((v (car vars)))
 	    `(if (null? ,rest)
@@ -596,7 +606,7 @@
     `(let ((,var ,rest))
        (cond ((null? ,var) ,default-exp)
 	     ((null? (cdr ,var)) (car ,var))
-	     (else (##sys#error (##core#immutable '"too many optional arguments") ,var))))))
+	     (else (##core#check (##sys#error (##core#immutable '"too many optional arguments") ,var)))))))
 
 
 ;;; (LET-OPTIONALS* args ((var1 default1) ... [rest]) body1 ...)
@@ -621,7 +631,7 @@
     `(let ((,rvar ,args))
        ,(let loop ([args rvar] [vardefs var/defs])
 	  (if (null? vardefs)
-	      `(if (null? ,args)
+	      `(if (##core#check (null? ,args))
 		   (let () ,@body)
 		   (##sys#error (##core#immutable '"too many optional arguments") ,args) )
 	      (let ([head (car vardefs)])
@@ -634,17 +644,21 @@
 
 
 ;;; case-lambda (SRFI-16):
-;
-; - Note: this macro needs SRFI-1
 
 (define-macro (case-lambda . clauses)
+  (define (genvars n)
+    (let loop ([i 0])
+      (if (fx>= i n)
+	  '()
+	  (cons (gensym) (loop (fx+ i 1))) ) ) )
   (##sys#check-syntax 'case-lambda clauses '#(_ 0))
+  (require 'srfi-1)			; Urgh...
   (let* ((mincount (apply min (map (lambda (c)
 				     (##sys#decompose-lambda-list 
 				      (car c)
 				      (lambda (vars argc rest) argc) ) )
 				   clauses) ) ) 
-	 (minvars (list-tabulate mincount (lambda (z) (gensym))))
+	 (minvars (genvars mincount))
 	 (rvar (gensym)) 
 	 (lvar (gensym)) )
     `(lambda ,(append minvars rvar)
@@ -655,9 +669,14 @@
 	      (car c)
 	      (lambda (vars argc rest)
 		(##sys#check-syntax 'case-lambda (car c) 'lambda-list)
-		`(if (,(if rest 'fx>= 'fx=) ,lvar ,(fx- argc mincount))
+		`(if ,(let ([a2 (fx- argc mincount)])
+			(if rest
+			    (if (zero? a2)
+				#t
+				`(fx>= ,lvar ,a2) )
+			    `(fx= ,lvar ,a2) ) )
 		     ,(receive
-		       (vars1 vars2) (split-at (take vars argc) mincount)
+		       (vars1 vars2) (split-at! (take vars argc) mincount)
 		       (let ((bindings
 			      (let build ((vars2 vars2) (vrest rvar))
 				(if (null? vars2)
@@ -672,14 +691,11 @@
 					      (build '() vrest2) ) ) ) ) ) ) )
 			 (if (null? vars1)
 			     bindings
-			     `(let ,(zip vars1 minvars) ,bindings) ) ) )
+			     `(let ,(map list vars1 minvars) ,bindings) ) ) )
 		     ,body) ) ) )
-	   '(##sys#error (##core#immutable '"no matching clause in call to 'case-lambda' form"))
+	   '(##core#check (##sys#error (##core#immutable '"no matching clause in call to 'case-lambda' form")))
 	   clauses) ) ) ) )
 
-(eval-when (compile load eval)
-  (register-feature! 'srfi-16) )
-
 
 ;;; Record printing:
 
@@ -715,19 +731,19 @@
   (let gather ([args (##sys#slot head 1)]
 	       [specs '()]
 	       [vars '()] )
-    (if (or (not (pair? args)) (not (pair? (##sys#slot args 0))))
+    (if (or (not (pair? args)) 
+	    (memq (car args) '(#!optional #!key #!rest)) )
 	(let ([name (##sys#slot head 0)])
 	  `(##core#set! ,name
-	     (##tinyclos#add-global-method
-	      (##core#global-ref ,name)
-	      ',name
-	      (list ,@(reverse specs))
-	      (##core#named-lambda ,name (call-next-method ,@(reverse vars) ,@args) ,@body) ) ) )
+			(##tinyclos#add-global-method
+			 (##core#global-ref ,name)
+			 ',name
+			 (list ,@(reverse specs))
+			 (##core#named-lambda ,name (call-next-method ,@(reverse vars) ,@args) ,@body) ) ) )
 	(let ([arg (##sys#slot args 0)])
-	  (##sys#check-syntax 'define-method arg '(_ symbol))
 	  (gather (##sys#slot args 1)
-		  (cons (##sys#slot (##sys#slot arg 1) 0) specs)
-		  (cons (##sys#slot arg 0) vars) ) ) ) ) )
+		  (cons (if (pair? arg) (cadr arg) '<top>) specs)
+		  (cons (if (pair? arg) (car arg) arg) vars) ) ) ) ) )
 
 
 ;;; Exceptions:
@@ -786,15 +802,21 @@
        ,@(let loop ([slots slots] [i 1])
 	   (if (null? slots)
 	       '()
-	       (let ([slot (car slots)])
-		 `((define (,(cadr slot) x)
-		     (##sys#check-structure x ',t)
-		     (##sys#slot x ,i) )
-		   ,@(if (pair? (cddr slot))
+	       (let* ([slot (car slots)]
+		      (setters (memq #:record-setters ##sys#features))
+		      (setr? (pair? (cddr slot))) 
+		      (getr `(lambda (x)
+			       (##sys#check-structure x ',t)
+			       (##sys#block-ref x ,i) ) ) )
+		 `(,@(if setr?
 			 `((define (,(caddr slot) x y)
 			     (##sys#check-structure x ',t)
-			      (##sys#setslot x ,i y)) )
+			     (##sys#block-set! x ,i y)) )
 			 '() )
+		   (define ,(cadr slot) 
+		     ,(if (and setr? setters)
+			  `(getter-with-setter ,getr ,(caddr slot))
+			  getr) )
 		   ,@(loop (cdr slots) (add1 i)) ) ) ) ) ) ) )
 
 
@@ -853,13 +875,64 @@
 	   (let ([v (gensym)])
 	     (loop (cdr xs) vars (cons (list v (car xs)) bs) (cons v vals) #f) ) ] ) ) ) )
 
+
+;;; SRFI-13:
+
+(define-macro (let-string-start+end s-e-r proc s-exp args-exp . body)
+  (if (pair? (cddr s-e-r))
+      `(receive (,(caddr s-e-r) ,(car s-e-r) ,(cadr s-e-r))
+	   (string-parse-start+end ,proc ,s-exp ,args-exp)
+	 ,@body)
+      `(receive ,s-e-r
+	   (string-parse-final-start+end ,proc ,s-exp ,args-exp)
+	 ,@body) ) )
+
+
+;;; Extension helper:
+
+(define-macro (define-extension name . clauses)
+  (let loop ((s '()) (d '()) (cs clauses) (exports #f))
+    (cond ((null? cs)
+	   (let ((exps (if exports `(declare (export ,@exports)) '(begin))))
+	     `(cond-expand
+	       (chicken-compile-shared ,exps ,@d)
+	       ((not compiling) ,@d)
+	       (else 
+		(declare (unit ,name))
+		,exps
+		(provide ',name) 
+		,@s) ) ) )
+	  ((and (pair? cs) (pair? (car cs)))
+	   (let ((t (caar cs))
+		 (next (cdr cs)) )
+	     (cond ((eq? 'static t) (loop (cons `(begin ,@(cdar cs)) s) d next exports))
+		   ((eq? 'dynamic t) (loop s (cons `(begin ,@(cdar cs)) d) next exports))
+		   ((eq? 'export t) (loop s d next (append (or exports '()) (cdar cs))))
+		   (else (syntax-error 'define-extension "invalid clause specifier" (caar cs))) ) ) )
+	  (else (syntax-error 'define-extension "invalid clause syntax" cs)) ) ) )
+
+
+;;;; SRFI-31
+
+(define-macro (rec head . args)
+  (if (pair? head)
+      `(letrec ((,(car head) (lambda ,(cdr head) ,@args))) ,(car head))
+      `(letrec ((,head ,@args)) ,head)))
+
 (eval-when (compile load eval)
-  (register-feature! 'srfi-26) )
+  (register-feature! 'srfi-8 'srfi-16 'srfi-26 'srfi-31 'srfi-15 'srfi-11) )
 
 
-;;; User defined optimizers:
+;;; Definitions available at macroexpansion-time:
+
+(define-macro (define-for-syntax head . body)
+  (let* ((body (if (null? body) '((void)) body))
+	 (name (if (pair? head) (car head) head)) 
+	 (body (if (pair? head) `(lambda ,(cdr head) ,@body) (car body))))
+    (if (symbol? name)
+	(##sys#setslot name 0 (eval body))
+	(syntax-error 'define-for-syntax "invalid identifier" name) )
+    (if ##sys#enable-runtime-macros
+	`(define ,name ,body)
+	'(begin) ) ) )
 
-(define-macro (define-optimizer head . body)
-  (##sys#check-syntax 'define-optimizer head '(symbol . lambda-list))
-  `(eval-when (compile)
-     (##compiler#register-optimizer ',(car head) (lambda ,(cdr head) ,@body)) ) )
