@@ -1,6 +1,6 @@
 ;;;; csi.scm - Simple interpreter stub for CHICKEN
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -33,19 +33,13 @@
 ; Germany
 
 
-(declare (uses syntax-case match-support))
+(declare (uses match-support))
 
 (declare
   (usual-integrations)
   (disable-interrupts)
   (run-time-macros)
   (foreign-declare #<<EOF
-#ifdef C_USE_C_DEFAULTS
-# include "c_defaults.h"
-#elif !defined(C_INSTALL_HOME)
-# define C_INSTALL_HOME NULL
-#endif
-
 #if defined(_MSC_VER) && defined(_WIN32)
 # include <direct.h>
 #else
@@ -63,99 +57,80 @@ EOF
 (include "banner")
 
 #{csi 
-  max-history-count print-usage print-banner write-to-echo-port
-  run hexdump del transcript-port
-  trace-hl-macros parse-option-string chop-separator lookup-script-file
+  print-usage print-banner
+  run hexdump del 
+  parse-option-string chop-separator lookup-script-file
   report describe dump hexdump bytevector-data get-config
-  advised-procedures do-advise do-unadvise deldups
+  deldups tty-input?
+  history-list history-count history-add history-ref
   trace-indent trace-indent-level traced-procedures traced-procedure-entry traced-procedure-exit}
 
 (declare
-  (hide parse-option-string trace-hl-macros bytevector-data transcript-port banner
-	write-to-echo-port trace-indent) )
+  (hide parse-option-string bytevector-data banner member* canonicalize-args do-trace do-untrace
+	findall trace-indent command-table) )
 
 
 ;;; Parameters:
 
 (define-constant init-file ".csirc")
-(define-constant max-history-count 16)
-
-(set! ##sys#repl-print-length-limit 512)
-
-(define-foreign-variable installation-home c-string "C_INSTALL_HOME")
 
+(set! ##sys#repl-print-length-limit 2048)
 (set! ##sys#features (cons #:csi ##sys#features))
 
 
 ;;; Print all sorts of information:
 
 (define (print-usage)
-  (display #<#EOF
-Usage: csi {FILENAME | OPTION}
+  (display
+"Usage: csi {FILENAME | OPTION}
 
   where OPTION may be one of the following:
 
-    -help                       display this text and exit
-    -version                    display version and exit
-    -case-insensitive           enable case-insensitive reading
-    -eval EXPRESSIONS           evaluate given expressions
-    -feature SYMBOL             register feature identifier
-    -quiet                      do not print banner
-    -no-init                    do not load initialization file `~/#{init-file}'
-    -no-feature SYMBOL          unregister feature identifier
-    -batch                      terminate after command-line processing
-    -syntax  -hygienic          install SYNTAX-CASE macro package
-    -no-warnings                disable all warnings
-    -keyword-style STYLE        enable alternative keyword-syntax (none, prefix or suffix)
-    -r5rs                       equivalent to `-hygienic -strict'
-    -script PATHNAME            use interpreter for shell scripts
-    -script-meta PATHNAME       `-script' with extra options
-    -strict                     disable non-standard macros and symbol syntax
-    -strict-srfi-0              disable non-standard macros except `cond-expand'
-    -strict-reader              disable non-standard read syntax
-    -strict-letrec              enable fully R5RS conformant `letrec'
+    -h  -help  --help           display this text and exit
+    -v  -version                display version and exit
+    -i  -case-insensitive       enable case-insensitive reading
+    -e  -eval EXPRESSION        evaluate given expression
+    -D  -feature SYMBOL         register feature identifier
+    -q  -quiet                  do not print banner
+    -n  -no-init                do not load initialization file `")
+  (display init-file)
+  (display 
+"'
+    -b  -batch                  terminate after command-line processing
+    -w  -no-warnings            disable all warnings
+    -k  -keyword-style STYLE    enable alternative keyword-syntax (none, prefix or suffix)
+    -s  -script PATHNAME        use interpreter for shell scripts
+    -R  -require-extension NAME require extension before executing code
+    -I  -include-path PATHNAME  add PATHNAME to include path
     --                          ignore all following options
 
-EOF
-) )
+") )
 
-(define print-banner
-  (let ([config 
-	 (let ([bp (build-platform)]
-	       [st (software-type)]
-	       [sv (software-version)]
-	       [mt (machine-type)] )
-	   (define (str x)
-	     (if (eq? x 'unknown)
-		 ""
-		 (string-append (##sys#slot x 1) "-") ) )
-	   (string-append (str sv) (str st) (str bp) (##sys#slot mt 1)) ) ] )
-    (lambda ()
-      (print banner "\nVersion " build-version ", Build " build-number " - " config "\n(c)2000-2004 Felix L. Winkelmann") ) ) )
+(define (print-banner)
+  (print banner (chicken-version #t) "\n(c)2000-2005 Felix L. Winkelmann") )
 
 
-;;; Reader for she-bang comment:
-
-(set! ##sys#she-bang-checked #f)
+;;; Reader for REPL history:
 
 (set! ##sys#user-read-hook
   (let ([read-char read-char]
 	[read read]
 	[old-hook ##sys#user-read-hook] )
     (lambda (char port)
-      (if (char=? char #\!)
-	  (do ((c (read-char port) (read-char port)))
-	      ((char=? c #\newline) 
-	       (set! ##sys#she-bang-checked #t)
-	       (read port) )
-	    (when (eof-object? c) (##sys#error "unexpected end of file")) )
-	  (old-hook char port) ) ) ) )
+      (cond [(or (char=? #\) char) (char-whitespace? char))
+	     `',(history-ref (fx- history-count 1)) ]
+	    [(char-numeric? char)
+	     (let ([n (read port)])
+	       (if (integer? n)
+		   `',(history-ref n)
+		   (##sys#error "invalid syntax in `#<number>' (REPL history)" n) ) ) ]
+	    [else (old-hook char port)] ) ) ) )
 
 
 ;;; Chop terminating separator from pathname:
 
 (define chop-separator 
-  (let ([pds pathname-directory-separator]
+  (let ([pds ##sys#pathname-directory-separator]
 	[substring substring] )
     (lambda (str)
       (let ([len (sub1 (##sys#size str))])
@@ -170,7 +145,7 @@ EOF
 
 (define lookup-script-file 
   (let* ([buf (make-string 256)]
-	 [_getcwd (foreign-lambda nonnull-c-string "_getcwd" pointer int)] )
+	 [_getcwd (foreign-lambda nonnull-c-string "_getcwd" scheme-pointer int)] )
     (define (addext name)
       (if (file-exists? name)
 	  name
@@ -185,13 +160,13 @@ EOF
     (lambda (name)
       (let ([path (getenv "PATH")])
 	(and (> (##sys#size name) 0)
-	     (cond [(char=? pathname-directory-separator (string-ref name 0)) (addext name)]
-		   [(string-index (lambda (c) (char=? c pathname-directory-separator)) name)
+	     (cond [(char=? ##sys#pathname-directory-separator (string-ref name 0)) (addext name)]
+		   [(string-index (lambda (c) (char=? c ##sys#pathname-directory-separator)) name)
 		    (and-let* ([p (_getcwd buf 256)])
-		      (addext (string-append (chop-separator p) (string pathname-directory-separator) name)) ) ]
+		      (addext (string-append (chop-separator p) (string ##sys#pathname-directory-separator) name)) ) ]
 		   [(addext name)]
 		   [else
-		    (let ([name2 (string-append (string pathname-directory-separator) name)])
+		    (let ([name2 (string-append (string ##sys#pathname-directory-separator) name)])
 		      (let loop ([ps (string-split path ";")])
 			(and (pair? ps)
 			     (let ([name2 (string-append (chop-separator (##sys#slot ps 0)) name2)])
@@ -199,199 +174,161 @@ EOF
 				   (loop (##sys#slot ps 1)) ) ) ) ) ) ] ) ) ) ) ) )
 				   
 
-;;; Add some 'comma'-commands:
+;;; REPL customization:
+
+(define history-list (make-vector 32))
+(define history-count 1)
+
+(define history-add
+  (let ([vector-resize vector-resize])
+    (lambda (vals)
+      (let ([x (if (null? vals) (##sys#void) (##sys#slot vals 0))]
+	    [size (##sys#size history-list)] )
+	(when (fx>= history-count size)
+	  (set! history-list (vector-resize history-list (fx* 2 size))) )
+	(vector-set! history-list history-count x)
+	(set! history-count (fx+ history-count 1))
+	x) ) ) )
+
+(define (history-ref index)
+  (let ([i (inexact->exact index)])
+    (if (and (fx> i 0) (fx<= i history-count))
+	(vector-ref history-list i) 
+	(##sys#error "history entry index out of range" index) ) ) )
+
+(repl-prompt
+ (let ([sprintf sprintf])
+   (lambda ()
+     (sprintf "#;~A> " history-count) ) ) )
+
+(define (tty-input?)
+  (or (##sys#fudge 12) (##sys#tty-port? ##sys#standard-input)) )
+
+(set! ##sys#break-on-error #f)
+
+(set! ##sys#read-prompt-hook
+  (let ([old ##sys#read-prompt-hook])
+    (lambda ()
+      (when (tty-input?) (old)) ) ) )
+
+(define command-table (make-hash-table eq?))
+
+(define toplevel-command
+  (let ((hash-table-set! hash-table-set!))
+    (lambda (name proc #!optional help)
+      (##sys#check-symbol name 'toplevel-command)
+      (when help (##sys#check-string help 'toplevel-command))
+      (hash-table-set! command-table name (cons proc help)) ) ) )
 
 (set! ##sys#repl-eval-hook
+  ;; `macroexpand' is intentionally not shadowed (psyntax redefines it).
   (let ((eval eval)
 	(load-noisily load-noisily)
 	(read read)
+	(hash-table-ref hash-table-ref)
+	(hash-table-walk hash-table-walk)
+	(read-line read-line)
 	(length length)
 	(display display)
 	(write write)
+	(string-split string-split)
 	(printf printf)
+	(pretty-print pretty-print)
 	(integer? integer?)
-	(butlast butlast)
-	(list-ref list-ref)
-	(values values)
-	(history-list '())
-	(history-count 0) 
-	(void void) )
+	(values values) )
     (lambda (form)
-
-      (define (filename x)
-	(cond ((string? x) x)
-	      ((symbol? x) (##sys#symbol->string x)) 
-	      (else (##sys#error "argument is not a valid filename" x)) ) )
-
-      (define (add-h form results)
-	(cond ((fx>= history-count max-history-count)
-	       (let ([tl (list-tail history-list (fx- history-count 1))])
-		 (set-cdr! tl '())
-		 (set! history-list (cons (cons form results) history-list)) ) )
-	      (else
-	       (set! history-list (cons (cons form results) history-list))
-	       (set! history-count (fx+ history-count 1)) ) ) )
-
-      (define (ref-h index)
-	(let ((i (fx- (inexact->exact index) 1)))
-	  (if (and (fx>= i 0) (fx< i (length history-list)))
-	      (list-ref history-list i) 
-	      (##sys#error "history entry index out of range" index) ) ) )
-	  
       (set! trace-indent-level 0)
-      (set! $ (lambda i
-		(let ((index (if (pair? i) (inexact->exact (##sys#slot i 0)) 1)))
-		  (if (integer? index)
-		      (car (ref-h index))
-		      (##sys#error "invalid history entry index" index) ) ) ) )
-      (set! & (lambda i
-		(let ((index (if (pair? i) (inexact->exact (##sys#slot i 0)) 1)))
-		  (if (integer? index)
-		      (apply values (cdr (ref-h index)))
-		      (##sys#error "invalid history entry index" index) ) ) ) )
-      (write-to-echo-port form)
       (cond ((eof-object? form) (exit))
 	    ((and (pair? form)
 		  (eq? 'unquote (##sys#slot form 0)) )
 	     (let ((cmd (cadr form)))
-	       (case cmd
-		 ((x)
-		  (let ([x (read)])
-		    (add-h `(pretty-print (macroexpand ',x)) (list (void)))
-		    (pretty-print (macroexpand x))
-		    (void) ) )
-		 ((p)
-		  (let* ([x (read)]
-			 [xe (eval x)] )
-		    (add-h `(pretty-print (eval ',x)) (list (void)))
-		    (pretty-print xe)
-		    (void) ) )
-		 ((d)
-		  (let* ([x (read)]
-			 [xe (eval x)] )
-		    (describe xe) ) )
-		 ((du)
-		  (let* ([x (read)]
-			 [xe (eval x)] )
-		    (dump xe) ) )
-		 ((dur)
-		  (let* ([x (read)]
-			 [n (read)]
-			 [xe (eval x)] 
-			 [xn (eval n)] )
-		    (dump xe xn) ) )
-		 ((r) (report))
-		 ((q) (exit))
-		 ((l)
-		  (let ((fn (filename (read))))
-		    (add-h `(load ,fn) (list (void)))
-		    (load fn)
-		    (void) ) )
-		 ((ln) 
-		  (let ((fn (filename (read))))
-		    (add-h `(load-noisily ,fn) (list (void)))
-		    (load-noisily fn)
-		    (void) ) )
-		 ((t)
-		  (let ((x (read)))
-		    (receive rs (time (eval x))
-		      (add-h `(time (eval ',x)) rs)
-		      (apply values rs) ) ) )
-		 ((s)
-		  (let* ((x (read))
-			 (str (if (symbol? x) (##sys#symbol->string x) x))
-			 (r (system str)) )
-		    (add-h `(system ,str) (list r))
-		    r) )
-		 ((h)
-		  (do ((i 1 (fx+ i 1))
-		       (hl history-list (##sys#slot hl 1)) )
-		      ((eq? hl '()))
-		    (let ((entry (##sys#slot hl 0)))
-		      (printf "~s~a: ~s" i (if (fx< i 10) #\space "") (car entry))
-		      (if (not (equal? (list (void)) (cdr entry)))
-			  (begin
-			    (display " ->")
-			    (for-each 
-			     (lambda (x)
-			       (##sys#with-print-length-limit
-				60
-				(lambda ()
-				  (display #\space)
-				  (write x) ) ) )
-			     (cdr entry) ) ) )
-		      (newline) ) )
-		  (void) )
-		 ((?)
-		  (display 
-		   "Toplevel commands:
-
- ,?            Show this text
- ,h            List history of last commands
- ,NUM          Evaluate NUMth command in history-list
- ,p EXP        Pretty print evaluated expression EXP
- ,d EXP        Describe result of evaluated expression EXP
- ,du EXP       Dump data of expression EXP
- ,dur EXP N    Dump range
- ,q            Quit interpreter
- ,l FILENAME   Load file with given filename (may be symbol or string)
- ,ln FILENAME  Load file and print result of each top-level expression
- ,r            Show system information
- ,s STRING     Execute shell-command
- ,t EXP        Evaluate form and print elapsed time
- ,x EXP        Pretty print macroexpanded expression EXP\n")
-		  (void) )
-		 (else
-		  (cond ((fixnum? cmd)
-			 (let* ((item (ref-h cmd))
-				(exp (car item)) )
-			   (receive rs (eval exp)
-			     (add-h exp rs)
+	       (cond ((and (symbol? cmd) (hash-table-ref/default command-table cmd #f)) =>
+		      (lambda (p)
+			((car p))
+			(##sys#void) ) )
+		     (else
+		      (case cmd
+			((x)
+			 (let ([x (read)])
+			   (pretty-print (macroexpand x))
+			   (##sys#void) ) )
+			((p)
+			 (let* ([x (read)]
+				[xe (eval x)] )
+			   (pretty-print xe)
+			   (##sys#void) ) )
+			((d)
+			 (let* ([x (read)]
+				[xe (eval x)] )
+			   (describe xe) ) )
+			((du)
+			 (let* ([x (read)]
+				[xe (eval x)] )
+			   (dump xe) ) )
+			((dur)
+			 (let* ([x (read)]
+				[n (read)]
+				[xe (eval x)] 
+				[xn (eval n)] )
+			   (dump xe xn) ) )
+			((r) (report))
+			((q) (exit))
+			((l)
+			 (let ((fns (string-split (read-line))))
+			   (for-each load fns)
+			   (##sys#void) ) )
+			((ln) 
+			 (let ((fns (string-split (read-line))))
+			   (for-each (cut load-noisily <> printer: (lambda (x) (pretty-print x) (print* "==> "))) fns)
+			   (##sys#void) ) )
+			((t)
+			 (let ((x (read)))
+			   (receive rs (time (eval x))
+			     (history-add rs)
 			     (apply values rs) ) ) )
+			((tr) (do-trace (string-split (read-line))))
+			((utr) (do-untrace (string-split (read-line))))
+			((s)
+			 (let* ((str (read-line))
+				(r (system str)) )
+			   (history-add (list r))
+			   r) )
+			((?)
+			 (display 
+			  "Toplevel commands:
+
+ ,?                Show this text
+ ,p EXP            Pretty print evaluated expression EXP
+ ,d EXP            Describe result of evaluated expression EXP
+ ,du EXP           Dump data of expression EXP
+ ,dur EXP N        Dump range
+ ,q                Quit interpreter
+ ,l FILENAME ...   Load file with given filename
+ ,ln FILENAME ...  Load file and print result of each top-level expression
+ ,r                Show system information
+ ,s TEXT ...       Execute shell-command
+ ,tr NAME ...      Trace procedures
+ ,utr NAME ...     Untrace procedures
+ ,t EXP            Evaluate form and print elapsed time
+ ,x EXP            Pretty print macroexpanded expression EXP\n")
+			 (hash-table-walk
+			  command-table
+			  (lambda (k v) 
+			    (let ((help (cdr v)))
+			      (if help
+				  (print #\space help)
+				  (print " ," k) ) ) ) )
+			 (##sys#void) )
 			(else
 			 (printf "Undefined toplevel command ~s - enter `,?' for help~%" form) 
-			 (void) ) ) ) ) ) )
+			 (##sys#void) ) ) ) ) ) )
 	    (else
 	     (receive rs (eval form)
-	       (add-h form rs) 
+	       (when (tty-input?) (history-add rs))
 	       (apply values rs) ) ) ) ) ) )
 
-(define transcript-port #f)
 
-(define write-to-echo-port
-  (let ([write write]
-	[newline newline] )
-    (lambda (x)
-      (when transcript-port 
-	(write x transcript-port)
-	(newline transcript-port) ) ) ) )
-
-(define transcript-on 
-  (let ([close-output-port close-output-port]
-	[open-output-file open-output-file] )
-    (lambda (fname)
-      (##sys#check-string fname 'transcript-on)
-      (when transcript-port (close-output-port transcript-port))
-      (set! transcript-port (open-output-file fname)) ) ) )
-
-(define transcript-off
-  (let ([close-output-port close-output-port])
-    (lambda ()
-      (when transcript-port (close-output-port transcript-port))
-      (set! transcript-port #f) ) ) )
-
-
-;;; Display prompt if stdin is a terminal:
-
-(set! ##sys#read-prompt-hook
-  (let ([display display])
-    (lambda ()
-      (when (or (##sys#fudge 12) (##sys#tty-port? (current-input-port)))
-	(display (repl-prompt))
-	(flush-output) ) ) ) )
-
-
-;;; Tracing/advise:
+;;; Tracing:
 
 (define (del x lst tst)
   (let loop ([lst lst])
@@ -404,7 +341,6 @@ EOF
 
 (define trace-indent-level 0)
 (define traced-procedures '())
-(define advised-procedures '())
 
 (define trace-indent
   (lambda ()
@@ -418,7 +354,7 @@ EOF
     (trace-indent)
     (set! trace-indent-level (add1 trace-indent-level))
     (write (cons name args))
-    (newline) 
+    (##sys#write-char-0 #\newline ##sys#standard-output) 
     (flush-output) ) )
 
 (define traced-procedure-exit
@@ -432,145 +368,41 @@ EOF
        (write x)
        (write-char #\space) )
      results)
-    (newline) 
+    (##sys#write-char-0 #\newline ##sys#standard-output) 
     (flush-output) ) )
 
-(define (do-advise name old)
-  (unless (assq name advised-procedures)
-    (when (not (procedure? old))
-      (##sys#signal-hook #:type-error 'advise "can not advise non-procedure" old) )
-    (set! advised-procedures (cons (cons name old) advised-procedures)) ) )
-
-(define (do-unadvise name prev)
-  (let ([a (assq name advised-procedures)])
-    (if a
-	(begin
-	  (set! advised-procedures (del a advised-procedures eq?))
-	  (##sys#slot a 1) )
-	(begin
-	  (printf "Procedure `~S' was never advised.~%" name)
-	  prev) ) ) )
-
-(define-macro (trace . names)
-  `(begin
-     ,@(map (lambda (s)
-	      (let ((name (gensym))
-		    (old (gensym)) )
-		`(let* ((,name ',s)
-			(,old ,s) )
-		   (if (not (procedure? ,old)) (##sys#error 'trace "can not trace non-procedure" ,old))
-		   (set! traced-procedures (cons (cons ,name ,old) traced-procedures))
-		   (set! ,s (lambda args
-			      (traced-procedure-entry ,name args)
-			      (call-with-values (lambda () (apply ,old args))
-				(lambda results
-				  (traced-procedure-exit ,name results) 
-				  (apply values results) ) ) ) ) ) ) )
-	    names) ) )
-
-(define-macro (untrace . names)
-  `(begin
-     ,@(map (lambda (s)
-	      (let ((var (gensym))
-		    (old (gensym)) )
-		`(let* ((,var ',s)
-			(,old (assq ,var traced-procedures)) )
-		   (if ,old
-		       (set! ,s (cdr ,old))
-		       (printf "Procedure `~s' was not traced.~%" ,var) )
-		   (set! traced-procedures (del ,old traced-procedures eq?) ) ) ) )
-	    names) ) )
-
-(define-macro (advise name mode proc)
-  (let ([tmp (gensym)]
-	[old (gensym)] )
-    `(let* ([,tmp ,proc]
-	    [,old ,name] )
-       (do-advise ',name ,old)
-       (set! ,name
-	 ,(case mode
-	    [(before)
-	     `(lambda args
-		(apply ,tmp args)
-		(apply ,old args) ) ]
-	    [(after)
-	     `(lambda args
-		(call-with-values
-		    (lambda () (apply ,old args))
-		  (lambda results
-		    (apply ,tmp results)
-		    (apply values results) ) ) ) ]
-	    [(around)
-	     `(lambda args
-		(apply ,tmp ,old args) ) ]
-	    [else (##sys#error 'advise "invalid mode" ',mode)] ) ) ) ) )
-
-(define-macro (unadvise . names)
-  `(begin
-     ,@(map (lambda (n) 
-	      `(set! ,n (do-unadvise ',n ,n) ) )
-	    names) ) )
-
-(define trace-hl-macros #<<EOF
-(begin
-
-(define-syntax trace
-  (lambda (x)
-    (syntax-case x ()
-      [(_) (syntax (##core#undefined))]
-      [(_ name ...)
-       (with-syntax ([(tmp ...) (generate-temporaries (syntax (name ...)))])
-	 (syntax
-	  (let ([tmp name] ...)
-	    (if (not (procedure? tmp)) (##sys#error 'trace "can not trace non-procedure" tmp)) ...
-	    (set! ##csi#traced-procedures (cons (cons 'name tmp) ##csi#traced-procedures)) ...
-	    (set! name (lambda args
-			 (##csi#traced-procedure-entry 'name args)
-			 (##sys#call-with-values (lambda () (##sys#apply tmp args))
+(define do-trace
+  (lambda (names)
+    (for-each
+     (lambda (s)
+       (let ((s (string->symbol s)))
+	 (if (assq s traced-procedures)
+	     (##sys#warn "procedure already traced" s)
+	     (let ((old (##sys#slot s 0)))
+	       (cond ((not (procedure? old)) (##sys#error "can not trace non-procedure" s))
+		     (else
+		      (set! traced-procedures (cons (cons s old) traced-procedures))
+		      (##sys#setslot
+		       s 0
+		       (lambda args
+			 (traced-procedure-entry s args)
+			 (call-with-values (lambda () (apply old args))
 			   (lambda results
-			     (##csi#traced-procedure-exit 'name results)
-			     (##sys#apply ##sys#values results) ) ) ) )
-	    ...) ) ) ] ) ) )
+			     (traced-procedure-exit s results) 
+			     (apply values results) ) ) ) ) ) ) ) ) ) )
+     names) ) )
 
-(define-syntax untrace
-  (lambda (x)
-    (syntax-case x ()
-      [(_) (syntax (##core#undefined))]
-      [(_ name ...)
-       (with-syntax ([(tmp ...) (generate-temporaries (syntax (name ...)))])
-	 (syntax
-	  (let ([tmp (assq 'name ##csi#traced-procedures)] ...)
-	    (if tmp (set! name (cdr tmp))) ...
-	    (set! ##csi#traced-procedures (delq tmp ##csi#traced-procedures eq?)) ...) ) ) ] ) ) )
-
-(define-syntax advise
-  (syntax-rules (before after around)
-    [(_ "1" name before tmp old)
-     (set! name (lambda args (apply tmp args) (apply old args))) ]
-    [(_ "1" name after tmp old)
-     (set! name
-       (lambda args
-	 (call-with-values (lambda () (apply old args))
-	   (lambda results 
-	     (apply tmp results)
-	     (apply values results) ) ) ) ) ]
-    [(_ "1" name around tmp old)
-     (set! name (lambda args (apply tmp old args))) ]
-    [(_ name mode proc)
-     (let ([tmp proc]
-	   [old name] )
-       (##csi#do-advise 'name old)
-       (advise "1" name mode tmp old) ) ] ) )
-
-(define-syntax unadvise
-  (syntax-rules ()
-    [(_ name ...)
-     (begin
-       (set! name (##csi#do-unadvise 'name name)) ...) ] ) )
-
-)
-EOF
-)
+(define do-untrace 
+  (lambda (names)
+    (for-each
+     (lambda (s)
+       (let* ((s (string->symbol s))
+	      (p (assq s traced-procedures)) )
+	 (cond ((not p) (##sys#warn "procedure not traced" s))
+	       (else
+		(##sys#setslot s 0 (cdr p))
+		(set! traced-procedures (del s traced-procedures eq?) ) ) ) ) )
+     (if (null? names) (map (lambda (x) (symbol->string (car x))) traced-procedures) names) ) ) )
 
 
 ;;; Parse options from string:
@@ -631,7 +463,7 @@ EOF
 		    (vector-ref minfo 1)
 		    (vector-ref minfo 2)
 		    (if (= 1 (##sys#fudge 18)) "downward" "upward") )
-	    (newline)
+	    (##sys#write-char-0 #\newline ##sys#standard-output)
 	    (when (##sys#fudge 14) (display "interrupts are enabled\n"))
 	    (when (##sys#fudge 15) (display "symbol gc is enabled\n")) 
 	    (##core#undefined) ) ) ) ) ) )
@@ -658,9 +490,7 @@ EOF
 	[length length]
 	[list-ref list-ref]
 	[string-ref string-ref]
-	[newline newline]
-	[hash-table-for-each hash-table-for-each] 
-	[void void] )
+	[hash-table-walk hash-table-walk] )
     (lambda (x . port)
       (let ([out (:optional port ##sys#standard-output)])
 
@@ -684,15 +514,16 @@ EOF
 	      [(eq? x #t) (fprintf out "boolean true~%")]
 	      [(eq? x #f) (fprintf out "boolean false~%")]
 	      [(null? x) (fprintf out "empty list~%")]
-	      [(eof-object? x) (fprintf out "end-of-file~%")]
-	      [(eq? (void) x) (fprintf out "unspecified object~%")]
+	      [(eof-object? x) (fprintf out "end-of-file object~%")]
+	      [(eq? (##sys#void) x) (fprintf out "unspecified object~%")]
 	      [(fixnum? x)
 	       (fprintf out "exact integer ~S, #x~X, #o~O, #b~B" x x x x)
-	       (when (fx< x 256) (fprintf out ", character ~S" (integer->char x)))
-	       (newline) ]
+	       (let ([code (integer->char x)])
+		 (when (fx< code #x10000) (fprintf out ", character ~S" code)) )
+	       (##sys#write-char-0 #\newline ##sys#standard-output) ]
 	      [(eq? x (##sys#slot '##sys#arbitrary-unbound-symbol 0))
 	       (fprintf out "unbound value~%") ]
-	      [(number? x) (fprintf out "inexact number ~S~%" x)]
+	      [(##sys#number? x) (fprintf out "number ~S~%" x)]
 	      [(string? x) (descseq "string" ##sys#size string-ref 0)]
 	      [(vector? x) (descseq "vector" ##sys#size ##sys#slot 0)]
 	      [(symbol? x)
@@ -705,6 +536,7 @@ EOF
 	      [(procedure? x)
 	       (let ([len (##sys#size x)])
 		 (if (and (> len 3)
+			  (memq #:tinyclos ##sys#features)
 			  (eq? ##tinyclos#entity-tag (##sys#slot x (fx- len 1))) )
 		     (describe-object x out)
 		     (descseq 
@@ -739,41 +571,46 @@ EOF
 	       (let ([len (##sys#size x)])
 		 (fprintf out "byte vector of size ~S:~%" len)
 		 (hexdump x len ##sys#byte out) ) ]
+	      [(##core#inline "C_lambdainfop" x)
+	       (fprintf out "lambda information: ~s~%" (##sys#lambda-info->string x)) ]
 	      [(##sys#structure? x 'hash-table)
-	       (let ([test
-		      (switch (##sys#slot x 3)
-			[eq? 'eq?]
-			[eqv? eqv?]
-			[equal? 'equal?] 
-			[else "<unknown>"] ) ] )
-		 (fprintf out "hash-table with ~S elements and comparison procedure ~A:~%" (##sys#slot x 2) test)
-		 (hash-table-for-each
-		  (lambda (k v) (fprintf out " ~S\t-> ~S~%" k v))
-		  x) ) ]
+	       (let ((n (##sys#slot x 2)))
+		 (fprintf out "hash-table with ~S element~a~%  comparison procedure: ~A~%"
+			  n (if (fx= n 1) "" "s")  (##sys#slot x 3)) )
+	       (fprintf out "  hash function: ~a~%" (##sys#slot x 4))
+	       (hash-table-walk
+		x
+		(lambda (k v) (fprintf out " ~S\t-> ~S~%" k v)) ) ]
 	      [(##sys#structure? x 'environment)
 	       (if (##sys#slot x 1)
 		   (fprintf out "environment~%") 
 		   (fprintf out "the interaction environment~%") ) ]
 	      [(##sys#structure? x 'array)
-	       (let ([shp (##sys#slot x 3)]
-		     [vec (##sys#slot x 1)] )
-		 (fprintf out "array of rank ~S with ~A elements~% shape: " (fx/ (##sys#size shp) 2) (##sys#size vec))
+	       (let* ([shp (##sys#slot x 3)]
+		      [vec (##sys#slot x 1)] 
+		      (n (##sys#siize vec)) )
+		 (fprintf out "array of rank ~S with ~A element~a~% shape: "
+			  (fx/ (##sys#size shp) 2)
+			  n
+			  (if (fx= n 1) "" "s") )
 		 (let loop ([s (vector->list shp)])
 		   (unless (null? s)
 		     (let ([next (##sys#slot s 1)])
 		       (fprintf out " ~A..~A " (##sys#slot s 0) (##sys#slot next 0))
 		       (loop (##sys#slot next 1)) ) ) )
-		 (newline out) ) ]
+		 (##sys#write-char-0 #\newline out) ) ]
+	      [(and (##sys#structure? x 'meroon-instance) (provided? 'meroon))
+	       (unveil x out) ]
 	      [(##sys#generic-structure? x)
 	       (let ([st (##sys#slot x 0)])
 		 (let ([data (assq st bytevector-data)] )
 		   (if data
 		       (apply descseq (append (map eval (cdr data)) (list 0)))
 		       (begin
-			 (fprintf out "structure of type ~S:~%" (##sys#slot x 0))
+			 (fprintf out "structure of type `~S':~%" (##sys#slot x 0))
 			 (descseq #f ##sys#size ##sys#slot 1) ) ) ) ) ]
 	      [else (fprintf out "unknown object~%")] )
-	(void) ) ) ) )
+	(##sys#void) ) ) ) )
 
 
 ;;; Display hexdump:
@@ -831,7 +668,7 @@ EOF
 	    (if (and (fx>= c 32) (fx< c 128))
 		(write-char (integer->char c) out)
 		(write-char #\. out) ) ) ) 
-	(write-char #\newline out) ) ) ) )
+	(##sys#write-char-0 #\newline out) ) ) ) )
 
 
 ;;; emacs interface (suggested by Linh Dang)
@@ -892,39 +729,66 @@ EOF
 		 (new-tail (recur (del x tail elt=))))
 	    (if (eq? tail new-tail) lis (cons x new-tail)))))))
 
+(define (member* keys set)
+  (let loop ((set set))
+    (and (pair? set)
+	 (let find ((ks keys))
+	   (cond ((null? ks) (loop (cdr set)))
+		 ((equal? (car ks) (car set)) set)
+		 (else (find (cdr ks))) ) ) ) ) )
+
+(define-constant short-options 
+  '(#\k #\s #\v #\h #\D #\e #\i #\R #\b #\n #\q #\w #\- #\I) )
+
+(define-constant long-options
+  '("-keyword-style" "-script" "-version" "-help" "--help" "--" "-feature" "-eval" "-case-insensitive"
+    "-require-extension" "-batch" "-quiet" "-no-warnings" "-no-init" "-include-path") )
+
+(define (canonicalize-args args)
+  (let loop ((args args))
+    (if (null? args)
+	'()
+	(let ((x (car args)))
+	  (cond ((or (string=? "-s" x) (string=? "-script" x)) args)
+		((and (fx> (##sys#size x) 2)
+		      (char=? #\- (##core#inline "C_subchar" x 0))
+		      (not (member x long-options)) )
+		 (if (char=? #\: (##core#inline "C_subchar" x 1))
+		     (loop (cdr args))
+		     (let ((cs (string->list (substring x 1))))
+		       (if (findall cs short-options)
+			   (append (map (cut string #\- <>) cs) (loop (cdr args)))
+			   (##sys#error "invalid option" x) ) ) ) )
+		(else (cons x (loop (cdr args)))))))))
+
+(define (findall chars clist)
+  (let loop ((chars chars))
+    (or (null? chars)
+	(and (memq (car chars) clist)
+	     (loop (cdr chars))))))
+
 (define (run)
   (let* ([extraopts (parse-option-string (or (getenv "CSI_OPTIONS") ""))]
-	 [args (cdr (argv))]
+	 [args (canonicalize-args (cdr (argv)))]
 	 [loadlater #f]
-	 [kwstyle (member "-keyword-style" args)]
-	 [script (or (member "-script" args)
-		     (member "-script-meta" args) ) ] )
+	 [kwstyle (member* '("-k" "-keyword-style") args)]
+	 [script (member* '("-s" "-script") args)])
     (cond [script
 	   (unless (pair? (cdr script)) (##sys#error "missing argument to `-script' option"))
 	   (command-line-arguments (cddr script))
 	   (register-feature! 'script)
 	   (set-cdr! (cdr script) '()) 
-	   (when (memq (software-type) '(windows msdos))
+	   (when (and (memq (software-type) '(windows msdos))
+		      (not (eq? (build-platform) 'cygwin)) )
 	     (and-let* ([sname (lookup-script-file (cadr script))])
-	       (set-car! (cdr script) sname) ) )
-	   (when (string=? "-script-meta" (car script))
-	     (set! loadlater (open-input-file (cadr script)))
-	     (read-line loadlater)
-	     (set! args (append (parse-option-string (read-line loadlater)) args))
-	     (set-cdr! script '()) ) ]
+	       (set-car! (cdr script) sname) ) ) ]
 	  [else
-	   (set! args (append extraopts args))
+	   (set! args (append (canonicalize-args extraopts) args))
 	   (and-let* ([p (member "--" args)])
 	     (set-cdr! p '()) ) ] )
-    (let* ([batch (or (member "-batch" args) script)]
-	   [quiet (or (member "-quiet" args) script)]
-	   [r5rs (member "-r5rs" args)]
-	   [strict (or r5rs (member "-strict" args))]
-	   [strictr (member "-strict-reader" args)]
-	   [strict-srfi0 (member "-strict-srfi-0" args)]
-	   [strict-letrec (or strict strict-srfi0 (member "-strict-letrec" args))]
-	   [hygienic (or r5rs (member "-hygienic" args) (member "-syntax" args))]
-	   [home (chop-separator (or (getenv "CHICKEN_HOME") installation-home ""))]
+    (let* ([eval? (member* '("-e" "-eval") args)]
+	   [batch (or script (member* '("-b" "-batch") args) eval?)]
+	   [quiet (or script (member* '("-q" "-quiet") args) eval?)]
 	   [ipath (map chop-separator (string-split (or (getenv "CHICKEN_INCLUDE_PATH") "") ";"))] )
       
       (define (collect-options opt)
@@ -937,59 +801,41 @@ EOF
 		[else '()] ) ) )
 
       (define (loadinit)
-	(let* ([prefix (chop-separator (or (getenv "HOME") "."))]
-	       [fn (string-append prefix (string pathname-directory-separator) init-file)] )
-	  (when (file-exists? fn) 
-	    (load fn) ) ) )
+	(let ([fn (##sys#string-append "./" init-file)])
+	  (if (file-exists? fn)
+	      (load fn)
+	      (let* ([prefix (chop-separator (or (getenv "HOME") "."))]
+		     [fn (string-append prefix (string ##sys#pathname-directory-separator) init-file)] )
+		(when (file-exists? fn) 
+		  (load fn) ) ) ) ) )
 
-      (when (or (member "-help" args) (member "--help" args))
+      (when (member* '("-h" "-help" "--help") args)
 	(print-usage)
 	(exit 0) )
-      (when (member "-version" args)
+      (when (member* '("-v" "-version") args)
 	(print-banner)
 	(exit 0) )
-      (when (member "-no-warnings" args)
+      (when (member* '("-w" "-no-warnings") args)
 	(unless quiet (display "Warnings are disabled\n"))
 	(set! ##sys#warnings-enabled #f) )
       (unless quiet
 	(load-verbose #t)
 	(print-banner) )
-      (when (member "-case-insensitive" args)
+      (when (member* '("-i" "-case-insensitive") args)
 	(unless quiet (display "Identifiers and symbols are case insensitive\n"))
 	(register-feature! 'case-insensitive)
 	(case-sensitive #f) )
       (for-each register-feature! (collect-options "-feature"))
-      (for-each unregister-feature! (collect-options "-no-feature"))
+      (for-each register-feature! (collect-options "-D"))
       (set! ##sys#include-pathnames 
 	(deldups
 	 (append (map chop-separator (collect-options "-include-path"))
+		 (map chop-separator (collect-options "-I"))
 		 ##sys#include-pathnames
-		 ipath
-		 (if home (list home) '()) )
+		 ipath)
 	 string=?) )
-      (unless hygienic
-	(set! ##sys#features (cons #:match (cons #:match-support ##sys#features)))
-	(provide 'match) )
-      (cond [hygienic
-	     (unless quiet
-	       (display "Using hygienic macros")
-	       (cond [strict (display " - strict mode\n")]
-		     [strict-srfi0 (display " - strict mode with SRFI-0\n")]
-		     [else (newline)] ) )
-	     (##syncase#install-macro-package (or strict strict-srfi0) (not strict) strict-letrec)
-	     (##syncase#install-macro-defs trace-hl-macros) ]
-	    [strict
-	     (unless quiet (display "Strict mode\n"))
-	     (set! ##sys#strict-mode #t)
-	     (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-0 #:srfi-16 #:srfi-9 #:srfi-26)
-	     (register-feature! #:strict)
-	     (##sys#undefine-non-standard-macros '()) ]
-	    [strict-srfi0
-	     (unless quiet (display "Strict mode with SRFI-0\n"))
-	     (set! ##sys#strict-mode #t)
-	     (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-16 #:srfi-9 #:srfi-26)
-	     (register-feature! #:strict)
-	     (##sys#undefine-non-standard-macros '(cond-expand)) ] )
+      (set! ##sys#features (cons #:match ##sys#features))
+      (provide 'match)
       (when kwstyle
 	(cond [(not (pair? (cdr kwstyle)))
 	       (##sys#error "missing argument to `-keyword-style' option") ]
@@ -999,33 +845,30 @@ EOF
 	       (keyword-style #:none) ]
 	      [(string=? "suffix" (cadr kwstyle))
 	       (keyword-style #:suffix) ] ) )
-      (when (or strict strictr strict-srfi0)
-	(strict-reader #t)
-	(keyword-style #:none)
-	(case-sensitive #f) )
-      (when (and strict-letrec (not hygienic))
-	(##sys#enable-strict-letrec) )
-      (unless (or (member "-no-init" args) script) (loadinit))
+      (unless (or (member* '("-n" "-no-init") args) script) (loadinit))
       (do ([args args (cdr args)])
 	  ((null? args)
 	   (when loadlater (load loadlater))
-	   (unless batch (repl)) )
+	   (unless batch 
+	     (repl)
+	     (##sys#write-char-0 #\newline ##sys#standard-output) ) )
 	(let* ([arg (car args)]
 	       [len (string-length arg)] )
 	  (cond ((member 
 		  arg 
-		  '("--" "-batch" "-quiet" "-no-init" "-hygienic" "-syntax" "-no-warnings" "-script"
-		    "-strict-reader" "-script-meta" "-strict-letrec" "-r5rs"
-		    "-case-insensitive" "-strict" "-strict-srfi-0") ) )
-		((member arg '("-feature" "-no-feature" "-include-path" "-keyword-style"))
+		  '("--" "-batch" "-quiet" "-no-init" "-no-warnings" "-script" "-b" "-q" "-n" "-w" "-s" "-i"
+		    "-case-insensitive") ) )
+		((member arg '("-feature" "-include-path" "-keyword-style" "-D" "-I" "-k"))
 		 (set! args (cdr args)) )
-		((string=? "-eval" arg)
+		((or (string=? "-R" arg) (string=? "-require-extension" arg))
+		 (eval `(##core#require-extension ',(string->symbol (cadr args))))
+		 (set! args (cdr args)) )
+		((or (string=? "-e" arg) (string=? "-eval" arg))
 		 (let ([in (open-input-string (cadr args))])
 		   (do ([x (read in) (read in)])
 		       ((eof-object? x))
 		     (eval x) )
 		   (set! args (cdr args)) ) )
-		((and (>= len 2) (string=? "-:" (substring arg 0 2))))
 		(else (load arg) ) ) ) ) ) ) )
 
 (run)
