@@ -1,6 +1,6 @@
 ;;; extras.scm - Optional non-standard extensions
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -50,9 +50,9 @@ static C_word fgets_into_buffer(C_word str, C_word port, C_word size)
 {
   int len, n = C_unfix(size);
   char *buf = C_c_string(str);
-  FILE *fp = C_port_file(port);
+  C_FILEPTR fp = C_port_file(port);
 
-  if(fgets(buf, n, fp) == NULL) return C_fix(0);
+  if(C_fgets(buf, n, fp) == NULL) return C_fix(0);
 
   len = C_strlen(buf);
 
@@ -83,7 +83,7 @@ EOF
   reverse-string-append generic-write hashtab-default-size hashtab-threshold hashtab-rehash hashtab-primes-table}
 
 (declare
-  (hide hashtab-threshold hashtab-rehash generic-write) )
+  (hide hashtab-threshold hashtab-rehash generic-write %hash) )
 
 (cond-expand
  [unsafe
@@ -111,15 +111,15 @@ EOF
   (let ([read read]
 	[reverse reverse] 
 	[call-with-input-file call-with-input-file] )
-    (lambda port
-      (let ([port (:optional port ##sys#standard-input)])
-	(define (slurp port)
-	  (do ([x (read port) (read port)]
-	       [xs '() (cons x xs)] )
-	      ((eof-object? x) (reverse xs)) ) )
-	(if (port? port)
-	    (slurp port)
-	    (call-with-input-file port slurp) ) ) ) ) )
+    (lambda (#!optional (port ##sys#standard-input) (reader read) max)
+      (define (slurp port)
+	(do ((x (read port) (reader port))
+	     (i 0 (fx+ i 1))
+	     (xs '() (cons x xs)) )
+	    ((or (eof-object? x) (and max (fx>= i max))) (reverse xs)) ) )
+      (if (port? port)
+	  (slurp port)
+	  (call-with-input-file port slurp) ) ) ) )
 
 
 ;;; Combinators:
@@ -173,9 +173,27 @@ EOF
 	    [(pred (##sys#slot lst 0)) (loop (##sys#slot lst 1))]
 	    [else #f] ) ) ) )
 
+(define (noop . _) (void))
+
+(define (each . procs)
+  (cond ((null? procs) (lambda _ (void)))
+	((null? (##sys#slot procs 1)) (##sys#slot procs 0))
+	(else
+	 (lambda args
+	   (let loop ((procs procs))
+	     (let ((h (##sys#slot procs 0))
+		   (t (##sys#slot procs 1)) )
+	       (if (null? t)
+		   (apply h args)
+		   (begin
+		     (apply h args)
+		     (loop t) ) ) ) ) ) ) ) )
+
 
 ;;; List operators:
 
+(define (atom? x) (##core#inline "C_i_atomp" x))
+
 (define (tail? x y)
   (##sys#check-list y 'tail?)
   (or (##core#inline "C_eqp" x '())
@@ -266,6 +284,9 @@ EOF
 	 (sort! (map (lambda (x) (cons (random 10000) x)) l)
 		(lambda (x y) (< (car x) (car y)))) ) ) )
 
+
+;;; Alists:
+
 (define (alist-update! x y lst . cmp)
   (let* ([cmp (if (pair? cmp) (car cmp) eqv?)]
 	 [aq (cond [(eq? eq? cmp) assq]
@@ -300,9 +321,20 @@ EOF
 				   (loop (##sys#slot lst 1)) ) ) ) ) ) ] ) ] 
 	 [item (aq x lst)] )
     (if item
-	(##sys#slot item 0)
+	(##sys#slot item 1)
 	default) ) )
 
+(define (rassoc x lst . tst)
+  (cond-expand [(not unsafe) (##sys#check-list lst 'rassoc)][else])
+  (let ([tst (if (pair? tst) (car tst) eqv?)])
+    (let loop ([l lst])
+      (and (pair? l)
+	   (let ([a (##sys#slot l 0)])
+	     (cond-expand [(not unsafe) (##sys#check-pair a 'rassoc)][else])
+	     (if (tst x (##sys#slot a 1))
+		 a
+		 (loop (##sys#slot l 1)) ) ) ) ) ) )
+
 
 ;;; Random numbers:
 
@@ -341,10 +373,11 @@ EOF
 	     [limit (and parg (pair? (cdr args)) (cadr args))]
 	     [buffer (make-string 256)]
 	     [len 256] )
+	(##sys#check-port p 'read-line)
 	(if (and FGETS_INTO_BUFFER (eq? 'stream (##sys#slot p 7)))
 	    (let loop ([len 256] [buffer buffer] [result ""] [f #f])
 	      (let ([n (##core#inline "fgets_into_buffer" buffer p len)])
-		(cond [(eq? 0 n) (if f result (##sys#fudge 1))]
+		(cond [(eq? 0 n) (if f result #!eof)]
 		      [(not n)
 		       (loop (fx* len 2) (make-string (fx* len 2))
 			     (##sys#string-append result (##sys#substring buffer 0 (fx- len 1)))
@@ -379,23 +412,32 @@ EOF
 
 (define read-lines
   (let ([read-line read-line]
+	[call-with-input-file call-with-input-file] 
 	[reverse reverse] )
     (lambda port-and-max
       (let* ([port (if (pair? port-and-max) (##sys#slot port-and-max 0) ##sys#standard-input)]
 	     [rest (and (pair? port-and-max) (##sys#slot port-and-max 1))]
 	     [max (if (pair? rest) (##sys#slot rest 0) #f)] )
-	(do ([ln (read-line port) (read-line port)]
-	     [lns '() (cons ln lns)]
-	     [n (or max 1000000) (fx- n 1)] )
-	    ((or (eof-object? ln) (eq? n 0)) (reverse lns)) ) ) ) ) )
+	(define (doread port)
+	  (do ([ln (read-line port) (read-line port)]
+	       [lns '() (cons ln lns)]
+	       [n (or max 1000000) (fx- n 1)] )
+	      ((or (eof-object? ln) (eq? n 0)) (reverse lns)) ) )
+	(if (string? port)
+	    (call-with-input-file port doread)
+	    (begin
+	      (##sys#check-port port 'read-lines)
+	      (doread port) ) ) ) ) ) )
 
 (define read-string
   (let ([open-output-string open-output-string]
 	[get-output-string get-output-string] )
     (lambda n-and-port
       (let-optionals n-and-port ([n #f] [p ##sys#standard-input])
+	(##sys#check-port p 'read-string)
 	(let ([str (open-output-string)])
 	  (when n (##sys#check-exact n 'read-string))
+	  (##sys#check-port p 'read-string)
 	  (let loop ([n n])
 	    (or (and (eq? n 0) (get-output-string str))
 		(let ([c (##sys#read-char-0 p)])
@@ -425,6 +467,7 @@ EOF
     (lambda (s . more)
       (##sys#check-string s 'write-string)
       (let-optionals more ([n #f] [port ##sys#standard-output])
+	(##sys#check-port port 'read-string)
 	(when n (##sys#check-exact n 'write-string))
 	(display 
 	 (if (and n (fx< n (##sys#size s)))
@@ -439,6 +482,7 @@ EOF
       (let ((p (if (##core#inline "C_eqp" port '())
 		   ##sys#standard-output
 		   (##sys#slot port 0) ) ) )
+	(##sys#check-port p 'write-line)
 	(##sys#check-string str 'write-line)
 	(display str p)
 	(newline p) ) ) ) )
@@ -615,12 +659,12 @@ EOF
 	      ((eof-object? obj)  (out "#<eof>" col))
 	      ((vector? obj)      (wr-lst (vector->list obj) (out "#" col)))
 	      ((boolean? obj)     (out (if obj "#t" "#f") col))
-	      ((number? obj)      (out (number->string obj) col))
+	      ((##sys#number? obj)      (out (##sys#number->string obj) col))
 	      ((symbol? obj)
 	       (let ([s (open-output-string)])
 		 (##sys#print obj #t s)
 		 (out (get-output-string s) col) ) )
-	      ((procedure? obj)   (out "#<procedure>" col))
+	      ((procedure? obj)   (out (##sys#procedure->string obj) col))
 	      ((string? obj)      (if display?
 				      (out obj col)
 				      (let loop ((i 0) (j 0) (col (out "\"" col)))
@@ -638,18 +682,32 @@ EOF
 						 (out (##sys#substring obj i j) col))))))
 	      ((char? obj)        (if display?
 				      (out (make-string 1 obj) col)
-				      (out (case obj
-					     ((#\space)   "space")
-					     ((#\newline) "newline")
-					     (else        (make-string 1 obj)))
-					   (out "#\\" col))))
+				      (let ([code (char->integer obj)])
+					(out "#\\" col)
+					(cond [(char-name obj) 
+					       => (lambda (cn) 
+						    (out (##sys#slot cn 1) col) ) ]
+					      [(fx< code 32)
+					       (out "x" col)
+					       (out (number->string code 16) col) ]
+					      [(fx> code 255)
+					       (out (if (fx> code #xffff) "U" "u") col)
+					       (out (number->string code 16) col) ]
+					      [else (out (make-string 1 obj) col)] ) ) ) )
+	      ((eof-object? obj)  (out "#<eof>" col))
 	      ((##core#inline "C_undefinedp" obj) (out "#<unspecified>" col))
+	      ((##core#inline "C_anypointerp" obj) (out (##sys#pointer->string obj) col))
 	      ((##sys#generic-structure? obj)
 	       (let ([o (open-output-string)])
 		 (##sys#user-print-hook obj #t o)
 		 (out (get-output-string o) col) ) )
 	      ((port? obj) (out (string-append "#<port " (##sys#slot obj 3) ">") col))
-	      ((eof-object? obj)  (out "#<eof>" col))
+	      ((##core#inline "C_lambdainfop" obj)
+	       (out "#<lambda info " col)
+	       (out (##sys#lambda-info->string obj) col)
+	       (out "#>" col) )
+	      ((eq? obj (##sys#slot '##sys#arbitrary-unbound-symbol 0))
+	       (out "#<unbound value>" col) )
 	      (else               (out "#<unprintable object>" col)) ) )
 
       (define (pp obj col)
@@ -670,7 +728,7 @@ EOF
 	(define (pr obj col extra pp-pair)
 	  (if (or (pair? obj) (vector? obj)) ; may have to split on multiple lines
 	      (let ((result '())
-		    (left (min (+ (- (- width col) extra) 1) max-expr-width)))
+		    (left (max (+ (- (- width col) extra) 1) max-expr-width)))
 		(generic-write obj display? #f
 			       (lambda (str)
 				 (set! result (cons str result))
@@ -855,11 +913,13 @@ EOF
 (define ->string 
   (let ([open-output-string open-output-string]
 	[display display]
+	[string string]
 	[get-output-string get-output-string] )
     (lambda (x)
       (cond [(string? x) x]
 	    [(symbol? x) (symbol->string x)]
-	    [(number? x) (number->string x)]
+	    [(char? x) (string x)]
+	    [(number? x) (##sys#number->string x)]
 	    [else 
 	     (let ([o (open-output-string)])
 	       (display x o)
@@ -992,7 +1052,7 @@ EOF
 
 ;;; Concatenate list of strings:
 
-(define (string-intersperse strs ds)
+(define (string-intersperse strs #!optional (ds " "))
   (##sys#check-list strs 'string-intersperse)
   (##sys#check-string ds 'string-intersperse)
   (let ((dslen (##sys#size ds)))
@@ -1148,10 +1208,10 @@ EOF
 		    ((#\S) (write (next) port))
 		    ((#\A) (display (next) port))
 		    ((#\C) (##sys#write-char-0 (next) port))
-		    ((#\B) (display (number->string (next) 2) port))
-		    ((#\O) (display (number->string (next) 8) port))
-		    ((#\X) (display (number->string (next) 16) port))
-		    ((#\!) (flush-output port))
+		    ((#\B) (display (##sys#number->string (next) 2) port))
+		    ((#\O) (display (##sys#number->string (next) 8) port))
+		    ((#\X) (display (##sys#number->string (next) 16) port))
+		    ((#\!) (##sys#flush-output port))
 		    ((#\?)
 		     (let* ([fstr (next)]
 			    [lst (next)] )
@@ -1159,6 +1219,7 @@ EOF
 		       (rec fstr lst) ) )
 		    ((#\~) (##sys#write-char-0 #\~ port))
 		    ((#\%) (newline port))
+		    ((#\% #\N) (newline port))
 		    (else
 		     (if (char-whitespace? dchar)
 			 (let skip ((c (fetch)))
@@ -1185,6 +1246,10 @@ EOF
 	(apply fprintf out fstr args)
 	(get-output-string out) ) ) ) )
 
+(define format sprintf)
+
+(register-feature! 'srfi-28)
+
 
 ;;; Defines: sorted?, merge, merge!, sort, sort!
 ;;; Author : Richard A. O'Keefe (based on Prolog code by D.H.D.Warren)
@@ -1361,9 +1426,22 @@ EOF
 
 ;;; Utility definitions:
 
-(define hashtab-default-size 301)
-(define hashtab-threshold 0.5)
-(define hashtab-primes-table '(301 613 997 1597 2011 2521 3001))
+(define-constant hashtab-default-size 307)
+(define-constant hashtab-threshold 0.5)
+
+; Predefined sizes for the hash tables:
+;
+; Start in 307; each element is the smallest prime that is at least twice as
+; bigger as the previous element in the list.  The last number is an
+; exception: it is the largest fixnum we can repressent.
+
+(define-constant hashtab-primes-table
+  '(307 617 1237 2477 4957 9923 19853 39709 79423 158849 317701 635413 1270849
+	2541701 5083423 10166857 20333759 40667527 81335063 162670129 325340273 
+	650680571 1073741823))
+
+(define-constant hashtab-max-size 1073741823)
+
 (define (hash-table? x) (##sys#structure? x 'hash-table))
 
 
@@ -1372,23 +1450,48 @@ EOF
 (define make-hash-table
   (let ([make-vector make-vector])
     (lambda test-and-size
-      (let-optionals test-and-size ([test eq?] [len hashtab-default-size])
+      (let-optionals test-and-size ([test equal?] [hashf %hash] [len hashtab-default-size])
 	(##sys#check-exact len 'make-hash-table)
-	(##sys#make-structure 'hash-table (make-vector len '()) 0 test) ) ) ) )
+	(##sys#make-structure 'hash-table (make-vector len '()) 0 test hashf) ) ) ) )
 
-(define clear-hash-table!
-  (let ([vector-fill! vector-fill!])
+(define hash-table-copy 
+  (let ((make-vector make-vector))
     (lambda (ht)
-      (##sys#check-structure ht 'hash-table 'clear-hash-table!)
-      (##sys#setslot ht 2 0)
-      (vector-fill! (##sys#slot ht 1) '()) ) ) )
+      (##sys#check-structure ht 'hash-table 'hash-table-copy)
+      (let* ((vec1 (##sys#slot ht 1))
+	     (len (##sys#size vec1))
+	     (vec2 (make-vector len '())) )
+	(do ((i 0 (fx+ i 1)))
+	    ((fx>= i len)
+	     (##sys#make-structure
+	      'hash-table vec2 (##sys#slot ht 2) (##sys#slot ht 3) (##sys#slot ht 4)) )
+	  (##sys#setslot 
+	   vec2 i
+	   (let copy ((lst (##sys#slot vec1 i)))
+	     (if (null? lst)
+		 '()
+		 (let ((b (##sys#slot lst 0)))
+		   (cons (cons (##sys#slot b 0) (##sys#slot b 1))
+			 (copy (##sys#slot lst 1))))))))))) )
+
+
+;;; Hash-table property-accessors
+
+(define (hash-table-equivalence-function ht)
+  (##sys#check-structure ht 'hash-table 'hash-table-equivalence-function) 
+  (##sys#slot ht 3) )
+
+(define (hash-table-hash-function ht)
+  (##sys#check-structure ht 'hash-table 'hash-table-hash-function) 
+  (##sys#slot ht 4) )
 
 
 ;;; Generation of hash-values:
 
 (define-constant hash-depth-limit 4)
+(define-constant default-hash-bound 536870912)
 
-(define hash
+(define %hash
   (lambda (x limit)
     (define (hash-with-test x d)
       (if (or (not (##core#inline "C_blockp" x)) (##core#inline "C_byteblockp" x) (symbol? x))
@@ -1403,7 +1506,6 @@ EOF
 	    ((eq? x '()) 258)
 	    ((##core#inline "C_eofp" x) 259)
 	    ((not (##core#inline "C_blockp" x)) 262)
-	    ((##sys#permanent? x) (##core#inline "C_hashptr" x))
 	    ((##core#inline "C_symbolp" x) (##core#inline "C_hash_string" (##sys#slot x 1)))
 	    ((list? x) (fx+ (length x) (hash-with-test (##sys#slot x 0) d)))
 	    ((pair? x) 
@@ -1423,68 +1525,53 @@ EOF
 			   (fx+ i 1)
 			   (fx- len 1) ) ) ) ) ) ) )
     (##sys#check-exact limit 'hash)
-    (##core#inline "C_fixnum_modulo" (bitwise-and #x00ffffff (rechash x 0)) limit) ) )
+    (##core#inline "C_fixnum_modulo" (fxand #x00ffffff (rechash x 0)) limit) ) )
+
+(define (hash x #!optional (bound default-hash-bound))
+  (##sys#check-exact bound 'hash)
+  (%hash x bound) )
+
+(define hash-by-identity hash)
+
+(define (string-hash s #!optional (bound default-hash-bound))
+  (##core#inline 
+   "C_fixnum_modulo"
+   (fxand #x00ffffff (##core#inline "C_hash_string" s))
+   limit) )
+
+(define (string-ci-hash s #!optional (bound default-hash-bound))
+  (##core#inline 
+   "C_fixnum_modulo"
+   (fxand #x00ffffff (##core#inline "C_hash_string_ci" s))
+   limit) )
 
 
 ;;; Access:
 
-(define (hash-table-count ht)
-  (##sys#check-structure ht 'hash-table 'hash-table-count)
+(define (hash-table-size ht)
+  (##sys#check-structure ht 'hash-table 'hash-table-size)
   (##sys#slot ht 2) )
 
-(define (hash-table-size ht)
-  (##sys#check-structure ht 'hash-table 'hash-table-count)
-  (##sys#size (##sys#slot ht 1)) )
-
-(define hash-table-ref
-  (let ([hash hash]
-	[eq0 eq?] )
-    (lambda (ht key . default)
-      (##sys#check-structure ht 'hash-table 'hash-table-ref)
-      (let* ([vec (##sys#slot ht 1)]
-	     [k (hash key (##sys#size vec))] 
-	     [def (and (pair? default) (car default))]
-	     [test (##sys#slot ht 3)] )
-	(if (eq? eq0 test)
-	    ;; Fast path (eq? test):
-	    (let loop ((bucket (##sys#slot vec k)))
-	      (if (eq? bucket '())
-		  def
-		  (let ((b (##sys#slot bucket 0)))
-		    (if (eq? key (##sys#slot b 0))
-			(##sys#slot b 1)
-			(loop (##sys#slot bucket 1)) ) ) ) )
-	    (let loop ((bucket (##sys#slot vec k)))
-	      (if (eq? bucket '())
-		  def
-		  (let ((b (##sys#slot bucket 0)))
-		    (if (test key (##sys#slot b 0))
-			(##sys#slot b 1)
-			(loop (##sys#slot bucket 1)) ) ) ) ) ) ) ) ) )
-
-(define hash-table-set! 
-  (let ([hash hash]
-	[eq0 eq?]
+(define hash-table-update! 
+  ;; This one was suggested by Sven Hartrumpf.
+  (let ([eq0 eq?]
 	[floor floor] )
-    (lambda (ht key val)
-      (##sys#check-structure ht 'hash-table 'hash-table-set!)
+    (lambda (ht key proc init)
+      (##sys#check-structure ht 'hash-table 'hash-table-update!)
       (let restart ()
 	(let* ((vec (##sys#slot ht 1))
+	       (hashf (##sys#slot ht 4))
 	       (len (##sys#size vec))
 	       (test (##sys#slot ht 3))
-	       (k (hash key len))
+	       (k (hashf key len))
 	       (c (fx+ (##sys#slot ht 2) 1)) )
-	  (if (fx>= c (inexact->exact (floor (* len hashtab-threshold))))
-	      (let* ((newlen
-		      (cond ((memq len hashtab-primes-table)
-			     => (lambda (n) 
-				  (let ((next (##sys#slot n 1)))
-				    (if (eq? next '())
-					(fx+ len 101) ; arbitrary
-					(##sys#slot next 0) ) ) ) )
-			    (else (fx+ len 101)) ) )
-		     (vec2 (make-vector newlen '())) )
-		(hashtab-rehash vec vec2)
+	  (if (and (fx>= c (inexact->exact (floor (* len hashtab-threshold))))
+		   (fx< len hashtab-max-size))
+	      (let ((vec2 (make-vector
+                            (##sys#hash-new-len hashtab-primes-table
+						(fxmin hashtab-max-size (fx* len 2)))
+                            '())))
+		(hashtab-rehash vec vec2 hashf)
 		(##sys#setslot ht 1 vec2)
 		(restart) ) 
 	      (let ((bucket0 (##sys#slot vec k)))
@@ -1492,33 +1579,91 @@ EOF
 		    ;; Fast path (eq? test):
 		    (let loop ((bucket bucket0))
 		      (cond ((eq? bucket '())
-			     (##sys#setslot vec k (cons (cons key val) bucket0))
+			     (##sys#setslot vec k (cons (cons key (init)) bucket0))
 			     (##sys#setslot ht 2 c) )
 			    (else
 			     (let ((b (##sys#slot bucket 0)))
 			       (if (eq? key (##sys#slot b 0))
-				   (##sys#setslot b 1 val)
+				   (##sys#setslot b 1 (proc (##sys#slot b 1)))
 				   (loop (##sys#slot bucket 1)) ) ) ) ) )
 		    (let loop ((bucket bucket0))
 		      (cond ((eq? bucket '())
-			     (##sys#setslot vec k (cons (cons key val) bucket0))
+			     (##sys#setslot vec k (cons (cons key (init)) bucket0))
 			     (##sys#setslot ht 2 c) )
 			    (else
 			     (let ((b (##sys#slot bucket 0)))
 			       (if (test key (##sys#slot b 0))
-				   (##sys#setslot b 1 val)
+				   (##sys#setslot b 1 (proc (##sys#slot b 1)))
 				   (loop (##sys#slot bucket 1)) ) ) ) ) ) ) ) ) ) ) ) ) )
 
-(define hash-table-remove!
-  (let ([hash hash]
-	[eq0 eq?]
+(define hash-table-update!/default 
+  (let ((hash-table-update! hash-table-update!))
+    (lambda (ht key func def)
+      (hash-table-update! ht key func (lambda () def)) ) ) )
+
+(define hash-table-set! 
+  (let ([hash-table-update! hash-table-update!])
+    (lambda (ht key val)
+      (hash-table-update! ht key (lambda (x) val) (lambda () val)) ) ) )
+
+(define hash-table-ref
+  (let ([eq0 eq?])
+    (getter-with-setter
+     (lambda (ht key . default)
+       (##sys#check-structure ht 'hash-table 'hash-table-ref)
+       (let* ([vec (##sys#slot ht 1)]
+	      [hashf (##sys#slot ht 4)]
+	      [k (hashf key (##sys#size vec))] 
+	      [def (if (pair? default) 
+		       (car default)
+		       (cut ##sys#signal-hook #:access-error 'hash-table-ref "hash-table does not contain key" key ht) ) ]
+	      [test (##sys#slot ht 3)] )
+	 (if (eq? eq0 test)
+	     ;; Fast path (eq? test):
+	     (let loop ((bucket (##sys#slot vec k)))
+	       (if (eq? bucket '())
+		   (def)
+		   (let ((b (##sys#slot bucket 0)))
+		     (if (eq? key (##sys#slot b 0))
+			 (##sys#slot b 1)
+			 (loop (##sys#slot bucket 1)) ) ) ) )
+	     (let loop ((bucket (##sys#slot vec k)))
+	       (if (eq? bucket '())
+		   (def)
+		   (let ((b (##sys#slot bucket 0)))
+		     (if (test key (##sys#slot b 0))
+			 (##sys#slot b 1)
+			 (loop (##sys#slot bucket 1)) ) ) ) ) ) ) )
+     hash-table-set!) ) )
+
+(define hash-table-ref/default
+  (let ((hash-table-ref hash-table-ref))
+    (lambda (ht key def)
+      (hash-table-ref ht key (lambda () def)) ) ) )
+
+(define hash-table-exists?
+  (let ((unique (vector 42))
+	(ref hash-table-ref/default) )
+    (lambda (ht key)
+      (##sys#check-structure ht 'hash-table 'hash-table-exists?)
+      (not (eq? unique (ref ht key unique))) ) ) )
+
+(define (##sys#hash-new-len tab req)
+  (if (or (fx>= (##sys#slot tab 0) req)
+	  (eq? (##sys#slot tab 1) '()))
+      (##sys#slot tab 0)
+      (##sys#hash-new-len (##sys#slot tab 1) req)))
+
+(define hash-table-delete!
+  (let ([eq0 eq?]
 	[floor floor])
     (lambda (ht key)
-      (##sys#check-structure ht 'hash-table 'hash-table-remove!)
+      (##sys#check-structure ht 'hash-table 'hash-table-delete!)
       (let* ((vec (##sys#slot ht 1))
+	     (hashf (##sys#slot ht 4))
 	     (len (##sys#size vec))
 	     (test (##sys#slot ht 3))
-	     (k (hash key len))
+	     (k (hashf key len))
 	     (c (fx- (##sys#slot ht 2) 1)))
 	(let ((bucket0 (##sys#slot vec k)))
 	  (if (eq? eq0 test)
@@ -1551,25 +1696,38 @@ EOF
 			  (loop bucket (##sys#slot bucket 1))))))))))))
 
 (define hashtab-rehash
-  (let ([hash hash])
-    (lambda (vec1 vec2)
-      (let ([len1 (##sys#size vec1)]
-	    [len2 (##sys#size vec2)] )
-	(do ([i 0 (fx+ i 1)])
-	    ((fx>= i len1))
-	  (let loop ([bucket (##sys#slot vec1 i)])
-	    (unless (null? bucket)
-	      (let* ([b (##sys#slot bucket 0)]
-		     [x (##sys#slot b 0)] 
-		     [k (hash x len2)] )
-		(##sys#setslot vec2 k (cons (cons x (##sys#slot b 1)) (##sys#slot vec2 k)))
-		(loop (##sys#slot bucket 1)) ) ) ) ) ) ) ) )
+  (lambda (vec1 vec2 hashf)
+    (let ([len1 (##sys#size vec1)]
+	  [len2 (##sys#size vec2)] )
+      (do ([i 0 (fx+ i 1)])
+	  ((fx>= i len1))
+	(let loop ([bucket (##sys#slot vec1 i)])
+	  (unless (null? bucket)
+	    (let* ([b (##sys#slot bucket 0)]
+		   [x (##sys#slot b 0)] 
+		   [k (hashf x len2)] )
+	      (##sys#setslot vec2 k (cons (cons x (##sys#slot b 1)) (##sys#slot vec2 k)))
+	      (loop (##sys#slot bucket 1)) ) ) ) ) ) ) )
+
+(define hash-table-merge!
+  (let ((hash-table-set! hash-table-set!))
+    (lambda (ht1 ht2)
+      (##sys#check-structure ht1 'hash-table 'hash-table-merge!)
+      (##sys#check-structure ht2 'hash-table 'hash-table-merge!)
+      (let* ((vec (##sys#slot ht2 1))
+	     (len (##sys#size vec)) )
+	(do ((i 0 (fx+ i 1)))
+	    ((fx>= i len) ht1)
+	  (do ((lst (##sys#slot vec i) (##sys#slot lst 1)))
+	      ((null? lst))
+	    (let ((b (##sys#slot lst 0)))
+	      (hash-table-set! ht1 (##sys#slot b 0) (##sys#slot b 1)) ) ) ) ) ) ) )
 
 
-;;; Conversion to list:
+;;; Conversion to and from lists:
 
-(define (hash-table->list ht)
-  (##sys#check-structure ht 'hash-table 'hash-table->list)
+(define (hash-table->alist ht)
+  (##sys#check-structure ht 'hash-table 'hash-table->alist)
   (let* ([vec (##sys#slot ht 1)]
 	 [len (##sys#size vec)] )
     (let loop ([i 0] [lst '()])
@@ -1582,12 +1740,50 @@ EOF
 		       (let ([x (##sys#slot bucket 0)])
 			 (cons (cons (##sys#slot x 0) (##sys#slot x 1)) lst) ) ) ) ) ) ) ) )
 
+(define alist->hash-table
+  (let ((make-hash-table make-hash-table)
+	(hash-table-set! hash-table-set!) )
+    (lambda (alist . rest)
+      (let ((hash (apply make-hash-table rest)))
+	(for-each (lambda (x) (hash-table-set! hash (car x) (cdr x))) alist)
+	hash))))
+
+(define (hash-table-keys ht)
+  (##sys#check-structure ht 'hash-table 'hash-table-keys)
+  (let* ([vec (##sys#slot ht 1)]
+	 [len (##sys#size vec)] )
+    (let loop ([i 0] [lst '()])
+      (if (fx>= i len)
+	  lst
+	  (let loop2 ([bucket (##sys#slot vec i)] [lst lst])
+	    (if (null? bucket)
+		(loop (fx+ i 1) lst)
+		(loop2 (##sys#slot bucket 1)
+		       (let ([x (##sys#slot bucket 0)])
+			 (cons (##sys#slot x 0) lst) ) ) ) ) ) ) ) )
+
+(define (hash-table-values ht)
+  (##sys#check-structure ht 'hash-table 'hash-table-values)
+  (let* ([vec (##sys#slot ht 1)]
+	 [len (##sys#size vec)] )
+    (let loop ([i 0] [lst '()])
+      (if (fx>= i len)
+	  lst
+	  (let loop2 ([bucket (##sys#slot vec i)] [lst lst])
+	    (if (null? bucket)
+		(loop (fx+ i 1) lst)
+		(loop2 (##sys#slot bucket 1)
+		       (let ([x (##sys#slot bucket 0)])
+			 (cons (##sys#slot x 1) lst) ) ) ) ) ) ) ) )
+
+(register-feature! 'srfi-69)
+
 
 ;;; Mapping over keys and elements:
 
-(define hash-table-for-each
-  (lambda (p ht)
-    (##sys#check-structure ht 'hash-table 'hash-table-for-each)
+(define hash-table-walk
+  (lambda (ht p)
+    (##sys#check-structure ht 'hash-table 'hash-table-walk)
     (let* ((vec (##sys#slot ht 1))
 	   (len (##sys#size vec)))
       (do ((i 0 (fx+ i 1)))
@@ -1595,29 +1791,24 @@ EOF
 	(##sys#for-each (lambda (bucket) 
 		      (p (##sys#slot bucket 0)
 			 (##sys#slot bucket 1) ) )
-		    (##sys#slot vec i) ) ) ) ) ) 
+		    (##sys#slot vec i) ) ) ) ) )
 
-
-;;; Using a hash-table as a disembodied property-list:
-
-(define get
-  (let ((hash-table-ref hash-table-ref))
-    (lambda (db key prop)
-      (let ((plist (hash-table-ref db key)))
-	(and plist
-	     (cond ((assq prop plist) => cdr)
-		   (else #f) ) ) ) ) ) )
-
-(define put!
-  (let ((hash-table-ref hash-table-ref)
-	(hash-table-set! hash-table-set!) )
-    (lambda (db key prop val)
-      (let ((plist (hash-table-ref db key)))
-	(if plist
-	    (cond ((assq prop plist) => (lambda (a) (##sys#setslot a 1 val)))
-		  (else 
-		   (##sys#setslot plist 1 (cons (cons prop val) (##sys#slot plist 1)))) )
-	    (hash-table-set! db key (cons (cons prop val) '())) ) ) ) ) )
+(define hash-table-fold
+  (lambda (ht p init)
+    (##sys#check-structure ht 'hash-table 'hash-table-fold)
+    (let* ((vec (##sys#slot ht 1))
+	   (len (##sys#size vec)))
+      (let loop ((i 0) (acc init))
+	(if (fx>= i len)
+	    acc
+	    (let fold2 ((buckets (##sys#slot vec i)) (acc acc))
+	      (if (null? buckets)
+		  (loop (fx+ i 1) acc)
+		  (let ((b (##sys#slot buckets 0)))
+		    (fold2 (##sys#slot buckets 1)
+			   (p (##sys#slot b 0)
+			      (##sys#slot b 1) 
+			      acc) ) ) ) ) ) ) ) ) )
 
 
 ; Support for queues
