@@ -1,6 +1,6 @@
 ;;;; lolevel.scm - Low-level routines for CHICKEN
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -37,7 +37,7 @@
   (unit lolevel)
   (uses extras)
   (usual-integrations)
-  (hide ipc-hook)
+  (hide ipc-hook xproc-tag)
   (foreign-declare #<<EOF
 #if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
 # include <sys/types.h>
@@ -47,17 +47,15 @@
 #endif
 #if defined(__i386__) && !defined(C_NONUNIX) && !defined(__CYGWIN__)
 # define C_valloc(n)               valloc(n)
-# define C_makeexecutable(a, n)    mprotect(a, n, PROT_READ | PROT_WRITE | PROT_EXEC)
 #elif defined(__i386__) || defined(_M_IX86)
 # define C_valloc(n)               malloc(n)
-# define C_makeexecutable(a, n)    0
 #else
 # define C_valloc(n)               NULL
-# define C_makeexecutable(a, n)    -1
 #endif
 
 #define C_pointer_to_object(ptr)   ((C_word*)C_block_item(ptr, 0))
 #define C_w2b(x)                   C_fix(C_wordstobytes(C_unfix(x)))
+#define C_pointer_eqp(x, y)        C_mk_bool(C_c_pointer_nn(x) == C_c_pointer_nn(y))
 EOF
 ) )
 
@@ -82,6 +80,8 @@ EOF
     (define-macro (##sys#check-exact . _) '(##core#undefined))
     (define-macro (##sys#check-port . _) '(##core#undefined))
     (define-macro (##sys#check-number . _) '(##core#undefined))
+    (define-macro (##sys#check-pointer . _) '(##core#undefined))
+    (define-macro (##sys#check-special . _) '(##core#undefined))
     (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
  [else] )
 
@@ -92,9 +92,9 @@ EOF
 
 (define move-memory!
   (let ([memmove1 (foreign-lambda void "C_memmove" c-pointer c-pointer int)]
-	[memmove2 (foreign-lambda void "C_memmove" c-pointer pointer int)]
-	[memmove3 (foreign-lambda void "C_memmove" pointer c-pointer int)]
-	[memmove4 (foreign-lambda void "C_memmove" pointer pointer int)]
+	[memmove2 (foreign-lambda void "C_memmove" c-pointer scheme-pointer int)]
+	[memmove3 (foreign-lambda void "C_memmove" scheme-pointer c-pointer int)]
+	[memmove4 (foreign-lambda void "C_memmove" scheme-pointer scheme-pointer int)]
 	[slot1structs '(mmap u8vector u16vector u32vector s8vector s16vector s32vector f32vector f64vector)] )
     (lambda (from to . n)
       (define (err) (##sys#error 'move-memory! "need number of bytes to move" from to))
@@ -132,6 +132,13 @@ EOF
 
 ;;; Pointer operations:
 
+(define (##sys#check-pointer ptr loc)
+  (unless (and (##core#inline "C_blockp" ptr)
+	       (or (##core#inline "C_pointerp" ptr)
+		   (##core#inline "C_swigpointerp" ptr)
+		   (##core#inline "C_taggedpointerp" ptr) ) )
+    (##sys#signal-hook #:type-error loc "bad argument type - not a pointer" ptr) ) )
+
 (define null-pointer ##sys#null-pointer)
 
 (define (pointer? x)
@@ -139,32 +146,21 @@ EOF
        (or (##core#inline "C_pointerp" x)
 	   (##core#inline "C_taggedpointerp" x) ) ) )
 
-(define address->pointer
-    (lambda (addr)
-      (cond-expand
-       [(not unsafe)
-	(when (not (integer? addr))
-	  (##sys#signal-hook #:type-error 'address->pointer "bad argument type - not an integer" addr) ) ]
-       [else] )
-      (##sys#address->pointer addr) ) )
+(define (address->pointer addr)
+  (cond-expand
+   [(not unsafe)
+    (when (not (integer? addr))
+      (##sys#signal-hook #:type-error 'address->pointer "bad argument type - not an integer" addr) ) ]
+   [else] )
+  (##sys#address->pointer addr) )
 
-(define pointer->address
-    (lambda (ptr)
-      (cond-expand 
-       [(not unsafe)
-	(when (not (and (##core#inline "C_blockp" ptr) (##core#inline "C_specialp" ptr)))
-	  (##sys#signal-hook #:type-error 'pointer->address "bad argument type - not a pointer" ptr) ) ]
-       [else] )
-      (##sys#pointer->address ptr) ) )
+(define (pointer->address ptr)
+  (##sys#check-special ptr 'pointer->address)
+  (##sys#pointer->address ptr) )
 
-(define null-pointer?
-    (lambda (ptr)
-      (cond-expand
-       [(not unsafe)
-	(when (not (and (##core#inline "C_blockp" ptr) (##core#inline "C_specialp" ptr)))
-	  (##sys#signal-hook #:type-error 'null-pointer? "bad argument type - not a pointer" ptr) ) ]
-       [else] )
-      (eq? 0 (##sys#pointer->address ptr) ) ) )
+(define (null-pointer? ptr)
+  (##sys#check-special ptr 'null-pointer?)
+  (eq? 0 (##sys#pointer->address ptr) ) )
 
 (define (object->pointer x)
   (and (##core#inline "C_blockp" x)
@@ -173,13 +169,14 @@ EOF
 	x) ) )
 
 (define (pointer->object ptr)
-  (cond-expand
-   [(not unsafe)
-    (when (not (and (##core#inline "C_blockp" ptr) (##core#inline "C_pointerp" ptr)))
-      (##sys#signal-hook #:type-error 'pointer->object "bad argument type - not a pointer" ptr) ) ]
-   [else] )
+  (##sys#check-pointer ptr 'pointer->object)
   (##core#inline "C_pointer_to_object" ptr) )
 
+(define (pointer=? p1 p2)
+  (##sys#check-special p1 'pointer=?)
+  (##sys#check-special p2 'pointer=?)
+  (##core#inline "C_pointer_eqp" p1 p2) )
+
 (define allocate (foreign-lambda c-pointer "C_malloc" int))
 (define free (foreign-lambda void "C_free" c-pointer))
 
@@ -204,14 +201,45 @@ EOF
 (define pointer-f32-set! (foreign-lambda* void ([c-pointer p] [double n]) "*((float *)p) = n;"))
 (define pointer-f64-set! (foreign-lambda* void ([c-pointer p] [float n]) "*((double *)p) = n;"))
 
-(define pointer-u8-ref (foreign-lambda* int ([c-pointer p]) "return(*((unsigned char *)p));"))
-(define pointer-s8-ref (foreign-lambda* int ([c-pointer p]) "return(*((char *)p));"))
-(define pointer-u16-ref (foreign-lambda* int ([c-pointer p]) "return(*((unsigned short *)p));"))
-(define pointer-s16-ref (foreign-lambda* int ([c-pointer p]) "return(*((short *)p));"))
-(define pointer-u32-ref (foreign-lambda* integer ([c-pointer p]) "return(*((C_u32 *)p));"))
-(define pointer-s32-ref (foreign-lambda* integer ([c-pointer p]) "return(*((C_s32 *)p));"))
-(define pointer-f32-ref (foreign-lambda* float ([c-pointer p]) "return(*((float *)p));"))
-(define pointer-f64-ref (foreign-lambda* double ([c-pointer p]) "return(*((double *)p));"))
+(define pointer-u8-ref
+  (getter-with-setter
+   (foreign-lambda* int ([c-pointer p]) "return(*((unsigned char *)p));")
+   pointer-u8-set!) )
+
+(define pointer-s8-ref
+  (getter-with-setter
+   (foreign-lambda* int ([c-pointer p]) "return(*((char *)p));")
+   pointer-s8-set!) )
+
+(define pointer-u16-ref
+  (getter-with-setter
+   (foreign-lambda* int ([c-pointer p]) "return(*((unsigned short *)p));")
+   pointer-u16-set!) )
+
+(define pointer-s16-ref
+  (getter-with-setter
+   (foreign-lambda* int ([c-pointer p]) "return(*((short *)p));")
+   pointer-s6-set!) )
+
+(define pointer-u32-ref
+  (getter-with-setter
+   (foreign-lambda* integer ([c-pointer p]) "return(*((C_u32 *)p));")
+   pointer-u32-set!) )
+
+(define pointer-s32-ref
+  (getter-with-setter
+   (foreign-lambda* integer ([c-pointer p]) "return(*((C_s32 *)p));")
+   pointer-s32-set!) )
+
+(define pointer-f32-ref
+  (getter-with-setter
+   (foreign-lambda* float ([c-pointer p]) "return(*((float *)p));")
+   pointer-f32-set!) )
+
+(define pointer-f64-ref
+  (getter-with-setter
+   (foreign-lambda* double ([c-pointer p]) "return(*((double *)p));")
+   pointer-f64-set!) )
 
 (define (tag-pointer ptr tag)
   (let ([tp (##sys#make-tagged-pointer tag)])
@@ -223,7 +251,7 @@ EOF
 (define (tagged-pointer? x tag)
   (and (##core#inline "C_blockp" x) 
        (##core#inline "C_taggedpointerp" x)
-       (eq? tag (##sys#slot x 1)) ) )
+       (equal? tag (##sys#slot x 1)) ) )
 
 (define (pointer-tag x)
   (if (and (##core#inline "C_blockp" x) (##core#inline "C_specialp" x))
@@ -234,55 +262,38 @@ EOF
 
 ;;; Procedures extended with data:
 
-(define extend-procedure
-  (let ([make-vector make-vector] )
-    (lambda (proc x)
-      (cond-expand 
-       [(not unsafe)
-	(unless (##core#inline "C_closurep" proc)
-	  (##sys#signal-hook #:type-error 'extend-procedure "bad argument type - not a procedure" proc) ) ]
-       [else] )
-      (let ([len (##sys#size proc)])
-	(if (and (fx> len 1) (eq? ##sys#snafu (##sys#slot proc (fx- len 1)))) 
-	    (begin
-	      (##sys#setslot proc (fx- len 2) x)
-	      proc)
-	    (let* ([len2 (fx+ len 2)]
-		   [p2 (make-vector len2)] )
-	      (do ([i 1 (fx+ i 1)])
-		  ((fx>= i len)
-		   (##sys#setslot p2 i x)
-		   (##sys#setslot p2 (fx+ i 1) ##sys#snafu)
-		   (##sys#vector->closure! p2 (##sys#peek-unsigned-integer proc 0))
-		   p2)
-		(##sys#setslot p2 i (##sys#slot proc i)) ) ) ) ) ) ) )
+(define xproc-tag (vector 'extended))
+
+(define (extend-procedure proc data)
+  #+(not unsafe)
+  (unless (##core#inline "C_closurep" proc)
+    (##sys#signal-hook #:type-error 'extend-procedure "bad argument type - not a procedure" proc) )
+  (##sys#decorate-lambda
+   proc
+   (lambda (x) (and (pair? x) (eq? xproc-tag (##sys#slot x 0)))) 
+   (lambda (x i)
+     (##sys#setslot x i (cons xproc-tag data))
+     x) ) )
 
 (define (extended-procedure? x)
   (and (##core#inline "C_blockp" x)
        (##core#inline "C_closurep" x)
-       (let ([len (##sys#size x)])
-	 (and (fx> len 1)
-	      (eq? ##sys#snafu (##sys#slot x (fx- len 1))) ) ) ) )
+       (##sys#lambda-decoration x (lambda (x) (and (pair? x) (eq? xproc-tag (##sys#slot x 0)))))
+       #t) )
 
-(define procedure-data 
-  (let ([extended-procedure? extended-procedure?] )
-    (lambda (proc)
-      (cond-expand 
-       [(not unsafe)
-	(unless (extended-procedure? proc)
-	  (##sys#signal-hook #:type-error 'procedure-data "bad argument type - not an extended procedure") ) ]
-       [else] )
-      (##sys#slot proc (fx- (##sys#size proc) 2)) ) ) )
+(define (procedure-data x)
+  (and (##core#inline "C_blockp" x)
+       (##core#inline "C_closurep" x)
+       (and-let* ((d (##sys#lambda-decoration x (lambda (x) (and (pair? x) (eq? xproc-tag (##sys#slot x 0)))))))
+	 (##sys#slot d 1) ) ) )
 
 (define set-procedure-data!
-  (let ([extended-procedure? extended-procedure?] )
+  (let ((extend-procedure extend-procedure))
     (lambda (proc x)
-      (cond-expand
-       [(not unsafe)
-	(unless (extended-procedure? proc) 
-	  (##sys#signal-hook #:type-error 'set-procedure-data! "bad argument type - not an extended procedure") ) ]
-       [else] )
-      (##sys#setslot proc (fx- (##sys#size proc) 2) x) ) ) )
+      (let ((p2 (extend-procedure proc x)))
+	(if (eq? p2 proc)
+	    proc
+	    (##sys#signal-hook #:type-error 'set-procedure-data! "bad argument type - not an extended procedure" proc) ) ) ) ) )
 
 
 ;;; Bytevector stuff:
@@ -318,15 +329,6 @@ EOF
 	    ((fx>= i n) bv)
 	  (##sys#setbyte bv i (##sys#slot bytes 0)) ) ) ) ) )
 
-(define byte-vector-ref
-  (lambda (bv i)
-    (##sys#check-byte-vector bv 'byte-vector-ref)
-    (##sys#check-exact i 'byte-vector-ref)
-    (let ([n (##sys#size bv)])
-      (if (or (fx< i 0) (fx>= i n))
-	  (##sys#error 'byte-vector-ref "out of range" bv i)
-	  (##sys#byte bv i) ) ) ) )
-
 (define byte-vector-set!
   (lambda (bv i x)
     (##sys#check-byte-vector bv 'byte-vector-set!)
@@ -337,6 +339,17 @@ EOF
 	  (##sys#error 'byte-vector-set! "out of range" bv i)
 	  (##sys#setbyte bv i x) ) ) ) )
 
+(define byte-vector-ref
+  (getter-with-setter
+   (lambda (bv i)
+     (##sys#check-byte-vector bv 'byte-vector-ref)
+     (##sys#check-exact i 'byte-vector-ref)
+     (let ([n (##sys#size bv)])
+       (if (or (fx< i 0) (fx>= i n))
+	   (##sys#error 'byte-vector-ref "out of range" bv i)
+	   (##sys#byte bv i) ) ) )
+   byte-vector-set!) )
+
 (define (byte-vector->list bv)
   (##sys#check-byte-vector bv 'byte-vector->list)
   (let ([len (##sys#size bv)])
@@ -392,13 +405,6 @@ EOF
            if((bv = (char *)malloc(size + 3 + sizeof(C_header))) == NULL) return(C_SCHEME_FALSE);
            bv = (char *)C_align((C_word)bv);
            ((C_SCHEME_BLOCK *)bv)->header = C_BYTEVECTOR_TYPE | size;
-           return((C_word)bv);") ]
-      [xalloc
-       (foreign-lambda* scheme-object ((int size))
-	 "char *bv;
-           if((bv = (char *)C_valloc(size + sizeof(C_header))) == NULL) return(C_SCHEME_FALSE);
-           if(C_makeexecutable(bv, size) == -1) return(C_SCHEME_FALSE);
-           ((C_SCHEME_BLOCK *)bv)->header = C_BYTEVECTOR_TYPE | size;
            return((C_word)bv);") ] )
   (define (make size init alloc loc)
     (##sys#check-exact size loc)
@@ -409,8 +415,7 @@ EOF
 		 (when (pair? init) (byte-vector-fill! bv (##sys#slot init 0)))
 		 bv]
 		[else (##sys#signal-hook #:runtime-error "can not allocate statically allocated bytevector" size)] ) ) ) )
-  (set! make-static-byte-vector (lambda (size . init) (make size init malloc 'make-static-byte-vector)))
-  (set! make-executable-byte-vector (lambda (size . init) (make size init xalloc 'make-executable-byte-vector))) )
+  (set! make-static-byte-vector (lambda (size . init) (make size init malloc 'make-static-byte-vector))))
 
 (define static-byte-vector->pointer 
   (lambda (bv)
@@ -424,24 +429,8 @@ EOF
 
 ;;; Accessors for arbitrary block objects:
 
-(let ([check
-       (lambda (x i)
-	 (when (or (not (##core#inline "C_blockp" x)) 
-		   (and (##core#inline "C_specialp" x) (fx= i 0))
-		   (##core#inline "C_byteblockp" x) ) 
-	   (##sys#error "slot not accessible" x i) )
-	 (when (or (fx< i 0) (fx>= i (##sys#size x)))
-	   (##sys#error "slot-reference out of range" x i) ) ) ] )
-  (set! block-ref
-    (lambda (x i)
-      (##sys#check-exact i 'block-ref)
-      (check x i)
-      (##sys#slot x i) ) )
-  (set! block-set!
-    (lambda (x i y)
-      (##sys#check-exact i 'block-set!)
-      (check x i)
-      (##sys#setslot x i y) ) ) )
+(define block-set! ##sys#block-set!)
+(define block-ref (getter-with-setter ##sys#block-ref ##sys#block-set!))
 
 (define number-of-slots 
   (lambda (x)
@@ -478,29 +467,6 @@ EOF
       (##sys#signal-hook #:type-error 'record->vector "bad argument type - not a record structure" x) ) )
 
 
-;;; Convert executable bytevector into a procedure:
-
-(define executable-byte-vector->procedure
-  (lambda (bv)
-    (##sys#check-byte-vector bv 'executable-byter-vector->procedure)
-    (if (##core#inline "C_permanentp" bv)
-	((foreign-lambda* scheme-object ((scheme-object bv) (scheme-object proc))
-	   "C_set_block_item(proc, 0, (C_word)(&C_block_item(bv, 0)));
-              return(proc);")
-	 bv (lambda () #f) )
-	(##sys#signal-hook #:type-error 'executable-byte-vector->procedure 
-			   "can not use non-static byte-vector for procedure code pointer" bv) ) ) )
-
-
-;;; Call code in executable bytevector:
-
-(define (invoke-executable-byte-vector bv . args)
-  (##sys#check-byte-vector bv 'invoke-executable-byte-vector)
-  ((foreign-lambda* scheme-object ((pointer bv) (pointer args))
-     "return(((C_word (*)(void *args))bv)(args));")
-   bv (##sys#list->vector args) ) )
-
-
 ;;; Copy arbitrary object:
 
 (define object-copy
@@ -526,17 +492,17 @@ EOF
 
 (define object-evict
   (let ([make-hash-table make-hash-table]
-	[hash-table-ref hash-table-ref]
+	[hash-table-ref/default hash-table-ref/default]
 	[hash-table-set! hash-table-set!] )
     (lambda (x . allocator)
       (let ([allocator 
 	     (if (pair? allocator) 
 		 (car allocator)
 		 (foreign-lambda c-pointer "C_malloc" int) ) ] 
-	    [tab (make-hash-table)] )
+	    [tab (make-hash-table eq?)] )
 	(let evict ([x x])
 	  (cond [(not (##core#inline "C_blockp" x)) x]
-		[(hash-table-ref tab x)]
+		[(hash-table-ref/default tab x #f)]
 		[else
 		 (let* ([n (##sys#size x)]
 			[bytes (if (##core#inline "C_byteblockp" x) (align-to-word n) (##core#inline "C_bytes" n))]
@@ -564,11 +530,11 @@ EOF
 		   (do ([i (if (##core#inline "C_specialp" x) 1 0) (fx+ i 1)])
 		       ((fx>= i n))
 		     (release (##sys#slot x i))) )
-		 (free (##sys#address->pointer (##core#inline "C_block_address" x))) ) ] ) ) ) ) )
+		 (free (##sys#address->pointer (##core#inline_allocate ("C_block_address" 4) x))) ) ] ) ) ) ) )
 
 (define object-evict-to-location
   (let ([make-hash-table make-hash-table]
-	[hash-table-ref hash-table-ref]
+	[hash-table-ref/default hash-table-ref/default]
 	[align-to-word align-to-word]
 	[hash-table-set! hash-table-set!] )
     (lambda (x ptr . limit)
@@ -584,11 +550,11 @@ EOF
 		    limit)
 		  #f) ]
 	     [ptr2 (##sys#address->pointer (##sys#pointer->address ptr))]
-	     [tab (make-hash-table)]
+	     [tab (make-hash-table eq?)]
 	     [x2
 	      (let evict ([x x])
 		(cond [(not (##core#inline "C_blockp" x)) x]
-		      [(hash-table-ref tab x)]
+		      [(hash-table-ref/default tab x #f)]
 		      [else
 		       (let* ([n (##sys#size x)]
 			      [bytes 
@@ -596,7 +562,14 @@ EOF
 				    (##core#inline "C_bytes" 1) ) ] )
 			 (when limit
 			   (set! limit (fx- limit bytes))
-			   (when (fx< limit 0) (##sys#error 'object-evict-to-location "can not evict object - limit exceeded" x)) )
+			   (when (fx< limit 0) 
+			     (signal
+			      (make-composite-condition
+			       (make-property-condition
+				'exn 'location 'object-evict-to-location
+				'message "can not evict object - limit exceeded" 
+				'arguments (list x limit))
+			       (make-property-condition 'evict 'limit limit) ) ) ) )
 			 (let ([y (##core#inline "C_evict_block" x ptr2)])
 			   (when (symbol? x) (##sys#setislot y 0 (##core#undefined)))
 			   (##sys#set-pointer-address! ptr2 (+ (##sys#pointer->address ptr2) bytes))
@@ -613,14 +586,14 @@ EOF
 
 (define object-size
   (let ([make-hash-table make-hash-table]
-	[hash-table-ref hash-table-ref]
+	[hash-table-ref/default hash-table-ref/default]
 	[align-to-word align-to-word]
 	[hash-table-set! hash-table-set!] )
     (lambda (x)
-      (let ([tab (make-hash-table)])
+      (let ([tab (make-hash-table eq?)])
 	(let evict ([x x])
 	  (cond [(not (##core#inline "C_blockp" x)) 0]
-		[(hash-table-ref tab x) 0]
+		[(hash-table-ref/default tab x #f) 0]
 		[else
 		 (let* ([n (##sys#size x)]
 			[bytes
@@ -640,16 +613,21 @@ EOF
   (let ([make-vector make-vector]
 	[make-hash-table make-hash-table]
 	[hash-table-set! hash-table-set!]
-	[hash-table-ref hash-table-ref] )
-    (lambda (x)
+	[hash-table-ref/default hash-table-ref/default] )
+    (lambda (x #!optional (full #f))
       (define (err x)
 	(##sys#signal-hook #:type-error 'object-unevict "can not copy object" x) )
-      (let ([tab (make-hash-table)])
+      (let ([tab (make-hash-table eq?)])
 	(let copy ([x x])
 	  (cond [(not (##core#inline "C_blockp" x)) x]
 		[(not (##core#inline "C_permanentp" x)) x]
-		[(##core#inline "C_byteblockp" x) x]
-		[(hash-table-ref tab x)]
+		[(hash-table-ref/default tab x #f)]
+		[(##core#inline "C_byteblockp" x) 
+		 (if full
+		     (let ([y (##core#inline "C_copy_block" x (##sys#make-string (##sys#size x)))])
+		       (hash-table-set! tab x y)
+		       y) 
+		     x) ]
 		[(symbol? x) 
 		 (let ([y (##sys#intern-symbol (##sys#slot x 1))])
 		   (hash-table-set! tab x y)
@@ -694,8 +672,8 @@ EOF
 (define (make-weak-locative obj . index)
   (##sys#make-locative obj (:optional index 0) #t 'make-weak-locative) )
 
-(define locative-ref (##core#primitive "C_locative_ref"))
 (define (locative-set! x y) (##core#inline "C_i_locative_set" x y))
+(define locative-ref (getter-with-setter (##core#primitive "C_locative_ref") locative-set!))
 (define (locative->object x) (##core#inline "C_i_locative_to_object" x))
 (define (locative? x) (and (##core#inline "C_blockp" x) (##core#inline "C_locativep" x)))
 
