@@ -1,6 +1,6 @@
 ;;;; optimizer.scm - The CHICKEN Scheme compiler (optimizations)
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -36,12 +36,12 @@
 (declare (unit optimizer))
 
 #{compiler
-  build-information compiler-arguments process-command-line perform-lambda-lifting!
+  compiler-arguments process-command-line perform-lambda-lifting!
   default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home decompose-lambda-list external-to-pointer
-  foreign-type-table-size copy-node! export-list
+  foreign-type-table-size copy-node! export-list inline-list not-inline-list
   unit-name insert-timer-checks used-units external-variables
   debug-info-index debug-info-vector-name profile-info-vector-name
   foreign-declarations emit-trace-info block-compilation analysis-database-size line-number-database-size
@@ -58,7 +58,7 @@
   reorganize-recursive-bindings substitution-table simplify-named-call
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
   transform-direct-lambdas! expand-foreign-callback-lambda* debug-lambda-list debug-variable-list debugging
-  debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
+  debugging-chicken bomb check-signature posq stringify symbolify flonum? build-lambda-list
   string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
   put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
@@ -67,10 +67,10 @@
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database
   expand-foreign-callback-lambda default-optimization-passes default-optimization-passes-when-trying-harder
-  copyright compiler-features units-used-by-default words-per-flonum rewrite
+  units-used-by-default words-per-flonum rewrite
   parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
-  make-random-name final-foreign-type
+  make-random-name final-foreign-type inline-max-size simplified-ops
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
   foreign-argument-conversion foreign-result-conversion foreign-type-convert-argument foreign-type-convert-result}
 
@@ -149,6 +149,7 @@
 ;   also: change procedure's lambda-list.
 
 (define simplifications (make-vector 301 '()))
+(define simplified-ops '())
 
 (define (perform-high-level-optimizations node db)
   (let ([removed-lets 0]
@@ -164,22 +165,21 @@
     (define (touch) (set! dirty #t))
 
     (define (simplify n)
-      (and (memq 'apply-simplifications compiler-features)
-	   (or (and-let* ([entry (##sys#hash-table-ref simplifications (node-class n))])
-		 (any (lambda (s)
-			(and-let* ([vars (second s)]
-				   [env (match-node n (first s) vars)] 
-				   [n2 (apply (third s) db
-					      (map (lambda (v) (cdr (assq v env))) vars) ) ] )
-			  (let* ([name (caar s)]
-				 [counter (assq name simplified-classes)] )
-			    (if counter
-				(set-cdr! counter (add1 (cdr counter)))
-				(set! simplified-classes (alist-cons name 1 simplified-classes)) )
-			    (touch)
-			    (simplify n2) ) ) )
-		      entry) )
-	       n) ) )
+      (or (and-let* ([entry (##sys#hash-table-ref simplifications (node-class n))])
+	    (any (lambda (s)
+		   (and-let* ([vars (second s)]
+			      [env (match-node n (first s) vars)] 
+			      [n2 (apply (third s) db
+					 (map (lambda (v) (cdr (assq v env))) vars) ) ] )
+		     (let* ([name (caar s)]
+			    [counter (assq name simplified-classes)] )
+		       (if counter
+			   (set-cdr! counter (add1 (cdr counter)))
+			   (set! simplified-classes (alist-cons name 1 simplified-classes)) )
+		       (touch)
+		       (simplify n2) ) ) )
+		 entry) )
+	  n) )
 
     (define (walk n)
       (if (memq n broken-constant-nodes)
@@ -299,62 +299,78 @@
 		       [lval (and (not (test var 'unknown)) (test var 'value))]
 		       [args (cdr subs)] )
 		  (cond [(test var 'contractable)
-			 (let* ((lparams (node-parameters lval))
-				(llist (third lparams)) )
+			 (let* ([lparams (node-parameters lval)]
+				[llist (third lparams)] )
 			   (check-signature var args llist)
 			   (debugging 'o "contracted procedure" var)
 			   (touch)
-			   (walk (inline-lambda-bindings llist args (first (node-subexpressions lval)))) ) ]
+			   (walk (inline-lambda-bindings llist args (first (node-subexpressions lval)) #f)) ) ]
 			[(and lval (eq? '##core#lambda (node-class lval)))
-			 (decompose-lambda-list
-			  (third (node-parameters lval))
-			  (lambda (vars argc rest)
-			    (cond [(test (first (node-parameters lval)) 'has-unused-parameters)
-				   (if (< (length args) argc) ; Expression was already optimized (should this happen?)
-				       (walk-generic n class params subs)
-				       (let loop ((vars vars) (argc argc) (args args) (used '()))
-					 (cond [(or (null? vars) (zero? argc))
-						(touch)
-						(make-node
-						 '##core#call
-						 params
-						 (map walk (cons fun (append-reverse used args))) ) ]
-					       [(test (car vars) 'unused)
-						(touch)
-						(debugging 'o "removed unused parameter to known procedure" (car vars) var)
-						(if (expression-has-side-effects? (car args) db)
+			 (let* ([lparams (node-parameters lval)]
+				[llist (third lparams)] )
+			   (decompose-lambda-list
+			    llist
+			    (lambda (vars argc rest)
+			      (let ([fid (first lparams)])
+				(cond [(and (test fid 'simple)
+					    (test var 'inlinable)
+					    (not (memq var not-inline-list))
+					    (or (memq var inline-list)
+						(< (fourth lparams) inline-max-size) ) )
+				       (debugging 'i "procedure inlinable" var fid (fourth lparams))
+				       (check-signature var args llist)
+				       (debugging 'o "inlining procedure" var)
+				       (touch)
+				       (walk (inline-lambda-bindings llist args (first (node-subexpressions lval)) #t)) ]
+				      [(test fid 'has-unused-parameters)
+				       (if (< (length args) argc) ; Expression was already optimized (should this happen?)
+					   (walk-generic n class params subs)
+					   (let loop ((vars vars) (argc argc) (args args) (used '()))
+					     (cond [(or (null? vars) (zero? argc))
+						    (touch)
 						    (make-node
-						     'let
-						     (list (gensym 't))
-						     (list (walk (car args))
-							   (loop (cdr vars) (sub1 argc) (cdr args) used) ) )
-						    (loop (cdr vars) (sub1 argc) (cdr args) used) ) ]
-					       [else (loop (cdr vars)
-							   (sub1 argc)
-							   (cdr args)
-							   (cons (car args) used) ) ] ) ) ) ]
-				  [(and (test (first (node-parameters lval)) 'explicit-rest)
-					(not (memq n rest-consers)) ) ; make sure we haven't inlined rest-list already
-				   (let ([n (length (third (node-parameters lval)))])
-				     (debugging 'o "consed rest parameter at call site" var n)
-				     (let-values ([(args rargs) (split-at args n)])
-				       (let ([n2 (make-node
-						  '##core#call
-						  params
-						  (map walk
-						       (cons fun
-							     (append 
-							      args
-							      (list
-							       (if (null? rargs)
-								   (qnode '())
-								   (make-node
-								    '##core#inline_allocate 
-								    (list "C_a_i_list" (* 3 (length rargs)))
-								    rargs) ) ) ) ) ) ) ] )
-					 (set! rest-consers (cons n2 rest-consers))
-					 n2) ) ) ]
-				  [else (walk-generic n class params subs)] ) ) ) ]
+						     '##core#call
+						     params
+						     (map walk (cons fun (append-reverse used args))) ) ]
+						   [(test (car vars) 'unused)
+						    (touch)
+						    (debugging 'o "removed unused parameter to known procedure" (car vars) var)
+						    (if (expression-has-side-effects? (car args) db)
+							(make-node
+							 'let
+							 (list (gensym 't))
+							 (list (walk (car args))
+							       (loop (cdr vars) (sub1 argc) (cdr args) used) ) )
+							(loop (cdr vars) (sub1 argc) (cdr args) used) ) ]
+						   [else (loop (cdr vars)
+							       (sub1 argc)
+							       (cdr args)
+							       (cons (car args) used) ) ] ) ) ) ]
+				      [(and (test fid 'explicit-rest)
+					    (not (memq n rest-consers)) ) ; make sure we haven't inlined rest-list already
+				       (let ([n (length llist)])
+					 (if (< (length args) n)
+					     (walk-generic n class params subs)
+					     (begin
+					       (debugging 'o "consed rest parameter at call site" var n)
+					       (let-values ([(args rargs) (split-at args n)])
+						 (let ([n2 (make-node
+							    '##core#call
+							    params
+							    (map walk
+								 (cons fun
+								       (append 
+									args
+									(list
+									 (if (null? rargs)
+									     (qnode '())
+									     (make-node
+									      '##core#inline_allocate 
+									      (list "C_a_i_list" (* 3 (length rargs)))
+									      rargs) ) ) ) ) ) ) ] )
+						   (set! rest-consers (cons n2 rest-consers))
+						   n2) ) ) ) ) ]
+				      [else (walk-generic n class params subs)] ) ) ) ) ) ]
 			[else (walk-generic n class params subs)] ) ) ]
 	       [(##core#lambda)
 		(if (first params)
@@ -389,8 +405,17 @@
 	(values node #t)
 	(begin
 	  (debugging 'p "traversal phase...")
+	  (set! simplified-ops '())
 	  (let ((node2 (walk node)))
 	    (when (pair? simplified-classes) (debugging 'o "simplifications" simplified-classes))
+	    (when (and (pair? simplified-ops) (debugging 'o "  call simplifications:"))
+	      (for-each
+	       (lambda (p)
+		 (print* #\tab (car p))
+		 (if (> (cdr p) 1)
+		     (print #\tab (cdr p))
+		     (newline) ) )
+	       simplified-ops) )
 	    (when (> replaced-vars 0) (debugging 'o "replaced variables" replaced-vars))
 	    (when (> removed-lets 0) (debugging 'o "removed binding forms" removed-lets))
 	    (when (> removed-ifs 0) (debugging 'o "removed conditional forms" removed-ifs))
@@ -508,7 +533,13 @@
    ,(lambda (db a b c d)
       (let loop ((entries (or (##sys#hash-table-ref substitution-table a) '())))
 	(cond ((null? entries) #f)
-	      ((simplify-named-call db d a b (caar entries) (cdar entries) c))
+	      ((simplify-named-call db d a b (caar entries) (cdar entries) c)
+	       => (lambda (r)
+		    (let ((as (assq a simplified-ops)))
+		      (if as 
+			  (set-cdr! as (add1 (cdr as)))
+			  (set! simplified-ops (alist-cons a 1 simplified-ops)) ) )
+		    r) )
 	      (else (loop (cdr entries))) ) ) ) ) )
 
 
@@ -532,7 +563,6 @@
    (var0 var1 var2 op const1 const2 body1 body2 d1 d2 rest)
    ,(lambda (db var0 var1 var2 op const1 const2 body1 body2 d1 d2 rest)
       (and (equal? op eq-inline-operator)
-	   (memq 'target-has-switch compiler-features)
 	   (immediate? const1)
 	   (immediate? const2)
 	   (= 1 (length (get db var1 'references)))
@@ -560,7 +590,6 @@
    (var op var0 const d body n clauses)
    ,(lambda (db var op var0 const d body n clauses)
       (and (equal? op eq-inline-operator)
-	   (memq 'target-has-switch compiler-features)
 	   (immediate? const)
 	   (= 1 (length (get db var 'references)))
 	   (make-node
@@ -623,6 +652,21 @@
 				     [else #f] ) ) ) ) ) ]
 		   [else #f] ) ) ) ) ) ) )
 
+ ;; (let ((<var1> <var2>))
+ ;;   (<var1> ...) )
+ ;; -> (<var2> ...)
+ ;; - <var1> used only once
+ #| this doesn't seem to work (Sven Hartrumpf):
+ `((let (var1) (##core#variable (var2))
+	(##core#call p (##core#variable (var1)) . more) ) ; `p' was `#t', bombed also
+   (var1 var2 p more)
+   ,(lambda (db var1 var2 p more)
+      (and (= 1 (length (get db var1 'references)))
+	   (make-node
+	    '##core#call p
+	    (cons (varnode var2) more) ) ) ) )
+ |#
+
  ;; (let ((<var> (##core#inline <op> ...)))
  ;;   (if <var> <x> <y>) )
  ;; -> (if (##core#inline <op> ...) <x> <y>)
@@ -792,7 +836,6 @@
     (##sys#hash-table-set! substitution-table name (append old (list class-and-args))) ) )
 
 (define (simplify-named-call db params name cont class classargs callargs)
-
   (define (test sym prop) (get db sym prop))
 
   (case class
@@ -1103,6 +1146,62 @@
 			      (list (qnode (third classargs)))
 			      tail) ) ) ) ) ) ) )
 
+    ;; (<op>) -> <id>
+    ;; (<op> <x>) -> <x>
+    ;; (<op> <x1> ...) -> (##core#inline_allocate (<genop> <words>) <x1> (##core#inline_allocate (<genop> <words>) ...))
+    ;; (<op> <x1> ...) -> (##core#inline <[u]fixop> <x1> (##core#inline <[u]fixop> ...)) [fixnum-mode (perhaps unsafe)]
+    ;; - Remove "<id>" from arguments.
+    ((21) ; classargs = (<id> <fixop> <ufixop> <genop> <words>)
+     (and inline-substitutions-enabled
+	  (or (test name 'standard-binding) (test name 'extended-binding))
+	  (let* ([id (first classargs)]
+		 [words (fifth classargs)]
+		 [genop (fourth classargs)]
+		 [fixop (if unsafe (third classargs) (second classargs))]
+		 [callargs 
+		  (remove
+		   (lambda (x)
+		     (and (eq? 'quote (node-class x))
+			  (eq? id (first (node-parameters x))) ) ) 
+		   callargs) ] )
+	    (cond [(null? callargs) (make-node '##core#call '(#t) (list cont (qnode id)))]
+		  [(null? (cdr callargs))
+		   (make-node '##core#call '(#t) (list cont (first callargs))) ]
+		  [else
+		   (make-node
+		    '##core#call '(#t)
+		    (list
+		     cont
+		     (fold-inner
+		      (lambda (x y)
+			(if (eq? number-type 'fixnum)
+			    (make-node '##core#inline (list fixop) (list x y))
+			    (make-node '##core#inline_allocate (list genop words) (list x y)) ) )
+		      callargs) ) ) ] ) ) ) )
+
+    ;; (<alloc-op> ...) -> (##core#inline_allocate (<aiop> <words>) ...)
+    ;; (<alloc-op> ...) -> (##core#inline <fxop> ...) [fixnum mode]
+    ((22) ; classargs = (<argc> <aiop> <safe> <words> <fxop>)
+     (let ([argc (first classargs)]
+	   [rargc (length callargs)]
+	   [w (fourth classargs)] )
+       (and inline-substitutions-enabled
+	    (= rargc argc)
+	    (or (test name 'extended-binding) (test name 'standard-binding))
+	    (or (third classargs) unsafe)
+	    (make-node
+	     '##core#call '(#t)
+	     (list cont 
+		   (if (eq? number-type 'fixnum)
+		       (make-node
+			'##core#inline
+			(list (fifth classargs))
+			callargs)
+		       (make-node
+			'##core#inline_allocate
+			(list (second classargs) w)
+			callargs) ) ) ) ) ) )
+
     (else (bomb "bad type (optimize)")) ) )
 
 
