@@ -1,6 +1,6 @@
 ;;;; partition.scm - The CHICKEN Scheme compiler (partitioning)
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without
@@ -169,7 +169,7 @@
 ;; Construct a bucket$.  GAINPROC is a one-arg procedure that
 ;; calculates the gain of a bucket item.
 (define (make-bucket$ GAINPROC KEYPROC)
-  (make-bucket$$ (make-hash-table) GAINPROC KEYPROC))
+  (make-bucket$$ (make-hash-table eq?) GAINPROC KEYPROC))
 
 ;; Gets the current gain of a bucket-item$
 (define (bucket$-gain bucketitem)
@@ -188,7 +188,7 @@
 (define (bucket$-add! bucket item)
   (let* [(db (bucket$$-db bucket))
          (gain ((bucket$$-gainproc bucket) item))
-         (dl (hash-table-ref db gain))]
+         (dl (hash-table-ref/default db gain #f))]
     (cond
       [(not dl)
        (set! dl (make-dlist$))
@@ -215,12 +215,12 @@
 (define (bucket$-remove-largest! bucket)
   (let [(n #f)
         (db (bucket$$-db bucket))]
-    (hash-table-for-each
+    (hash-table-walk
+     db
       (lambda (gain dlist)
         (and (or (not n) (> gain n)) ;; is larger or not set
              (not (dlist$-null? dlist)) ;; is not empty dlist
-             (set! n gain))) ;; this is now largest
-      db)
+             (set! n gain))) );; this is now largest
     (or n (error "No items remain in bucket to remove-largest!"))
     (cons (dlist$-first! (hash-table-ref db n)) n)))
   
@@ -259,13 +259,13 @@
 
 ;; Get a cell from graph
 (define (graph$-get graph id)
-  (hash-table-ref (graph$$-cells graph) id))
+  (hash-table-ref/default (graph$$-cells graph) id #f))
 
 ;; List operations
 (define (graph$-length graph)
   (graph$$-num-cells graph))
 (define (graph$->list graph)
-  (map cdr (hash-table->list (graph$$-cells graph))))
+  (map cdr (hash-table->alist (graph$$-cells graph))))
 (define (graph$-for-each proc graph)
   (for-each
     proc
