@@ -51,7 +51,7 @@
 (register-feature! 'regex 'pregexp)
 
 
-(define-constant *pregexp-comment-char* #\;)
+(define-constant *pregexp-comment-char* #\#)
 (define-constant *pregexp-return-char* #\return)
 (define-constant *pregexp-tab-char* #\tab)
 
@@ -628,11 +628,129 @@
                               r)))))
                 (loop (+ i 1) (string-append r (string c)))))))))
 
-(define pregexp
+(define (extract-bit-field size position n)
+  (fxand (fxnot (arithmetic-shift -1 size))
+	 (arithmetic-shift n (- position))))
+
+(define utf8-start-byte->length
+  (let ((table '#(
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; 0x
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; 1x
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; 2x
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; 3x
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; 4x
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; 5x
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; 6x
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; 7x
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; 8x
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; 9x
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; ax
+1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ; bx
+2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 ; cx
+2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 ; dx
+3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 ; ex
+4 4 4 4 4 4 4 4 5 5 5 5 6 6 0 0 ; fx
+)))
+    (lambda (i) (vector-ref table i))))
+
+(define (string-ref-at-byte s byte)
+  (let* ((c (string-ref s byte))
+         (ci (char->integer c))
+         (len (utf8-start-byte->length ci)))
+    (if (<= len 1)
+      c
+      (let ((end (+ byte len)))
+        (if (> end (string-length s))
+          (##sys#error 'regexp "utf8 trailing char overflow" s byte)
+          (let loop ((i (+ byte 1)) (res (extract-bit-field (- 7 len) 0 c)))
+            (if (= i end)
+              (integer->char res)
+              (loop (+ i 1)
+                    (fxior
+                     (fxshl res 6)
+                     (fxand #b00111111
+			    (char->integer (string-ref s i))))))))))))
+
+(define-constant pat-utf8
+  (let ((pat-ascii "[\x01-\x7F]")
+        (pat-utf8-2 "[\xC2-\xDF][\x80-\xC1]")
+        (pat-utf8-3 "[\xE0-\xEF][\x80-\xC1]{2}"))
+    (reverse
+     (string->list
+      (string-append "(?:" pat-ascii "|" pat-utf8-2 "|" pat-utf8-3 ")")))))
+
+(define (utf8-pattern->byte-pattern s)
+  (let ((size (string-length s)))
+    (define (scan i res)
+      (if (= i size)
+        (list->string (reverse res))
+        (let ((c (string-ref s i)))
+          (case c
+            ((#\.) ; XXXX optimize the .+ cases
+             (if (and (< (+ i 1) size) (eqv? #\* (string-ref s (+ i 1))))
+               (scan (+ i 2) (cons #\* (cons #\. res)))
+               (scan (+ i 1) (append pat-utf8 res))))
+            ((#\\)
+             (scan (+ i 2) (cons (string-ref s (+ i 1)) (cons #\\ res))))
+            ((#\[) (class (+ i 1) '() #t res))
+            (else (scan (+ i 1) (cons c res)))))))
+    (define (class->group ls)
+      (let loop ((ls ls) (acc '()))
+        (if (null? ls)
+          (string->list
+           (string-append "(?:" (string-intersperse acc "|") ")"))
+          (let ((c (car ls)) (rst (cdr ls)))
+            (if (and (pair? rst) (eqv? #\- (car rst)))
+              (let ((end (cadr rst)))
+                (if (or (> (char->integer c) 128)
+                        (> (char->integer end) 128))
+                  (##sys#error 'regexp "full unicode classes not supported" s)
+                  (loop (cddr rst)
+                        (cons (string-append "[" (char->string end)
+                                             "-" (char->string c) "]")
+                              acc))))
+              (loop (cdr ls) (cons (char->string c) acc)))))))
+    (define (class i acc ascii? res)
+      (if (= i size)
+        (##sys#error 'regexp "incomplete character class" s)
+        (let ((c (string-ref s i)))
+          (case c
+            ((#\])
+             (if ascii?
+               (scan (+ i 1) (cons #\] (append acc (cons #\[ res))))
+               (scan (+ i 1) (append (reverse (class->group acc)) res))))
+            ((#\\)
+             (let ((next (string-ref s (+ i 1))))
+               (if (< (char->integer next) 128)
+                 (class (+ i 2) (cons next (cons #\\ acc)) ascii? res)
+                 (class (+ i 1) (cons #\\ acc) ascii? res))))
+            (else
+             (if (< (char->integer c) 128)
+               (class (+ i 1) (cons c acc) ascii? res)
+               (class (+ i (utf8-start-byte->length (char->integer c)))
+                      (cons (string-ref-at-byte s i) acc)
+                      #f res)))))))
+    (scan 0 '())))
+
+(define %pregexp
   (lambda (s)
     (set! *pregexp-space-sensitive?* #t) ;in case it got corrupted
     (list ':sub (car (pregexp-read-pattern s 0 (string-length s))))))
 
+(define (pregexp s . o)
+  (%pregexp
+   (if (null? o)
+     s
+     (let ((s1 (if (car o) (string-append "(?i:" s ")") s))
+           (o1 (cdr o)))
+       (if (null? o1)
+         s1
+         (let ((s2 (if (car o1) (string-append "(?x:" s1 ")") s1))
+               (o2 (cdr o1)))
+           (if (and (pair? o2) (car o2))
+             (utf8-pattern->byte-pattern s2)
+             s2)))))))
+
 (define regexp pregexp)
 (define (regexp? x) (and (pair? x) (eq? ':sub (##sys#slot x 0))))
 
@@ -738,174 +856,4 @@
 	(and r (map (lambda (p) (list (car p) (cdr p))) r) ) ) ) ) )
 
 
-;;; Split string into fields:
-
-(define string-split-fields
-  (let ([reverse reverse]
-	[substring substring]
-	[string-search-positions string-search-positions] )
-    (lambda (regexp str . mode-and-start)
-      (##sys#check-string str 'string-split-fields)
-      (let* ([argc (length mode-and-start)]
-	     [len (##sys#size str)]
-	     [mode (if (fx> argc 0) (car mode-and-start) #t)]
-	     [start (if (fx> argc 1) (cadr mode-and-start) 0)] 
-	     [fini (case mode
-		     [(#:suffix)
-		      (lambda (ms start)
-			(if (fx< start len)
-			    (##sys#error 'string-split-fields "record does not end with suffix" str regexp)
-			    (reverse ms) ) ) ]
-		     [(#:infix)
-		      (lambda (ms start)
-			(if (fx>= start len)
-			    (reverse ms)
-			    (reverse (cons (substring str start len) ms)) ) ) ]
-		     [else (lambda (ms start) (reverse ms)) ] ) ]
-	     [fetch (case mode
-		      [(#:infix #:suffix) (lambda (start from to) (substring str start from))]
-		      [else (lambda (start from to) (substring str from to))] ) ] )
-	(let loop ([ms '()] [start start])
-	  (let ([m (string-search-positions regexp str start)])
-	    (if m
-		(let* ([mp (##sys#slot m 0)]
-		       [from (##sys#slot mp 0)]
-		       [to (cadr mp)] )
-		  (if (fx= from to)
-		      (if (fx= to len)
-			  (fini ms start)
-			  (loop (cons (fetch start (fx+ from 1) (fx+ to 2)) ms) (fx+ to 1)) )
-		      (loop (cons (fetch start from to) ms) to) ) )
-		(fini ms start) ) ) ) ) ) ) )
-
-
-;;; Substitute matching strings:
-
-(define string-substitute
-  (let ([substring substring]
-	[reverse reverse]
-	[make-string make-string]
-	[string-search-positions string-search-positions] )
-    (lambda (regex subst string . flag)
-      (##sys#check-string subst 'string-substitute)
-      (let* ([which (if (pair? flag) (car flag) 1)]
-	     [substlen (##sys#size subst)]
-	     [substlen-1 (fx- substlen 1)]
-	     [result '()] 
-	     [total 0] )
-
-	(define (push x) 
-	  (set! result (cons x result))
-	  (set! total (fx+ total (##sys#size x))) )
-	
-	(define (substitute matches)
-	  (let loop ([start 0] [index 0])
-	    (if (fx>= index substlen-1)
-		(push (if (fx= start 0) subst (substring subst start substlen)))
-		(let ([c (##core#inline "C_subchar" subst index)]
-		      [index+1 (fx+ index 1)] )
-		  (if (char=? c #\\) 
-		      (let ([c2 (##core#inline "C_subchar" subst index+1)])
-			(if (not (char=? #\\ c2))
-			    (let ([mi (list-ref matches (fx- (char->integer c2) 48))])
-			      (push (substring subst start index))
-			      (push (substring string (car mi) (cadr mi))) 
-			      (loop (fx+ index 2) index+1) )
-			    (loop start (fx+ index+1 1)) ) )
-		      (loop start index+1) ) ) ) ) )
-
-	(define (concatenate strs)
-	  (let ([str (make-string total)])
-	    (let loop ([ss strs] [index 0])
-	      (if (null? ss) 
-		  str
-		  (let* ([si (car ss)]
-			 [len (##sys#size si)] )
-		    (##core#inline "C_substring_copy" si str 0 len index)
-		    (loop (cdr ss) (fx+ index len)) ) ) ) ) )
-
-	(let loop ([index 0] [count 1])
-	  (let ([matches (string-search-positions regex string index)])
-	    (cond [matches
-		   (let* ([range (car matches)]
-			  [upto (cadr range)] )
-		     (cond [(or (not (fixnum? which)) (fx= count which))
-			    (push (substring string index (car range)))
-			    (substitute matches)
-			    (loop upto #f) ]
-			   [else
-			    (push (substring string index upto))
-			    (loop upto (fx+ count 1)) ] ) ) ]
-		  [else
-		   (push (substring string index (##sys#size string)))
-		   (concatenate (reverse result)) ] ) ) ) ) ) ) )
-
-(define string-substitute*
-  (let ([string-search-positions string-search-positions])
-    (lambda (str smap)
-      (##sys#check-string str 'string-substitute*)
-      (##sys#check-list smap 'string-substitute*)
-      (let ([len (##sys#size str)])
-	(define (collect i from total fs)
-	  (if (fx>= i len)
-	      (##sys#fragments->string
-	       total
-	       (reverse 
-		(if (fx> i from) 
-		    (cons (##sys#substring str from i) fs)
-		    fs) ) )
-	      (let loop ([smap smap] [pos len])
-		(if (null? smap)
-		    (collect pos from (fx+ total (fx- pos i)) fs)
-		    (let* ([p (car smap)]
-			   [sm (car p)]
-			   [st (cdr p)] 
-			   [m (string-search-positions sm str i)] 
-			   [ma (and m (##sys#slot m 0))] )
-		      (if (and ma (fx= i (##sys#slot ma 0)))
-			  (let ([i2 (##sys#slot (##sys#slot ma 1) 0)])
-			    (when (fx> i from)
-			      (set! fs 
-				(cons (##sys#substring str from i) fs)) ) 
-			    (collect 
-			     i2 i2
-			     (fx+ total (string-length st))
-			     (cons st fs) ) ) 
-			  (loop (cdr smap) 
-				(if ma 
-				    (fxmin pos (##sys#slot ma 0)) 
-				    pos) ) ) ) ) ) ) ) 
-	(collect 0 0 0 '()) ) ) ) )
-
-
-;;; Some useful things:
-
-(define pattern->regexp
-  (let ([list->string list->string]
-	[string->list string->list] )
-    (lambda (s)
-      (##sys#check-string s 'pattern->regexp)
-      (list->string
-       (let loop ([cs (string->list s)])
-	 (if (null? cs)
-	     '()
-	     (let ([c (car cs)]
-		   [rest (cdr cs)] )
-	       (cond [(char=? c #\*) `(#\. #\* ,@(loop rest))]
-		     [(char=? c #\?) (cons '#\. (loop rest))]
-		     [(or (char-alphabetic? c) (char-numeric? c)) (cons c (loop rest))]
-		     [else `(#\\ ,c ,@(loop rest))] ) ) ) ) ) ) ) )
-
-(define grep
-  (let ([string-match string-match])
-    (lambda (rx lst)
-      (##sys#check-list lst 'grep)
-      (let loop ([lst lst])
-	(if (null? lst)
-	    '()
-	    (let ([x (car lst)]
-		  [r (cdr lst)] )
-	      (if (string-match rx x)
-		  (cons x (loop r))
-		  (loop r) ) ) ) ) ) ) )
-
+(include "regex-common.scm")
