@@ -1,6 +1,6 @@
 ;;;; profiler.scm - Support code for profiling applications
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -54,6 +54,7 @@
 
 (define ##sys#profile-vector-list '())
 (define ##sys#profile-name #f)
+(define ##sys#profile-append-mode #f)
 
 
 ;;; Initialize profile counter vector:
@@ -111,19 +112,23 @@
 	[write-char write-char]
 	[write write] )
     (lambda ()
-      (with-output-to-file ##sys#profile-name
-	(lambda () 
-	  (for-each
-	   (lambda (vec)
-	     (let ([len (##sys#size vec)])
-	       (do ([i 0 (+ i profile-info-entry-size)])
-		   ((>= i len))
-		 (write-char #\()
-		 (write (##sys#slot vec i))
-		 (write-char #\space)
-		 (write (##sys#slot vec (add1 i)))
-		 (write-char #\space)
-		 (write (##sys#slot vec (+ i 3)))
-		 (write-char #\))
-		 (write-char #\newline) ) ) ) 
-	   ##sys#profile-vector-list) ) ) ) ) )
+      (apply
+       with-output-to-file ##sys#profile-name
+       (lambda () 
+	 (for-each
+	  (lambda (vec)
+	    (let ([len (##sys#size vec)])
+	      (do ([i 0 (+ i profile-info-entry-size)])
+		  ((>= i len))
+		(write-char #\()
+		(write (##sys#slot vec i))
+		(write-char #\space)
+		(write (##sys#slot vec (add1 i)))
+		(write-char #\space)
+		(write (##sys#slot vec (+ i 3)))
+		(write-char #\))
+		(write-char #\newline) ) ) ) 
+	  ##sys#profile-vector-list) )
+       (if ##sys#profile-append-mode
+	   '(append:)
+	   '() ) ) ) ) )
