@@ -1,6 +1,6 @@
 ; scheduler.scm - Basic scheduler for multithreading
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -40,13 +40,26 @@
   (usual-integrations)
   (hide ##sys#ready-queue-head ##sys#ready-queue-tail ##sys#timeout-list
 	##sys#update-thread-state-buffer ##sys#restore-thread-state-buffer
-	##sys#remove-from-ready-queue ##sys#unblock-threads-for-i/o
+	##sys#remove-from-ready-queue ##sys#unblock-threads-for-i/o ##sys#force-primordial
 	##sys#fdset-input-set ##sys#fdset-output-set ##sys#fdset-clear
 	##sys#fdset-select-timeout ##sys#fdset-restore) 
   (foreign-declare #<<EOF
+#ifdef HAVE_ERRNO_H
+# include <errno.h>
+extern int errno;
+# define C_signal_interrupted_p     C_mk_bool(errno == EINTR)
+#else
+# define C_signal_interrupted_p     C_SCHEME_FALSE
+#endif
+
 #ifdef _WIN32
+# if _MSC_VER > 1300
 # include <winsock2.h>
-# define C_msleep(n)     (Sleep(C_unfix(n)), C_SCHEME_UNDEFINED)
+# include <ws2tcpip.h>
+# else
+# include <winsock.h>
+# endif
+# define C_msleep(n)     (Sleep(C_unfix(n)), C_SCHEME_TRUE)
 #else
 # include <unistd.h>
 # include <sys/types.h>
@@ -55,15 +68,16 @@
 static C_word C_msleep(C_word ms);
 C_word C_msleep(C_word ms) {
 #ifdef __CYGWIN__
-  usleep(ms * 1000);
+  if(usleep(C_unfix(ms) * 1000) == -1) return C_SCHEME_FALSE;
 #else
   struct timespec ts;
   unsigned long mss = C_unfix(ms);
   ts.tv_sec = mss / 1000;
   ts.tv_nsec = (mss % 1000) * 1000000;
-  nanosleep(&ts, NULL);
+  
+  if(nanosleep(&ts, NULL) == -1) return C_SCHEME_FALSE;
 #endif
-  return C_SCHEME_UNDEFINED;
+  return C_SCHEME_TRUE;
 }
 #endif
 static fd_set C_fdset_input, C_fdset_output, C_fdset_input_2, C_fdset_output_2;
@@ -92,6 +106,7 @@ EOF
     (##core#inline "C_set_initial_timer_interrupt_period" (##sys#slot thread 9))
     ((##sys#slot thread 1)) )
   (let* ([ct ##sys#current-thread]
+	 [eintr #f]
 	 [cts (##sys#slot ct 3)] )
     (dbg "scheduling, current: " ct ", ready: " ##sys#ready-queue-head)
     (##sys#update-thread-state-buffer ct)
@@ -124,11 +139,16 @@ EOF
 				       (null? ##sys#fd-list) 
 				       (pair? ##sys#timeout-list))
 			      (let ([tmo1 (caar ##sys#timeout-list)])
-				(##core#inline "C_msleep" (fxmax 0 (- tmo1 now))) ) ) ) )
+				(set! eintr
+				  (and (not (##core#inline "C_msleep" (fxmax 0 (- tmo1 now))))
+				       (foreign-value "C_signal_interrupted_p" bool) ) ) ) ) ) )
 		      (loop (cdr lst)) ) ) ) ) ) )
       ;; Unblock threads blocked by I/O:
-      (unless (null? ##sys#fd-list)
-	(##sys#unblock-threads-for-i/o) )
+      (if eintr
+	  (##sys#force-primordial)
+	  (begin
+	    (unless (null? ##sys#fd-list)
+	      (##sys#unblock-threads-for-i/o) ) ) )
       ;; Fetch and activate next ready thread:
       (let loop2 ()
 	(let ([nt (##sys#remove-from-ready-queue)])
@@ -139,6 +159,10 @@ EOF
 		[(eq? (##sys#slot nt 3) 'ready) (switch nt)]
 		[else (loop2)] ) ) ) ) ) )
 
+(define (##sys#force-primordial)
+  (dbg "primordial thread forced due to interrupt")
+  (##sys#thread-unblock! ##sys#primordial-thread) )
+
 (define ##sys#ready-queue-head '())
 (define ##sys#ready-queue-tail '())
 
@@ -228,7 +252,6 @@ EOF
 
 (define (##sys#thread-basic-unblock! t)
   (dbg "unblocking: " t)
-  (##sys#setislot t 3 'ready)
   (##sys#setislot t 11 #f)
   (##sys#setislot t 4 #f)
   (##sys#add-to-ready-queue t) )
@@ -236,17 +259,28 @@ EOF
 (define ##sys#default-exception-handler
   (let ([print-error-message print-error-message]
 	[display display]
+	[print-call-chain print-call-chain]
 	[open-output-string open-output-string]
 	[get-output-string get-output-string] )
     (lambda (arg)
       (let ([ct ##sys#current-thread])
 	(dbg "exception: " ct " -> " (if (##sys#structure? arg 'condition) (##sys#slot arg 2) arg))
-	(when ##sys#warnings-enabled
-	  (let ([o (open-output-string)])
-	    (display "Warning (" o)
-	    (display ct o)
-	    (display "): " o)
-	    (print-error-message arg ##sys#standard-error (get-output-string o)) ) )
+	(cond [(foreign-value "C_abort_on_thread_exceptions" bool)
+	       (let* ([pt ##sys#primordial-thread]
+		      [ptx (##sys#slot pt 1)] )
+		 (##sys#setslot 
+		  pt 1 
+		  (lambda ()
+		    (##sys#signal arg)
+		    (ptx) ) )
+		 (##sys#thread-unblock! pt) ) ]
+	      [##sys#warnings-enabled
+	       (let ([o (open-output-string)])
+		 (display "Warning (" o)
+		 (display ct o)
+		 (display "): " o)
+		 (print-error-message arg ##sys#standard-error (get-output-string o))
+		 (print-call-chain ##sys#standard-error 0 ct) ) ] )
 	(##sys#setslot ct 7 arg)
 	(##sys#thread-kill! ct 'terminated)
 	(##sys#schedule) ) ) ) )
@@ -306,36 +340,37 @@ EOF
   (dbg "fd-list: " ##sys#fd-list)
   (let* ([to? (pair? ##sys#timeout-list)]
 	 [rq? (pair? ##sys#ready-queue-head)]
-	 [n (##sys#fdset-select-timeout		; we use FD_SETSIZE, but really should use max fd
+	 [n (##sys#fdset-select-timeout	; we use FD_SETSIZE, but really should use max fd
 	     (or rq? to?)
 	     (if (and to? (not rq?))
 		 (let* ([tmo1 (caar ##sys#timeout-list)]
 			[now (##sys#fudge 16)])
 		   (fxmax 0 (- tmo1 now)) )
-		 0) ) ] )                    ; otherwise immediate timeout.
+		 0) ) ] )		; otherwise immediate timeout.
     (dbg n " fds ready")
-    (when (fx> n 0)
-      (set! ##sys#fd-list
-	(let loop ([n n] [lst ##sys#fd-list])
-	  (if (zero? n)
-	      lst
-	      (let* ([a (car lst)]
-		     [fd (car a)]
-		     [inf (##core#inline "C_fd_test_input" fd)]
-		     [outf (##core#inline "C_fd_test_output" fd)] )
-		(dbg "fd " fd " ready: input=" inf ", output=" outf)
-		(if (or inf outf)
-		    (let loop2 ([threads (cdr a)])
-		      (if (null? threads) 
-			  (begin
-			    (##sys#fdset-clear fd)
-			    (loop (sub1 n) (cdr lst)) )
-			  (let* ([t (car threads)]
-				 [p (##sys#slot t 11)] )
-			    (when (and (pair? p) (eq? fd (car p)))
-			      (##sys#thread-basic-unblock! t) )
-			    (loop2 (cdr threads)) ) ) )
-		    (cons a (loop n (cdr lst))) ) ) ) ) ) )
+    (cond [(eq? -1 n) (##sys#force-primordial)]
+	  [(fx> n 0)
+	   (set! ##sys#fd-list
+	     (let loop ([n n] [lst ##sys#fd-list])
+	       (if (zero? n)
+		   lst
+		   (let* ([a (car lst)]
+			  [fd (car a)]
+			  [inf (##core#inline "C_fd_test_input" fd)]
+			  [outf (##core#inline "C_fd_test_output" fd)] )
+		     (dbg "fd " fd " ready: input=" inf ", output=" outf)
+		     (if (or inf outf)
+			 (let loop2 ([threads (cdr a)])
+			   (if (null? threads) 
+			       (begin
+				 (##sys#fdset-clear fd)
+				 (loop (sub1 n) (cdr lst)) )
+			       (let* ([t (car threads)]
+				      [p (##sys#slot t 11)] )
+				 (when (and (pair? p) (eq? fd (car p)))
+				   (##sys#thread-basic-unblock! t) )
+				 (loop2 (cdr threads)) ) ) )
+			 (cons a (loop n (cdr lst))) ) ) ) ) ) ] )
     (##sys#fdset-restore) ) )
 
 
@@ -374,8 +409,12 @@ EOF
     (set! ##sys#timeout-list (##sys#delq t ##sys#timeout-list))
     (set! ##sys#fd-list 
       (let loop ([fdl ##sys#fd-list])
-	(cond [(null? fdl)]
-	      [(eq? t (##sys#slot (##sys#slot fdl 0) 1)) (##sys#slot fdl 1)]
-	      [else (cons (##sys#slot fdl 0) (loop (##sys#slot fdl 1)))] ) ) )
+	(if (null? fdl)
+	    '()
+	    (let ([a (##sys#slot fdl 0)])
+	      (cons
+	       (cons (##sys#slot a 0)
+		     (##sys#delq t (##sys#slot a 1)) )
+	       (loop (##sys#slot fdl 1)) ) ) ) ) )
     (##sys#setislot t 12 '())
     (##sys#thread-basic-unblock! t) ) )
