@@ -54,18 +54,9 @@
 
 (register-feature! 'srfi-13)
 
-(include "srfi-13-syntax")
 
-
-(cond-expand [hygienic-macros
-(define char-cased? char-alphabetic?)
-(define char-titlecase char-upcase)
-] [else
 (define-inline (char-cased? c) (char-alphabetic? c))
 (define-inline (char-titlecase c) (char-upcase c))
-] )
-
-(eval-when (eval) (include "srfi-14"))
 
 
 ;;; SRFI 13 string library reference implementation		-*- Scheme -*-
@@ -185,20 +176,6 @@
 ;;; Support for START/END substring specs
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-(cond-expand [hygienic-macros
-
-;;; This one parses out a *pair* of final start/end indices. 
-;;; Not exported; for internal use.
-(define-syntax let-string-start+end2
-  (syntax-rules ()
-    ((l-s-s+e2 (start1 end1 start2 end2) proc s1 s2 args body ...)
-     (let ((procv proc)) ; Make sure PROC is only evaluated once.
-       (let-string-start+end (start1 end1 rest) procv s1 args
-         (let-string-start+end (start2 end2) procv s2 rest
-           body ...))))))
-
-] [else
-
 (eval-when (compile eval)
   (define-macro (let-string-start+end2 s-e proc s1 s2 args . body)
     (let ([procv (gensym)]
@@ -210,9 +187,6 @@
 	   ,(cddr s-e) ,procv ,s2 ,rest
 	   ,@body) ) ) ) ) )
 
-] )
-
-
 
 ;;; Returns three values: rest start end
 
@@ -305,11 +279,11 @@
 ;;; overhead for END parameter.
 (define (%substring/shared s start end)
   (if (and (zero? start) (= end (string-length s))) s
-      (substring s start end)))
+      (##sys#substring s start end)))
 
 (define (string-copy s . maybe-start+end)
   (let-string-start+end (start end) string-copy s maybe-start+end
-    (substring s start end)))
+    (##sys#substring s start end)))
 
 ;This library uses the R5RS SUBSTRING, but doesn't export it.
 ;Here is a definition, just for completeness.
@@ -967,7 +941,7 @@
 		(if (>= i bound) (- i 1) (lp (+ i i))))))
     (let lp ((i start) (ans 0))
       (if (>= i end) (modulo ans bound)
-	  (lp (+ i 1) (bitwise-and mask (+ (* 37 ans) (iref s i))))))))
+	  (lp (+ i 1) (fxand mask (+ (* 37 ans) (iref s i))))))))
 
 (define (string-hash s . maybe-bound+start+end)
   (let-optionals* maybe-bound+start+end ((bound 4194304); (and (integer? bound)
@@ -1036,7 +1010,7 @@
 
 (define (string-titlecase s . maybe-start+end)
   (let-string-start+end (start end) string-titlecase! s maybe-start+end
-    (let ((ans (substring s start end)))
+    (let ((ans (##sys#substring s start end)))
       (%string-titlecase! ans 0 (- end start))
       ans)))
 
@@ -1163,7 +1137,7 @@
 					   (begin (string-set! temp i c)
 						  (+ i 1))))
 				     0 s start end)))
-	  (if (= ans-len slen) temp (substring temp 0 ans-len)))
+	  (if (= ans-len slen) temp (##sys#substring temp 0 ans-len)))
 
 	(let* ((cset (cond ((char-set? criteria) criteria)
 			   ((char? criteria) (char-set criteria))
@@ -1191,7 +1165,7 @@
 						  (+ i 1))
 					   i))
 				     0 s start end)))
-	  (if (= ans-len slen) temp (substring temp 0 ans-len)))
+	  (if (= ans-len slen) temp (##sys#substring temp 0 ans-len)))
 
 	(let* ((cset (cond ((char-set? criteria) criteria)
 			   ((char? criteria) (char-set criteria))
@@ -1473,7 +1447,7 @@
 (define (make-kmp-restart-vector pattern . maybe-c=+start+end)
   (let-optionals* maybe-c=+start+end
                   ((c= char=?) rest) ; (procedure? c=))
-     (receive (start end) (string-parse-start+end make-kmp-restart-vector pattern rest)
+     (receive (rest2 start end) (string-parse-start+end make-kmp-restart-vector pattern rest)
        (let* ((rvlen (- end start))
 	   (rv (make-vector rvlen -1)))
       (if (> rvlen 0)
@@ -1543,7 +1517,7 @@
   (let-optionals* c=+p-start+s-start+s-end
       ((c=      char=?) ; (procedure? c=))
        (p-start 0) rest) ; (and (integer? p-start) (exact? p-start) (<= 0 p-start)))
-    (receive (s-start s-end) (string-parse-start+end string-kmp-partial-search s rest)
+    (receive (rest2 s-start s-end) (string-parse-start+end string-kmp-partial-search s rest)
     ;; Enough prelude. Here's the actual code.
     (let ((patlen (vector-length rv)))
       (let lp ((si s-start)		; An index into S.
@@ -1715,7 +1689,7 @@
     ;; string starts.
     (let lp ((len 0) (nzlist #f) (lis string-list))
       (if (pair? lis)
-	  (let ((slen (string-length (car string-list))))
+	  (let ((slen (string-length (car lis))))
 	    (lp (+ len slen)
 		(if (or nzlist (zero? slen)) nzlist lis)
 		(cdr lis)))
@@ -1779,9 +1753,9 @@
 		   (cond ((string-skip-right s token-chars start tend-1) =>
 			  (lambda (tstart-1)
 			    (lp tstart-1
-				(cons (substring s (+ 1 tstart-1) tend)
+				(cons (##sys#substring s (+ 1 tstart-1) tend)
 				      ans))))
-			 (else (cons (substring s start tend) ans))))))
+			 (else (cons (##sys#substring s start tend) ans))))))
 	      (else ans))))))
 
 
@@ -1840,7 +1814,7 @@
 
 	    ;; Selected text falls entirely within one span.
 	    ((= (floor (/ from slen)) (floor (/ to slen)))
-	     (substring s (+ start (modulo from slen))
+	     (##sys#substring s (+ start (modulo from slen))
 			  (+ start (modulo to   slen))))
 
 	    ;; Selected text requires multiple spans.
