@@ -7,9 +7,10 @@
   (disable-interrupts)
   (standard-bindings)
   (extended-bindings) 
-  (hide %char-set:s/check %string-iter %char-set-diff+intersection! %char->latin %latin1->char
+  (hide %char-set:s/check %string-iter %char-set-diff+intersection! %char->latin1 %latin1->char
 	%ucs-range->char-set! %string->char-set! %list->char-set! %set-char-set! %char-set-unfold!
-	%char-set-algebra %char-set-cursor-next %char-set-filter! %set-char-set) )
+	%char-set-algebra %char-set-cursor-next %char-set-filter! %set-char-set c0 c1 %string-copy
+	%default-base) )
 
 (cond-expand 
  [paranoia]
@@ -36,13 +37,8 @@
 (register-feature! 'srfi-14)
 
 
-(cond-expand [hygienic-macros
-(define %latin1->char integer->char)
-(define %char->latin1 char->integer)
-] [else
 (define (%latin1->char n) (integer->char n))
 (define (%char->latin1 c) (char->integer c))
-] )
 
 
 ;;; SRFI-14 character-sets library				-*- Scheme -*-
@@ -150,7 +146,7 @@
 (define (%char-set:s/check cs proc)
   (let lp ((cs cs))
     (if (char-set? cs) (char-set:s cs)
-	(lp (##sys#error "Not a char-set" cs proc)))))
+	(lp (##sys#error proc "Not a char-set" cs)))))
 
 
 
@@ -158,19 +154,11 @@
 ;;; underlying string representation of char sets. They should be
 ;;; inlined if possible.
 
-(cond-expand [hygienic-macros
-(define (si=0? s i) (zero? (%char->latin1 (string-ref s i))))
-(define (si=1? s i) (not (si=0? s i)))
-(define (si s i) (%char->latin1 (string-ref s i)))
-(define (%set0! s i) (string-set! s i c0))
-(define (%set1! s i) (string-set! s i c1))
-] [else
 (define-inline (si=0? s i) (zero? (%char->latin1 (string-ref s i))))
 (define-inline (si=1? s i) (not (si=0? s i)))
 (define-inline (si s i) (%char->latin1 (string-ref s i)))
 (define-inline (%set0! s i) (string-set! s i c0))
 (define-inline (%set1! s i) (string-set! s i c1))
-] )
 
 (define c0 (%latin1->char 0))
 (define c1 (%latin1->char 1))
@@ -178,48 +166,41 @@
 ;;; These do various "s[i] := s[i] op val" operations -- see 
 ;;; %CHAR-SET-ALGEBRA. They are used to implement the various
 ;;; set-algebra procedures.
-(cond-expand [hygienic-macros
-(define (setv!   s i v) (string-set! s i (%latin1->char v))) ; SET to a Value.
-(define (%not!   s i v) (setv! s i (- 1 v)))
-(define (%and!   s i v) (if (zero? v) (%set0! s i)))
-(define (%or!    s i v) (if (not (zero? v)) (%set1! s i)))
-(define (%minus! s i v) (if (not (zero? v)) (%set0! s i)))
-(define (%xor!   s i v) (if (not (zero? v)) (setv! s i (- 1 (si s i)))))
-] [else
 (define-inline (setv!   s i v) (string-set! s i (%latin1->char v))) ; SET to a Value.
 (define-inline (%not!   s i v) (setv! s i (- 1 v)))
 (define-inline (%and!   s i v) (if (zero? v) (%set0! s i)))
 (define-inline (%or!    s i v) (if (not (zero? v)) (%set1! s i)))
 (define-inline (%minus! s i v) (if (not (zero? v)) (%set0! s i)))
 (define-inline (%xor!   s i v) (if (not (zero? v)) (setv! s i (- 1 (si s i)))))
-] )
 
 (define (char-set-copy cs)
-  (make-char-set (%string-copy (%char-set:s/check cs char-set-copy))))
+  (make-char-set (%string-copy (%char-set:s/check cs 'char-set-copy))))
 
-(define (char-set= rest)
-  (or (null? rest)
-      (let* ((cs1  (car rest))
-	     (rest (cdr rest))
-	     (s1 (%char-set:s/check cs1 char-set=)))
-	(let lp ((rest rest))
-	  (or (not (pair? rest))
-	      (and (string=? s1 (%char-set:s/check (car rest) char-set=))
-		   (lp (cdr rest))))))))
+(define char-set= 
+  (lambda rest
+    (or (null? rest)
+	(let* ((cs1  (car rest))
+	       (rest (cdr rest))
+	       (s1 (%char-set:s/check cs1 'char-set=)))
+	  (let lp ((rest rest))
+	    (or (not (pair? rest))
+		(and (string=? s1 (%char-set:s/check (car rest) 'char-set=))
+		     (lp (cdr rest)))))))))
 
-(define (char-set<= cs1 . rest)
-  (or (null? rest)
-      (let ((cs1  (car rest))
-	    (rest (cdr rest)))
-	(let lp ((s1 (%char-set:s/check cs1 char-set<=))  (rest rest))
-	  (or (not (pair? rest))
-	      (let ((s2 (%char-set:s/check (car rest) char-set<=))
-		    (rest (cdr rest)))
-		(if (eq? s1 s2) (lp s2 rest) ; Fast path
-		    (let lp2 ((i 255))	; Real test
-		      (if (< i 0) (lp s2 rest)
-			  (and (<= (si s1 i) (si s2 i))
-			       (lp2 (- i 1))))))))))) )
+(define char-set<=
+  (lambda rest
+    (or (null? rest)
+	(let ((cs1  (car rest))
+	      (rest (cdr rest)))
+	  (let lp ((s1 (%char-set:s/check cs1 'char-set<=))  (rest rest))
+	    (or (not (pair? rest))
+		(let ((s2 (%char-set:s/check (car rest) 'char-set<=))
+		      (rest (cdr rest)))
+		  (if (eq? s1 s2) (lp s2 rest) ; Fast path
+		      (let lp2 ((i 255))	; Real test
+			(if (< i 0) (lp s2 rest)
+			    (and (<= (si s1 i) (si s2 i))
+				 (lp2 (- i 1))))))))))) ))
 
 ;;; Hash
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -238,7 +219,7 @@
   (let ((bound (:optional maybe-bound 4194304)))
     (if (zero? bound) (set! bound 4194304))
     (##sys#check-exact bound 'char-set-hash)
-    (let* ((s (%char-set:s/check cs char-set-hash))
+    (let* ((s (%char-set:s/check cs 'char-set-hash))
 	   ;; Compute a 111...1 mask that will cover BOUND-1:
 	   (mask (let lp ((i #x10000))	; Let's skip first 16 iterations, eh?
 		   (if (>= i bound) (- i 1) (lp (+ i i))))))
@@ -246,23 +227,24 @@
 	(if (< i 0) (modulo ans bound)
 	    (lp (- i 1)
 		(if (si=0? s i) ans
-		    (bitwise-and mask (+ (* 37 ans) i)))))))) )
+		    (fxand mask (+ (* 37 ans) i)))))))) )
 
 
 (define (char-set-contains? cs char)
-  (si=1? (%char-set:s/check cs char-set-contains?)
+  (##sys#check-char char 'char-set-contains?)
+  (si=1? (%char-set:s/check cs 'char-set-contains?)
 ;	 (%char->latin1 (check-arg char? char char-set-contains?))))
 	 (%char->latin1 char) ) )
 
 (define (char-set-size cs)
-  (let ((s (%char-set:s/check cs char-set-size)))
+  (let ((s (%char-set:s/check cs 'char-set-size)))
     (let lp ((i 255) (size 0))
       (if (< i 0) size
 	  (lp (- i 1) (+ size (si s i)))))))
 
 (define (char-set-count pred cset)
 ;  (check-arg procedure? pred char-set-count)
-  (let ((s (%char-set:s/check cset char-set-count)))
+  (let ((s (%char-set:s/check cset 'char-set-count)))
     (let lp ((i 255) (count 0))
       (if (< i 0) count
 	  (lp (- i 1)
@@ -286,13 +268,13 @@
   cs)
 
 (define (char-set-adjoin cs . chars)
-  (%set-char-set  %set1! char-set-adjoin cs chars))
+  (%set-char-set  %set1! 'char-set-adjoin cs chars))
 (define (char-set-adjoin! cs . chars)
-  (%set-char-set! %set1! char-set-adjoin! cs chars))
+  (%set-char-set! %set1! 'char-set-adjoin! cs chars))
 (define (char-set-delete cs . chars)
-  (%set-char-set  %set0! char-set-delete cs chars))
+  (%set-char-set  %set0! 'char-set-delete cs chars))
 (define (char-set-delete! cs . chars)
-  (%set-char-set! %set0! char-set-delete! cs chars))
+  (%set-char-set! %set0! 'char-set-delete! cs chars))
 
 
 ;;; Cursors
@@ -306,16 +288,17 @@
 ;;; (But first mask out the bits already scanned by the cursor first.)
 
 (define (char-set-cursor cset)
-  (%char-set-cursor-next cset 256 char-set-cursor))
+  (%char-set-cursor-next cset 256 'char-set-cursor))
   
 (define (end-of-char-set? cursor) (< cursor 0))
 
 (define (char-set-ref cset cursor) (%latin1->char cursor))
 
 (define (char-set-cursor-next cset cursor)
+  (##sys#check-exact cursor 'char-set-cursor-next)
 ;  (check-arg (lambda (i) (and (integer? i) (exact? i) (<= 0 i 255))) cursor
 ;	     char-set-cursor-next)
-  (%char-set-cursor-next cset cursor char-set-cursor-next))
+  (%char-set-cursor-next cset cursor 'char-set-cursor-next))
 
 (define (%char-set-cursor-next cset cursor proc)	; Internal
   (let ((s (%char-set:s/check cset proc)))
@@ -329,7 +312,7 @@
 
 (define (char-set-for-each proc cs)
 ;  (check-arg procedure? proc char-set-for-each)
-  (let ((s (%char-set:s/check cs char-set-for-each)))
+  (let ((s (%char-set:s/check cs 'char-set-for-each)))
     (let lp ((i 255))
       (cond ((>= i 0)
 	     (if (si=1? s i) (proc (%latin1->char i)))
@@ -337,7 +320,7 @@
 
 (define (char-set-map proc cs)
 ;  (check-arg procedure? proc char-set-map)
-  (let ((s (%char-set:s/check cs char-set-map))
+  (let ((s (%char-set:s/check cs 'char-set-map))
 	(ans (make-string 256 c0)))
     (let lp ((i 255))
       (cond ((>= i 0)
@@ -348,7 +331,7 @@
 
 (define (char-set-fold kons knil cs)
 ;  (check-arg procedure? kons char-set-fold)
-  (let ((s (%char-set:s/check cs char-set-fold)))
+  (let ((s (%char-set:s/check cs 'char-set-fold)))
     (let lp ((i 255) (ans knil))
       (if (< i 0) ans
 	  (lp (- i 1)
@@ -357,7 +340,7 @@
 
 (define (char-set-every pred cs)
 ;  (check-arg procedure? pred char-set-every)
-  (let ((s (%char-set:s/check cs char-set-every)))
+  (let ((s (%char-set:s/check cs 'char-set-every)))
     (let lp ((i 255))
       (or (< i 0)
 	  (and (or (si=0? s i) (pred (%latin1->char i)))
@@ -365,7 +348,7 @@
 
 (define (char-set-any pred cs)
 ;  (check-arg procedure? pred char-set-any)
-  (let ((s (%char-set:s/check cs char-set-any)))
+  (let ((s (%char-set:s/check cs 'char-set-any)))
     (let lp ((i 255))
       (and (>= i 0)
 	   (or (and (si=1? s i) (pred (%latin1->char i)))
@@ -388,7 +371,7 @@
 
 (define (char-set-unfold! p f g seed base-cset)
   (%char-set-unfold! char-set-unfold! p f g
-		     (%char-set:s/check base-cset char-set-unfold!)
+		     (%char-set:s/check base-cset 'char-set-unfold!)
 		     seed)
   base-cset)
 
@@ -411,12 +394,12 @@
     (make-char-set bs)))
 
 (define (list->char-set! chars base-cs)
-  (%list->char-set! chars (%char-set:s/check base-cs list->char-set!))
+  (%list->char-set! chars (%char-set:s/check base-cs 'list->char-set!))
   base-cs)
 
 
 (define (char-set->list cs)
-  (let ((s (%char-set:s/check cs char-set->list)))
+  (let ((s (%char-set:s/check cs 'char-set->list)))
     (let lp ((i 255) (ans '()))
       (if (< i 0) ans
 	  (lp (- i 1)
@@ -428,6 +411,7 @@
 ;;; string <--> char-set
 
 (define (%string->char-set! str bs proc)
+  (##sys#check-string str proc)
 ;  (check-arg string? str proc)
   (do ((i (- (string-length str) 1) (- i 1)))
       ((< i 0))
@@ -435,17 +419,17 @@
 
 (define (string->char-set str . maybe-base)
   (let ((bs (%default-base maybe-base string->char-set)))
-    (%string->char-set! str bs string->char-set)
+    (%string->char-set! str bs 'string->char-set)
     (make-char-set bs)))
 
 (define (string->char-set! str base-cs)
-  (%string->char-set! str (%char-set:s/check base-cs string->char-set!)
-		      string->char-set!)
+  (%string->char-set! str (%char-set:s/check base-cs 'string->char-set!)
+		      'string->char-set!)
   base-cs)
 
 
 (define (char-set->string cs)
-  (let* ((s (%char-set:s/check cs char-set->string))
+  (let* ((s (%char-set:s/check cs 'char-set->string))
 	 (ans (make-string (char-set-size cs))))
     (let lp ((i 255) (j 0))
       (if (< i 0) ans
@@ -458,10 +442,10 @@
 ;;; -- UCS-range -> char-set
 
 (define (%ucs-range->char-set! lower upper error? bs proc)
+  (##sys#check-exact lower proc)
+  (##sys#check-exact upper proc)
 ;  (check-arg (lambda (x) (and (integer? x) (exact? x) (<= 0 x))) lower proc)
 ;  (check-arg (lambda (x) (and (integer? x) (exact? x) (<= lower x))) upper proc)
-  (##sys#check-exact lower)
-  (##sys#check-exact upper)
 
   (if (and (< lower upper) (< 256 upper) error?)
       (##sys#error "Requested UCS range contains unavailable characters -- this implementation only supports Latin-1"
@@ -473,13 +457,13 @@
 (define (ucs-range->char-set lower upper . rest)
   (let-optionals* rest ((error? #f) rest)
     (let ((bs (%default-base rest ucs-range->char-set)))
-      (%ucs-range->char-set! lower upper error? bs ucs-range->char-set)
+      (%ucs-range->char-set! lower upper error? bs 'ucs-range->char-set)
       (make-char-set bs))))
 
 (define (ucs-range->char-set! lower upper error? base-cs)
   (%ucs-range->char-set! lower upper error?
-			 (%char-set:s/check base-cs ucs-range->char-set!)
-			 ucs-range->char-set)
+			 (%char-set:s/check base-cs 'ucs-range->char-set!)
+			 'ucs-range->char-set)
   base-cs)
 
 
@@ -496,15 +480,15 @@
 (define (char-set-filter predicate domain . maybe-base)
   (let ((bs (%default-base maybe-base char-set-filter)))
     (%char-set-filter! predicate
-		       (%char-set:s/check domain char-set-filter!)
+		       (%char-set:s/check domain 'char-set-filter!)
 		       bs
 		       char-set-filter)
     (make-char-set bs)))
 
 (define (char-set-filter! predicate domain base-cs)
   (%char-set-filter! predicate
-		     (%char-set:s/check domain char-set-filter!)
-		     (%char-set:s/check base-cs char-set-filter!)
+		     (%char-set:s/check domain 'char-set-filter!)
+		     (%char-set:s/check base-cs 'char-set-filter!)
 		     char-set-filter!)
   base-cs)
 
@@ -557,13 +541,13 @@
 ;;; -- Invert
 
 (define (char-set-complement cs)
-  (let ((s (%char-set:s/check cs char-set-complement))
+  (let ((s (%char-set:s/check cs 'char-set-complement))
 	(ans (make-string 256)))
     (%string-iter (lambda (i v) (%not! ans i v)) s)
     (make-char-set ans)))
 
 (define (char-set-complement! cset)
-  (let ((s (%char-set:s/check cset char-set-complement!)))
+  (let ((s (%char-set:s/check cset 'char-set-complement!)))
     (%string-iter (lambda (i v) (%not! s i v)) s))
   cset)
 
@@ -571,14 +555,14 @@
 ;;; -- Union
 
 (define (char-set-union! cset1 . csets)
-  (%char-set-algebra (%char-set:s/check cset1 char-set-union!)
-		     csets %or! char-set-union!)
+  (%char-set-algebra (%char-set:s/check cset1 'char-set-union!)
+		     csets %or! 'char-set-union!)
   cset1)
 
 (define (char-set-union . csets)
   (if (pair? csets)
-      (let ((s (%string-copy (%char-set:s/check (car csets) char-set-union))))
-	(%char-set-algebra s (cdr csets) %or! char-set-union)
+      (let ((s (%string-copy (%char-set:s/check (car csets) 'char-set-union))))
+	(%char-set-algebra s (cdr csets) %or! 'char-set-union)
 	(make-char-set s))
       (char-set-copy char-set:empty)))
 
@@ -586,14 +570,14 @@
 ;;; -- Intersection
 
 (define (char-set-intersection! cset1 . csets)
-  (%char-set-algebra (%char-set:s/check cset1 char-set-intersection!)
-		     csets %and! char-set-intersection!)
+  (%char-set-algebra (%char-set:s/check cset1 'char-set-intersection!)
+		     csets %and! 'char-set-intersection!)
   cset1)
 
 (define (char-set-intersection . csets)
   (if (pair? csets)
-      (let ((s (%string-copy (%char-set:s/check (car csets) char-set-intersection))))
-	(%char-set-algebra s (cdr csets) %and! char-set-intersection)
+      (let ((s (%string-copy (%char-set:s/check (car csets) 'char-set-intersection))))
+	(%char-set-algebra s (cdr csets) %and! 'char-set-intersection)
 	(make-char-set s))
       (char-set-copy char-set:full)))
 
@@ -601,14 +585,14 @@
 ;;; -- Difference
 
 (define (char-set-difference! cset1 . csets)
-  (%char-set-algebra (%char-set:s/check cset1 char-set-difference!)
-		     csets %minus! char-set-difference!)
+  (%char-set-algebra (%char-set:s/check cset1 'char-set-difference!)
+		     csets %minus! 'char-set-difference!)
   cset1)
 
 (define (char-set-difference cs1 . csets)
   (if (pair? csets)
-      (let ((s (%string-copy (%char-set:s/check cs1 char-set-difference))))
-	(%char-set-algebra s csets %minus! char-set-difference)
+      (let ((s (%string-copy (%char-set:s/check cs1 'char-set-difference))))
+	(%char-set-algebra s csets %minus! 'char-set-difference)
 	(make-char-set s))
       (char-set-copy cs1)))
 
@@ -616,14 +600,14 @@
 ;;; -- Xor
 
 (define (char-set-xor! cset1 . csets)
-  (%char-set-algebra (%char-set:s/check cset1 char-set-xor!)
-		      csets %xor! char-set-xor!)
+  (%char-set-algebra (%char-set:s/check cset1 'char-set-xor!)
+		      csets %xor! 'char-set-xor!)
   cset1)
 
 (define (char-set-xor . csets)
   (if (pair? csets)
-      (let ((s (%string-copy (%char-set:s/check (car csets) char-set-xor))))
-	(%char-set-algebra s (cdr csets) %xor! char-set-xor)
+      (let ((s (%string-copy (%char-set:s/check (car csets) 'char-set-xor))))
+	(%char-set-algebra s (cdr csets) %xor! 'char-set-xor)
 	(make-char-set s))
       (char-set-copy char-set:empty)))
 
@@ -641,19 +625,19 @@
 	    csets))
 
 (define (char-set-diff+intersection! cs1 cs2 . csets)
-  (let ((s1 (%char-set:s/check cs1 char-set-diff+intersection!))
-	(s2 (%char-set:s/check cs2 char-set-diff+intersection!)))
+  (let ((s1 (%char-set:s/check cs1 'char-set-diff+intersection!))
+	(s2 (%char-set:s/check cs2 'char-set-diff+intersection!)))
     (%string-iter (lambda (i v) (if (zero? v)
 				    (%set0! s2 i)
 				    (if (si=1? s2 i) (%set0! s1 i))))
 		  s1)
-    (%char-set-diff+intersection! s1 s2 csets char-set-diff+intersection!))
+    (%char-set-diff+intersection! s1 s2 csets 'char-set-diff+intersection!))
   (values cs1 cs2))
 
 (define (char-set-diff+intersection cs1 . csets)
-  (let ((diff (string-copy (%char-set:s/check cs1 char-set-diff+intersection)))
+  (let ((diff (string-copy (%char-set:s/check cs1 'char-set-diff+intersection)))
 	(int  (make-string 256 c0)))
-    (%char-set-diff+intersection! diff int csets char-set-diff+intersection)
+    (%char-set-diff+intersection! diff int csets 'char-set-diff+intersection)
     (values (make-char-set diff) (make-char-set int))))
 
 
