@@ -1,6 +1,6 @@
 ;;;; srfi-18.scm - Simple thread unit - felix
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -117,8 +117,8 @@ EOF
 
 (define (seconds->time n)
   (##sys#check-number n 'seconds->time)
-  (let* ([n2 (- n C_startup_time_seconds)] ; seconds since startup
-	 [ms (truncate (* 1000 (##sys#flonum-fraction n)))] ; milliseconds
+  (let* ([n2 (max 0 (- n C_startup_time_seconds))] ; seconds since startup
+	 [ms (truncate (* 1000 (##sys#flonum-fraction (##sys#exact->inexact n))))] ; milliseconds
 	 [n3 (inexact->exact (truncate (+ (* n2 1000) ms)))] ) ; milliseconds since startup
     (##sys#make-structure 'time n3 (truncate n) (inexact->exact ms)) ) )
 
@@ -200,13 +200,16 @@ EOF
   (##sys#slot x 6) )
 
 (define thread-start!
-  (lambda (thread)
-    (##sys#check-structure thread 'thread 'thread-start!)
-    (unless (eq? 'created (##sys#slot thread 3))
-      (##sys#error 'thread-start! "thread can not be started a second time" thread) )
-    (##sys#setslot thread 3 'ready)
-    (##sys#add-to-ready-queue thread) 
-    thread) )
+  (let ([make-thread make-thread])
+    (lambda (thread)
+      (if (procedure? thread)
+	  (set! thread (make-thread thread))
+	  (##sys#check-structure thread 'thread 'thread-start!) )
+      (unless (eq? 'created (##sys#slot thread 3))
+	(##sys#error 'thread-start! "thread can not be started a second time" thread) )
+      (##sys#setslot thread 3 'ready)
+      (##sys#add-to-ready-queue thread) 
+      thread) ) )
 
 (define (thread-yield!)
   (##sys#call-with-current-continuation
@@ -447,7 +450,7 @@ EOF
 ;;; Change continuation of thread to signal an exception:
 
 (define (thread-deliver-signal! thread exn)
-  (##sys#check-structure thread 'thread 'thread-deliver-exception!)
+  (##sys#check-structure thread 'thread 'thread-deliver-signal!)
   (if (eq? thread ##sys#current-thread)
       (##sys#signal exn)
       (let ([old (##sys#slot thread 1)])
@@ -455,4 +458,18 @@ EOF
 	 thread 1
 	 (lambda ()
 	   (##sys#signal exn)
-	   (old) ) ) ) ) )
+	   (old) ) )
+	(##sys#thread-unblock! thread) ) ) )
+
+
+;;; Don't block in the repl: (by Chris Double)
+
+(unless (eq? (build-platform) 'msvc)
+  (set! ##sys#read-prompt-hook
+    (let ([old ##sys#read-prompt-hook]
+	  [thread-yield! thread-yield!] )
+      (lambda ()
+	(when (or (##sys#fudge 12) (##sys#tty-port? ##sys#standard-input))
+	  (old)
+	  (##sys#thread-block-for-i/o! ##sys#current-thread 0 #t)
+	  (thread-yield!)))) ) )
