@@ -1,6 +1,6 @@
 ;;;; srfi-4.scm - Homogeneous numeric vectors
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -46,6 +46,13 @@
 #define C_s8peek(b, i)         C_fix(((char *)C_data_pointer(b))[ C_unfix(i) ])
 #define C_u16peek(b, i)        C_fix(((unsigned short *)C_data_pointer(b))[ C_unfix(i) ])
 #define C_s16peek(b, i)        C_fix(((short *)C_data_pointer(b))[ C_unfix(i) ])
+#ifdef C_SIXTY_FOUR
+# define C_a_u32peek(ptr, d, b, i) C_fix(((C_u32 *)C_data_pointer(b))[ C_unfix(i) ])
+# define C_a_s32peek(ptr, d, b, i) C_fix(((C_s32 *)C_data_pointer(b))[ C_unfix(i) ])
+#else
+# define C_a_u32peek(ptr, d, b, i) C_unsigned_int_to_num(ptr, ((C_u32 *)C_data_pointer(b))[ C_unfix(i) ])
+# define C_a_s32peek(ptr, d, b, i) C_int_to_num(ptr, ((C_s32 *)C_data_pointer(b))[ C_unfix(i) ])
+#endif
 #define C_f32peek(b, i)        (C_temporary_flonum = ((float *)C_data_pointer(b))[ C_unfix(i) ], C_SCHEME_UNDEFINED)
 #define C_f64peek(b, i)        (C_temporary_flonum = ((double *)C_data_pointer(b))[ C_unfix(i) ], C_SCHEME_UNDEFINED)
 #define C_u8poke(b, i, x)      ((((unsigned char *)C_data_pointer(b))[ C_unfix(i) ] = C_unfix(x)), C_SCHEME_UNDEFINED)
@@ -116,8 +123,8 @@ EOF
 (define (##sys#s8vector-ref v i) (##core#inline "C_s8peek" (##core#inline "C_slot" v 1) i))
 (define (##sys#u16vector-ref v i) (##core#inline "C_u16peek" (##core#inline "C_slot" v 1) i))
 (define (##sys#s16vector-ref v i) (##core#inline "C_s16peek" (##core#inline "C_slot" v 1) i))
-(define (##sys#u32vector-ref v i) (##sys#peek-unsigned-integer (##core#inline "C_slot" v 1) i))
-(define (##sys#s32vector-ref v i) (##sys#peek-signed-integer (##core#inline "C_slot" v 1) i))
+(define (##sys#u32vector-ref v i) (##core#inline_allocate ("C_a_u32peek" 4) (##core#inline "C_slot" v 1) i))
+(define (##sys#s32vector-ref v i) (##core#inline_allocate ("C_a_s32peek" 4) (##core#inline "C_slot" v 1) i))
 
 (define (##sys#f32vector-ref v i)
   (##core#inline "C_f32peek" (##core#inline "C_slot" v 1) i)
@@ -212,15 +219,6 @@ EOF
 		     x
 		     (exact->inexact x) ) ) ) ) )
  
-  (set! u8vector-ref (get u8vector-length ##sys#u8vector-ref 'u8vector-ref))
-  (set! s8vector-ref (get s8vector-length ##sys#s8vector-ref 's8vector-ref))
-  (set! u16vector-ref (get u16vector-length ##sys#u16vector-ref 'u16vector-ref))
-  (set! s16vector-ref (get s16vector-length ##sys#s16vector-ref 's16vector-ref))
-  (set! u32vector-ref (get u32vector-length ##sys#u32vector-ref 'u32vector-ref))
-  (set! s32vector-ref (get s32vector-length ##sys#s32vector-ref 's32vector-ref))
-  (set! f32vector-ref (get f32vector-length ##sys#f32vector-ref 'f32vector-ref))
-  (set! f64vector-ref (get f64vector-length ##sys#f64vector-ref 'f64vector-ref)) 
-
   (set! u8vector-set! (setu u8vector-length ##sys#u8vector-set! 'u8vector-set!))
   (set! s8vector-set! (set s8vector-length ##sys#s8vector-set! 's8vector-set!))
   (set! u16vector-set! (setu u16vector-length ##sys#u16vector-set! 'u16vector-set!))
@@ -228,116 +226,175 @@ EOF
   (set! u32vector-set! (setuw u32vector-length ##sys#u32vector-set! 'u32vector-set!))
   (set! s32vector-set! (setw s32vector-length ##sys#s32vector-set! 's32vector-set!))
   (set! f32vector-set! (setf f32vector-length ##sys#f32vector-set! 'f32vector-set!))
-  (set! f64vector-set! (setf f64vector-length ##sys#f64vector-set! 'f64vector-set!)) )
+  (set! f64vector-set! (setf f64vector-length ##sys#f64vector-set! 'f64vector-set!))
+
+  (set! u8vector-ref
+	(getter-with-setter (get u8vector-length ##sys#u8vector-ref 'u8vector-ref)
+			    u8vector-set!) )
+  (set! s8vector-ref
+	(getter-with-setter (get s8vector-length ##sys#s8vector-ref 's8vector-ref)
+			    s8vector-set!) )
+  (set! u16vector-ref
+	(getter-with-setter (get u16vector-length ##sys#u16vector-ref 'u16vector-ref)
+			    u16vector-set!) )
+  (set! s16vector-ref
+	(getter-with-setter (get s16vector-length ##sys#s16vector-ref 's16vector-ref)
+			    s16vector-set!) )
+  (set! u32vector-ref
+	(getter-with-setter
+	 (get u32vector-length ##sys#u32vector-ref 'u32vector-ref)
+	 u32vector-set!) )
+  (set! s32vector-ref
+	(getter-with-setter 
+	 (get s32vector-length ##sys#s32vector-ref 's32vector-ref)
+	 s32vector-set!) )
+  (set! f32vector-ref
+	(getter-with-setter 
+	 (get f32vector-length ##sys#f32vector-ref 'f32vector-ref)
+	 f32vector-set!) )
+  (set! f64vector-ref 
+	(getter-with-setter
+	 (get f64vector-length ##sys#f64vector-ref 'f64vector-ref)
+	 f64vector-set!) ) )
+
 
 
 ;;; Basic constructors:
 
-(let ([alloc
-       (lambda (len)
-	 (let ([bv (##sys#allocate-vector len #t #f #t)])
-	   (##core#inline "C_string_to_bytevector" bv)
-	   bv) ) ] )
+(let* ([ext-alloc
+	(foreign-lambda* scheme-object ([int bytes])
+	  "C_word *buf = (C_word *)C_malloc(bytes + sizeof(C_header));"
+	  "if(buf == NULL) return(C_SCHEME_FALSE);"
+	  "C_block_header(buf) = C_make_header(C_BYTEVECTOR_TYPE, bytes);"
+	  "return(buf);") ]
+       [ext-free
+	(foreign-lambda* void ([scheme-object bv])
+	  "C_free((void *)C_block_item(bv, 1));") ]
+       [set-finalizer! set-finalizer!]
+       [alloc
+	(lambda (loc len ext?)
+	  (if ext? 
+	      (let ([bv (ext-alloc len)])
+		(or bv
+		    (##sys#error loc "not enough memory - can not allocate external number vector" len)) )
+	      (let ([bv (##sys#allocate-vector len #t #f #t)]) ; this could be made better...
+		(##core#inline "C_string_to_bytevector" bv) 
+		bv) ) ) ] )
+
+  (set! release-number-vector
+    (lambda (v)
+      (if (and (##sys#generic-structure? v)
+	       (memq (##sys#slot v 0) '(u8vector u16vector s8vector s16vector u32vector s32vector f32vector f64vector)) )
+	  (ext-free v)
+	  (##sys#error 'release-number-vector "bad argument type - not a number vector" v)) ) )
 
   (set! make-u8vector
-    (lambda (len . init)
+    (lambda (len #!optional (init #f)  (ext? #f) (fin? #t))
       (##sys#check-exact len 'make-u8vector)
-      (let ((v (##sys#make-structure 'u8vector (alloc len))))
-	(if (##core#inline "C_eqp" init '())
+      (let ((v (##sys#make-structure 'u8vector (alloc 'make-u8vector len ext?))))
+	(when (and ext? fin?) (set-finalizer! v ext-free))
+	(if (not init)
 	    v
-	    (let ((x (##core#inline "C_slot" init 0)))
-	      (##sys#check-exact-interval x 0 #xff 'mnake.u8vector)
+	    (begin
+	      (##sys#check-exact-interval init 0 #xff 'make-u8vector)
 	      (do ((i 0 (##core#inline "C_fixnum_plus" i 1)))
 		  ((##core#inline "C_fixnum_greater_or_equal_p" i len) v)
-		(##sys#u8vector-set! v i x) ) ) ) ) ) )
+		(##sys#u8vector-set! v i init) ) ) ) ) ) )
 
   (set! make-s8vector
-    (lambda (len . init)
+    (lambda (len #!optional (init #f)  (ext? #f) (fin #t))
       (##sys#check-exact len 'make-s8vector)
-      (let ((v (##sys#make-structure 's8vector (alloc len))))
-	(if (##core#inline "C_eqp" init '())
+      (let ((v (##sys#make-structure 's8vector (alloc 'make-s8vector len ext?))))
+	(when (and ext? fin?) (set-finalizer! v ext-free))
+	(if (not init)
 	    v
-	    (let ((x (##core#inline "C_slot" init 0)))
-	      (##sys#check-exact-interval x -128 127 'make-s8vector)
+	    (begin
+	      (##sys#check-exact-interval init -128 127 'make-s8vector)
 	      (do ((i 0 (##core#inline "C_fixnum_plus" i 1)))
 		  ((##core#inline "C_fixnum_greater_or_equal_p" i len) v)
-		(##sys#s8vector-set! v i x) ) ) ) ) ) )
+		(##sys#s8vector-set! v i init) ) ) ) ) ) )
 
   (set! make-u16vector
-    (lambda (len . init)
+    (lambda (len #!optional (init #f)  (ext? #f) (fin #t))
       (##sys#check-exact len 'make-u16vector)
-      (let ((v (##sys#make-structure 'u16vector (alloc (##core#inline "C_fixnum_shift_left" len 1)))))
-	(if (##core#inline "C_eqp" init '())
+      (let ((v (##sys#make-structure 'u16vector (alloc 'make-u16vector (##core#inline "C_fixnum_shift_left" len 1) ext?))))
+	(when (and ext? fin?) (set-finalizer! v ext-free))
+	(if (not init)
 	    v
-	    (let ((x (##core#inline "C_slot" init 0)))
-	      (##sys#check-exact-interval x 0 #xffff 'make-u16vector)
+	    (begin
+	      (##sys#check-exact-interval init 0 #xffff 'make-u16vector)
 	      (do ((i 0 (##core#inline "C_fixnum_plus" i 1)))
 		  ((##core#inline "C_fixnum_greater_or_equal_p" i len) v)
-		(##sys#u16vector-set! v i x) ) ) ) ) ) )
+		(##sys#u16vector-set! v i init) ) ) ) ) ) )
 
   (set! make-s16vector
-    (lambda (len . init)
+    (lambda (len #!optional (init #f)  (ext? #f) (fin #t))
       (##sys#check-exact len 'make-s16vector)
-      (let ((v (##sys#make-structure 's16vector (alloc (##core#inline "C_fixnum_shift_left" len 1)))))
-	(if (##core#inline "C_eqp" init '())
+      (let ((v (##sys#make-structure 's16vector (alloc 'make-s16vector (##core#inline "C_fixnum_shift_left" len 1) ext?))))
+	(when (and ext? fin?) (set-finalizer! v ext-free))
+	(if (not init)
 	    v
-	    (let ((x (##core#inline "C_slot" init 0)))
-	      (##sys#check-exact-interval x -32768 32767 'make-s16vector)
+	    (begin
+	      (##sys#check-exact-interval init -32768 32767 'make-s16vector)
 	      (do ((i 0 (##core#inline "C_fixnum_plus" i 1)))
 		  ((##core#inline "C_fixnum_greater_or_equal_p" i len) v)
-		(##sys#s16vector-set! v i x) ) ) ) ) ) )
+		(##sys#s16vector-set! v i init) ) ) ) ) ) )
 
   (set! make-u32vector
-    (lambda (len . init)
-      (##sys#check-exact len 'make-ui32vector)
-      (let ((v (##sys#make-structure 'u32vector (alloc (##core#inline "C_fixnum_shift_left" len 2)))))
-	(if (##core#inline "C_eqp" init '())
+    (lambda (len #!optional (init #f)  (ext? #f) (fin #t))
+      (##sys#check-exact len 'make-u32vector)
+      (let ((v (##sys#make-structure 'u32vector (alloc 'make-u32vector (##core#inline "C_fixnum_shift_left" len 2) ext?))))
+	(when (and ext? fin?) (set-finalizer! v ext-free))
+	(if (not init)
 	    v
-	    (let ((x (##core#inline "C_slot" init 0)))
-	      (##sys#check-exact x 'make-u32vector)
+	    (begin
+	      (##sys#check-exact init 'make-u32vector)
 	      (do ((i 0 (##core#inline "C_fixnum_plus" i 1)))
 		  ((##core#inline "C_fixnum_greater_or_equal_p" i len) v)
-		(##sys#u32vector-set! v i x) ) ) ) ) ) )
+		(##sys#u32vector-set! v i init) ) ) ) ) ) )
 
   (set! make-s32vector
-    (lambda (len . init)
+    (lambda (len #!optional (init #f)  (ext? #f) (fin #t))
       (##sys#check-exact len 'make-s32vector)
-      (let ((v (##sys#make-structure 's32vector (alloc (##core#inline "C_fixnum_shift_left" len 2)))))
-	(if (##core#inline "C_eqp" init '())
+      (let ((v (##sys#make-structure 's32vector (alloc 'make-s32vector (##core#inline "C_fixnum_shift_left" len 2) ext?))))
+	(when (and ext? fin?) (set-finalizer! v ext-free))
+	(if (not init)
 	    v
-	    (let ((x (##core#inline "C_slot" init 0)))
-	      (##sys#check-exact x 'make-s32vector)
+	    (begin
+	      (##sys#check-exact init 'make-s32vector)
 	      (do ((i 0 (##core#inline "C_fixnum_plus" i 1)))
 		  ((##core#inline "C_fixnum_greater_or_equal_p" i len) v)
-		(##sys#s32vector-set! v i x) ) ) ) ) ) )
+		(##sys#s32vector-set! v i init) ) ) ) ) ) )
 
   (set! make-f32vector
-    (lambda (len . init)
+    (lambda (len #!optional (init #f)  (ext? #f) (fin #t))
       (##sys#check-exact len 'make-f32vector)
-      (let ((v (##sys#make-structure 'f32vector (alloc (##core#inline "C_fixnum_shift_left" len 2)))))
-	(if (##core#inline "C_eqp" init '())
+      (let ((v (##sys#make-structure 'f32vector (alloc 'make-f32vector (##core#inline "C_fixnum_shift_left" len 2) ext?))))
+	(when (and ext? fin?) (set-finalizer! v ext-free))
+	(if (not init)
 	    v
-	    (let ((x (##core#inline "C_slot" init 0)))
-	      (##sys#check-number x 'make-f32vector)
-	      (if (not (##core#inline "C_blockp" x))
-		  (set! x (exact->inexact x)) )
+	    (begin
+	      (##sys#check-number init 'make-f32vector)
+	      (unless (##core#inline "C_blockp" init)
+		(set! init (exact->inexact init)) )
 	      (do ((i 0 (##core#inline "C_fixnum_plus" i 1)))
 		  ((##core#inline "C_fixnum_greater_or_equal_p" i len) v)
-		(##sys#f32vector-set! v i x) ) ) ) ) ) )
+		(##sys#f32vector-set! v i init) ) ) ) ) ) )
 
   (set! make-f64vector
-    (lambda (len . init)
+    (lambda (len #!optional (init #f)  (ext? #f) (fin #t))
       (##sys#check-exact len 'make-f64vector)
-      (let ((v (##sys#make-structure 'f64vector (alloc (##core#inline "C_fixnum_shift_left" len 3)))))
-	(if (##core#inline "C_eqp" init '())
+      (let ((v (##sys#make-structure 'f64vector (alloc 'make-f64vector (##core#inline "C_fixnum_shift_left" len 3) ext?))))
+	(when (and ext? fin?) (set-finalizer! v ext-free))
+	(if (not init)
 	    v
-	    (let ((x (##core#inline "C_slot" init 0)))
-	      (##sys#check-number x 'make-f64vector)
-	      (if (not (##core#inline "C_blockp" x))
-		  (set! x (exact->inexact x)) )
+	    (begin
+	      (##sys#check-number init 'make-f64vector)
+	      (unless (##core#inline "C_blockp" init)
+		(set! init (exact->inexact init)) )
 	      (do ((i 0 (##core#inline "C_fixnum_plus" i 1)))
 		  ((##core#inline "C_fixnum_greater_or_equal_p" i len) v)
-		(##sys#f64vector-set! v i x) ) ) ) ) ) ) )
+		(##sys#f64vector-set! v i init) ) ) ) ) ) ) )
 
 
 ;;; Creating vectors from a list:
