@@ -1,6 +1,6 @@
 ;;;; support.scm - Miscellaneous support code for the CHICKEN compiler
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -37,20 +37,20 @@
 
 
 #{compiler
-  build-information compiler-arguments process-command-line dump-nodes
+  compiler-arguments process-command-line dump-nodes dump-undefined-globals
   default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home optimization-iterations compiler-cleanup-hook decompose-lambda-list
-  foreign-type-table-size file-io-only banner custom-declare-alist
+  foreign-type-table-size file-io-only banner custom-declare-alist parse-easy-ffi
   unit-name insert-timer-checks used-units source-filename pending-canonicalizations
   foreign-declarations block-compilation analysis-database-size line-number-database-size
-  target-heap-size target-stack-size try-harder default-installation-home 
+  target-heap-size target-stack-size try-harder default-installation-home check-global-exports
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
   dependency-list broken-constant-nodes inline-substitutions-enabled no-c-syntax-checks
-  always-bound-to-procedure block-variable-literal copy-node! valid-c-identifier?
+  always-bound-to-procedure block-variable-literal copy-node! valid-c-identifier? tree-copy copy-node-tree-and-rename
   direct-call-ids foreign-type-table first-analysis scan-sharp-greater-string enable-sharp-greater-read-syntax
   expand-profile-lambda profile-lambda-list profile-lambda-index profile-info-vector-name
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
@@ -58,9 +58,9 @@
   reorganize-recursive-bindings substitution-table simplify-named-call check-c-syntax
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
   transform-direct-lambdas! finish-foreign-result compressable-literal csc-control-file
-  debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
+  debugging-chicken bomb check-signature posq stringify symbolify flonum? build-lambda-list
   string->c-identifier c-ify-string words words->bytes check-and-open-input-file close-checked-input-file fold-inner
-  constant? compute-module-exports register-optimizer user-level-optimizers
+  constant? basic-literal? source-info->string
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
   put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
   build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node expression-has-side-effects?
@@ -68,11 +68,11 @@
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database estimate-foreign-result-location-size
   real-name real-name-table set-real-name! real-name2 display-real-name-table display-line-number-database
-  copyright compiler-features default-declarations units-used-by-default words-per-flonum
+  default-declarations units-used-by-default words-per-flonum emit-control-file-item
   foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   default-optimization-iterations chop-separator chop-extension follow-without-loop dump-exported-globals
-  generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration optimizers-used
+  generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
   foreign-argument-conversion foreign-result-conversion final-foreign-type debugging export-list block-globals
   make-random-name foreign-type-convert-result foreign-type-convert-argument process-custom-declaration}
 
@@ -87,8 +87,6 @@
 (define (compiler-cleanup-hook) #f)
 
 (define debugging-chicken '())
-(define warnings-enabled #t)
-(define optimizers-used #f)
 
 (define (bomb . msg-and-args)
   (if (pair? msg-and-args)
@@ -107,8 +105,8 @@
 	 (flush-output)
 	 #t) ) )
 
-(define (warning msg . args)
-  (when warnings-enabled
+(define (warning msg . args)		; this redefines library's `warning'
+  (when ##sys#warnings-enabled
     (let ((out (current-error-port)))
       (apply fprintf out (string-append "Warning: " msg) args)
       (newline out) ) ) )
@@ -126,11 +124,17 @@
       (for-each (lambda (x) (write x out) (newline out)) args)
       (exit 70) ) ) )
 
+(define (map-llist proc llist)
+  (let loop ([llist llist])
+    (cond [(null? llist) '()]
+	  [(symbol? llist) (proc llist)]
+	  [else (cons (proc (car llist)) (loop (cdr llist)))] ) ) )
+
 (define (check-signature var args llist)
   (define (err)
     (quit "Arguments to inlined call of `~A' do not match parameter-list ~A" 
 	  (real-name var)
-	  (map real-name (cdr llist)) ) )
+	  (map-llist real-name (cdr llist)) ) )
   (let loop ([as args] [ll llist])
     (cond [(null? ll) (unless (null? as) (err))]
 	  [(symbol? ll)]
@@ -175,7 +179,7 @@
 	  '(#\")
 	  (let* ((c (car chars))
 		 (code (char->integer c)) )
-	    (if (or (< code 32) (memq c '(#\" #\' #\\ #\?)))
+	    (if (or (< code 32) (> code 128) (memq c '(#\" #\' #\\ #\?)))
 		(append '(#\\)
 			(cond ((< code 8) '(#\0 #\0))
 			      ((< code 64) '(#\0))
@@ -266,6 +270,15 @@
 		    [else #f] ) ) ] )
     (and f (> count t) count) ) )
 
+(define (basic-literal? x)
+  (or (null? x)
+      (symbol? x)
+      (constant? x)
+      (and (vector? x) (every basic-literal? (vector->list x)))
+      (and (pair? x) 
+	   (basic-literal? (car x))
+	   (basic-literal? (cdr x)) ) ) )
+
 
 ;;; Expression manipulation:
 
@@ -427,7 +440,7 @@
 
 (define display-analysis-database
   (let ((names '((captured . cpt) (assigned . set) (boxed . box) (global . glo) (assigned-locally . stl)
-		 (contractable . con) (standard-binding . stb) (foldable . fld) (simple . sim)
+		 (contractable . con) (standard-binding . stb) (foldable . fld) (simple . sim) (inlinable . inl)
 		 (side-effecting . sef) (collapsable . col) (removable . rem) (constant . con)
 		 (undefined . und) (replacing . rpg) (unused . uud) (extended-binding . xtb)
 		 (customizable . cst) (has-unused-parameters . hup) (boxed-rest . bxr) ) ) )
@@ -444,7 +457,7 @@
 		 (begin
 		   (case (caar es)
 		     ((captured assigned boxed global contractable standard-binding foldable assigned-locally
-		       side-effecting collapsable removable undefined replacing unused simple
+		       side-effecting collapsable removable undefined replacing unused simple inlinable
 		       has-unused-parameters extended-binding customizable constant boxed-rest)
 		      (printf "\t~a" (cdr (assq (caar es) names))) )
 		     ((unknown)
@@ -478,11 +491,10 @@
 (define-record node
   class					; symbol
   parameters				; (value...)
-  subexpressions			; (node...)
-  user-field)				; anything
+  subexpressions)			; (node...)
 
 (define (make-node c p s)
-  (##sys#make-structure 'node c p s #f) ) ; this kludge is for allowing the inlined `make-node'
+  (##sys#make-structure 'node c p s) ) ; this kludge is for allowing the inlined `make-node'
 
 (define (varnode var) (make-node '##core#variable (list var) '()))
 (define (qnode const) (make-node 'quote (list const) '()))
@@ -546,14 +558,8 @@
 				#t]
 			       [else #f] )
 			 (if ln
-			     (let ([rn (real-name name)]
-				   [lns (number->string ln)] )
-			       (string-append 
-				source-filename ": " 
-				lns
-				(make-string (max 0 (- 4 (string-length lns))) #\space)
-				" "
-				(or rn (##sys#symbol->qualified-string name)) ) )
+			     (let ([rn (real-name name)])
+			       (list source-filename ln (or rn (##sys#symbol->qualified-string name))) )
 			     (##sys#symbol->qualified-string name) ) )
 		   (map walk x) ) ) ) ) )
 	    (else (make-node '##core#call '(#f) (map walk x))) ) )
@@ -567,7 +573,9 @@
 	  (params (node-parameters n)) 
 	  (class (node-class n)) )
       (case class
-	((if ##core#closure ##core#box ##core#cond) (cons class (map walk subs)))
+	((if ##core#box ##core#cond) (cons class (map walk subs)))
+	((##core#closure)
+	 `(##core#closure ,params ,@(map walk subs)) )
 	((##core#variable ##core#global-ref) (car params))
 	((quote) `(quote ,(car params)))
 	((let)
@@ -600,28 +608,73 @@
 	 (list (proc (first vars) (second vars))
 	       (fold (cdr vars)) ) ) ) ) )
 
-(define (inline-lambda-bindings llist args body)
+(define (inline-lambda-bindings llist args body copy?)
   (decompose-lambda-list
    llist
    (lambda (vars argc rest)
      (receive (largs rargs) (split-at args argc)
-       (fold-right
-	(lambda (var val body) (make-node 'let (list var) (list val body)) )
-	(if rest
-	    (make-node
-	     'let (list rest)
-	     (list (if (null? rargs)
-		       (qnode '())
-		       (make-node '##core#inline_allocate (list "C_a_i_list" (* 3 (length rargs))) rargs) )
-		   body) )
-	    body)
-	(take vars argc)
-	largs) ) ) ) )
+       (let* ([rlist (if copy? (map gensym vars) vars)]
+	      [body (if copy? 
+			(copy-node-tree-and-rename body vars rlist)
+			body) ] )
+	 (fold-right
+	  (lambda (var val body) (make-node 'let (list var) (list val body)) )
+	  (if rest
+	      (make-node
+	       'let (list (last rlist))
+	       (list (if (null? rargs)
+			 (qnode '())
+			 (make-node '##core#inline_allocate (list "C_a_i_list" (* 3 (length rargs))) rargs) )
+		     body) )
+	      body)
+	  (take rlist argc)
+	  largs) ) ) ) ) )
+
+(define (copy-node-tree-and-rename node vars aliases)
+  (let ([rlist (map cons vars aliases)])
+    (define (rename v rl) (alist-ref v rl eq? v))
+    (define (walk n rl)
+      (let ([subs (node-subexpressions n)]
+	    [params (node-parameters n)]
+	    [class (node-class n)] )
+	(case class
+	  [(##core#variable) (varnode (rename (first params) rl))]
+	  [(set!) (make-node 'set! (list (rename (first params) rl)) (map (cut walk <> rl) subs))]
+	  [(let) 
+	   (let* ([v (first params)]
+		  [a (gensym v)]
+		  [rl2 (alist-cons v a rl)] )
+	     (make-node 'let (list a) (map (cut walk <> rl2) subs)) ) ]
+	  [(##core#lambda)
+	   (decompose-lambda-list
+	    (third params)
+	    (lambda (vars argc rest)
+	      (let* ([as (map gensym vars)]
+		     [rl2 (append as rl)] )
+		(make-node 
+		 '##core#lambda
+		 (list (first params) (second params) 
+		       (build-lambda-list as argc (and rest (rename rest rl2)))
+		       (fourth params) )
+		 (map (cut walk <> rl2) subs) ) ) ) ) ]
+	  [else (make-node class (tree-copy params) (map (cut walk <> rl) subs))] ) ) )
+    (walk node rlist) ) )
+
+(define (tree-copy t)
+  (let rec ([t t])
+    (if (pair? t)
+	(cons (rec (car t)) (rec (cdr t)))
+	t) ) )
 
 (define (copy-node! from to)
   (node-class-set! to (node-class from))
   (node-parameters-set! to (node-parameters from))
-  (node-subexpressions-set! to (node-subexpressions from)) )
+  (node-subexpressions-set! to (node-subexpressions from)) 
+  (let ([len-from (##sys#size from)]
+	[len-to (##sys#size to)] )
+    (do ([i 4 (fx+ i 1)])
+	((or (fx>= i len-from) (fx>= i len-to)))
+      (##sys#setslot to i (##sys#slot from i)) ) ) )
 
 
 ;;; Match node-structure with pattern:
@@ -677,8 +730,10 @@
 
 (define (simple-lambda-node? node)
   (let* ([params (node-parameters node)]
-	 [k (first (third params))] )
-    (and (second params)
+	 [llist (third params)]
+	 [k (and (pair? llist) (first llist))] ) ; leaf-routine has no continuation argument
+    (and k 
+	 (second params)
 	 (let rec ([n node])
 	   (case (node-class n)
 	     [(##core#call)
@@ -689,7 +744,42 @@
 		     (every rec (cdr subs)) ) ) ]
 	     [(##core#callunit) #f]
 	     [else (every rec (node-subexpressions n))] ) ) ) ) )
+
+
+;;; Some safety checks and database dumping:
   
+(define (dump-exported-globals db)
+  (unless block-compilation
+    (##sys#hash-table-for-each
+     (lambda (sym plist)
+       (when (and (assq 'global plist) 
+		  (assq 'assigned plist)
+		  (or (and export-list (memq sym export-list))
+		      (not (memq sym block-globals)) ) )
+	 (write sym) 
+	 (newline) ) )
+     db) ) )
+
+(define (dump-undefined-globals db)
+  (##sys#hash-table-for-each
+   (lambda (sym plist)
+     (when (and (assq 'global plist)
+		(not (assq 'assigned plist)) )
+       (write sym)
+       (newline) ) )
+   db) )
+
+(define (check-global-exports db)
+  (when export-list
+    (let ([exps export-list])
+      (##sys#hash-table-for-each
+       (lambda (sym plist)
+	 (when (and (memq sym exps) (not (assq 'assigned plist)))
+	   (warning "exported global variable `~S' is used but not defined" sym) )
+	 (set! exps (delete sym exps eq?)) )
+       db)
+      (for-each (cut warning "exported global variable `~S' is not defined" <>) exps) ) ) )
+
 
 ;;; Compute general statistics from analysis database:
 ;
@@ -743,18 +833,6 @@
      (printf ";   known call sites: \t~s\n" sites) 
      (printf ";   database entries: \t~s\n" entries) ) ) )
 
-(define (dump-exported-globals db)
-  (unless block-compilation
-    (##sys#hash-table-for-each
-     (lambda (sym plist)
-       (when (and (assq 'global plist) 
-		  (assq 'assigned plist)
-		  (or (and export-list (memq sym export-list))
-		      (not (memq sym block-globals)) ) )
-	 (write sym) 
-	 (newline) ) )
-     db) ) )
-
 
 ;;; Pretty-print expressions:
 
@@ -781,86 +859,93 @@
       (follow-without-loop
        type
        (lambda (t next)
-	 (case t
-	   [(char unsigned-char) (if unsafe param `(##sys#foreign-char-argument ,param))]
-	   [(int unsigned-int short unsigned-short byte unsigned-byte) (if unsafe param `(##sys#foreign-fixnum-argument ,param))]
-	   [(float double) (if unsafe param `(##sys#foreign-flonum-argument ,param))]
-	   [(pointer byte-vector)
-	    (let ([tmp (gensym)])
-	      `(let ([,tmp ,param])
-		 (if ,tmp
-		     ,(if unsafe
-			  tmp
-			  `(##sys#foreign-block-argument ,tmp) )
-		     '#f) ) ) ]
-	   [(nonnull-pointer nonnull-byte-vector)
-	    (if unsafe
-		param
-		`(##sys#foreign-block-argument ,param) ) ]
-	   [(u8vector u16vector s8vector s16vector u32vector s32vector f32vector f64vector)
-	    (let ([tmp (gensym)])
-	      `(let ([,tmp ,param])
-		 (if ,tmp
-		     ,(if unsafe
-			  tmp
-			  `(##sys#foreign-number-vector-argument ',t ,tmp) )
-		     '#f) ) ) ]
-	   [(nonnull-u8vector nonnull-u16vector nonnull-s8vector nonnull-s16vector nonnull-u32vector nonnull-s32vector 
-			      nonnull-f32vector nonnull-f64vector)
-	    (if unsafe
-		param
-		`(##sys#foreign-number-vector-argument 
-		  ',(##sys#slot (assq t tmap) 1)
-		  ,param) ) ]
-	   [(integer long) (if unsafe param `(##sys#foreign-integer-argument ,param))]
-	   [(unsigned-integer unsigned-long)
-	    (if unsafe
-		param
-		`(##sys#foreign-unsigned-integer-argument ,param) ) ]
-	   [(c-pointer)
-	    (let ([tmp (gensym)])
-	      `(let ([,tmp ,param])
-		 (if ,tmp
-		     (##sys#foreign-pointer-argument ,tmp)
-		     '#f) ) ) ]
-	   [(nonnull-c-pointer)
-	    `(##sys#foreign-pointer-argument ,param) ]
-	   [(c-string c-string*)
-	    (let ([tmp (gensym)])
-	      `(let ([,tmp ,param])
-		 (if ,tmp
-		     ,(if unsafe 
-			  `(##sys#make-c-string ,tmp)
-			  `(##sys#make-c-string (##sys#foreign-string-argument ,tmp)) )
-		     '#f) ) ) ]
-	   [(nonnull-c-string nonnull-c-string*)
-	    (if unsafe 
-		`(##sys#make-c-string ,param)
-		`(##sys#make-c-string (##sys#foreign-string-argument ,param)) ) ]
-	   [else
-	    (cond [(and (symbol? t) (##sys#hash-table-ref foreign-type-table t))
-		   => (lambda (t)
-			(next (if (vector? t) (vector-ref t 0) t)) ) ]
-		  [(pair? t)
-		   (match t
-		     [((or 'ref 'pointer 'function 'c-pointer) . _)
-		      (let ([tmp (gensym)])
-			`(let ([,tmp ,param])
-			   (if ,tmp
-			       (##sys#foreign-pointer-argument ,tmp)
-			       '#f) ) )  ]
-		     [('instance . _)
-		      (let ([tmp (gensym)])
-			`(let ([,tmp ,param])
-			   (if ,tmp
-			       (slot-ref ,param 'this)
-			       '#f) ) ) ]
-		     [('nonnull-instance . _)
-		      `(slot-ref ,param 'this) ]
-		     [((or 'nonnull-pointer 'nonnull-c-pointer) . _)
-		      `(##sys#foreign-pointer-argument ,param) ]
-		     [_ param] ) ]
-		  [else param] ) ] ) )
+	 (let repeat ([t t])
+	   (case t
+	     [(char unsigned-char) (if unsafe param `(##sys#foreign-char-argument ,param))]
+	     [(int unsigned-int short unsigned-short byte unsigned-byte int32 unsigned-int32)
+	      (if unsafe param `(##sys#foreign-fixnum-argument ,param))]
+	     [(float double number) (if unsafe param `(##sys#foreign-flonum-argument ,param))]
+	     [(pointer byte-vector scheme-pointer) ; pointer is DEPRECATED
+	      (let ([tmp (gensym)])
+		`(let ([,tmp ,param])
+		   (if ,tmp
+		       ,(if unsafe
+			    tmp
+			    `(##sys#foreign-block-argument ,tmp) )
+		       '#f) ) ) ]
+	     [(nonnull-pointer nonnull-scheme-pointer nonnull-byte-vector) ; nonnull-pointer is DEPRECATED
+	      (if unsafe
+		  param
+		  `(##sys#foreign-block-argument ,param) ) ]
+	     [(u8vector u16vector s8vector s16vector u32vector s32vector f32vector f64vector)
+	      (let ([tmp (gensym)])
+		`(let ([,tmp ,param])
+		   (if ,tmp
+		       ,(if unsafe
+			    tmp
+			    `(##sys#foreign-number-vector-argument ',t ,tmp) )
+		       '#f) ) ) ]
+	     [(nonnull-u8vector nonnull-u16vector nonnull-s8vector nonnull-s16vector nonnull-u32vector nonnull-s32vector 
+				nonnull-f32vector nonnull-f64vector)
+	      (if unsafe
+		  param
+		  `(##sys#foreign-number-vector-argument 
+		    ',(##sys#slot (assq t tmap) 1)
+		    ,param) ) ]
+	     [(integer long integer32) (if unsafe param `(##sys#foreign-integer-argument ,param))]
+	     [(unsigned-integer unsigned-integer32 unsigned-long)
+	      (if unsafe
+		  param
+		  `(##sys#foreign-unsigned-integer-argument ,param) ) ]
+	     [(c-pointer)
+	      (let ([tmp (gensym)])
+		`(let ([,tmp ,param])
+		   (if ,tmp
+		       (##sys#foreign-pointer-argument ,tmp)
+		       '#f) ) ) ]
+	     [(nonnull-c-pointer)
+	      `(##sys#foreign-pointer-argument ,param) ]
+	     [(c-string c-string*)
+	      (let ([tmp (gensym)])
+		`(let ([,tmp ,param])
+		   (if ,tmp
+		       ,(if unsafe 
+			    `(##sys#make-c-string ,tmp)
+			    `(##sys#make-c-string (##sys#foreign-string-argument ,tmp)) )
+		       '#f) ) ) ]
+	     [(nonnull-c-string nonnull-c-string*)
+	      (if unsafe 
+		  `(##sys#make-c-string ,param)
+		  `(##sys#make-c-string (##sys#foreign-string-argument ,param)) ) ]
+	     [(symbol)
+	      (if unsafe 
+		  `(##sys#make-c-string (##sys#symbol->string ,param))
+		  `(##sys#make-c-string (##sys#foreign-string-argument (##sys#symbol->string ,param))) ) ]
+	     [else
+	      (cond [(and (symbol? t) (##sys#hash-table-ref foreign-type-table t))
+		     => (lambda (t)
+			  (next (if (vector? t) (vector-ref t 0) t)) ) ]
+		    [(pair? t)
+		     (match t
+		       [((or 'ref 'pointer 'function 'c-pointer) . _)
+			(let ([tmp (gensym)])
+			  `(let ([,tmp ,param])
+			     (if ,tmp
+				 (##sys#foreign-pointer-argument ,tmp)
+				 '#f) ) )  ]
+		       [((or 'instance 'instance-ref) . _)
+			(let ([tmp (gensym)])
+			  `(let ([,tmp ,param])
+			     (if ,tmp
+				 (slot-ref ,param 'this)
+				 '#f) ) ) ]
+		       [('nonnull-instance . _)
+			`(slot-ref ,param 'this) ]
+		       [('const t) (repeat t)]
+		       [((or 'nonnull-pointer 'nonnull-c-pointer) . _)
+			`(##sys#foreign-pointer-argument ,param) ]
+		       [_ param] ) ]
+		    [else param] ) ] ) ) )
        (lambda () (quit "foreign type `~S' refers to itself" type)) ) ) ) )
 
 
@@ -898,22 +983,22 @@
    type
    (lambda (t next)
      (case t
-       ((char int short bool void unsigned-short scheme-object unsigned-char unsigned-int byte unsigned-byte) 0)
-       ((c-string nonnull-c-string c-pointer nonnull-c-pointer
-		  c-string* nonnull-c-string*)
+       ((char int short bool void unsigned-short scheme-object unsigned-char unsigned-int byte unsigned-byte
+	      int32 unsigned-int32) 
+	0)
+       ((c-string nonnull-c-string c-pointer nonnull-c-pointer symbol c-string* nonnull-c-string*)
 	(words->bytes 3) )
-       ((unsigned-integer long integer unsigned-long)
+       ((unsigned-integer long integer unsigned-long integer32 unsigned-integer32)
 	(words->bytes 4) )
-       ((float double) 
-	(words->bytes 4) )		; possibly 8-byte aligned 32-bit double
+       ((float double number) 
+	(words->bytes 4) )		; possibly 8-byte aligned 64-bit double
        (else
 	(cond [(and (symbol? t) (##sys#hash-table-ref foreign-type-table t))
 	       => (lambda (t2)
 		    (next (if (vector? t2) (vector-ref t2 0) t2)) ) ]
 	      [(pair? t)
 	       (case (car t)
-		 [(ref nonnull-pointer pointer c-pointer nonnull-c-pointer function instance
-		       nonnull-instance) 
+		 [(ref nonnull-pointer pointer c-pointer nonnull-c-pointer function instance instance-ref nonnull-instance) 
 		  (words->bytes 3) ]
 		 [else 0] ) ]
 	      [else 0] ) ) ) )
@@ -927,10 +1012,11 @@
    (lambda (t next)
      (case t
        ((char int short bool unsigned-short unsigned-char unsigned-int long unsigned-long byte unsigned-byte
-	      c-pointer pointer nonnull-c-pointer unsigned-integer integer float c-string
-	      nonnull-c-string c-string* nonnull-c-string*) 
+	      c-pointer pointer nonnull-c-pointer unsigned-integer integer float c-string symbol
+	      scheme-pointer nonnull-scheme-pointer int32 unsigned-int32 integer32 unsigned-integer32
+	      nonnull-c-string c-string* nonnull-c-string*) ; pointer and nonnull-pointer are DEPRECATED
 	(words->bytes 1) )
-       ((double)
+       ((double number)
 	(words->bytes 2) )
        (else
 	(cond [(and (symbol? t) (##sys#hash-table-ref foreign-type-table t))
@@ -952,9 +1038,10 @@
     [(nonnull-c-string) `(##sys#peek-nonnull-c-string ,body '0)]
     [(c-string*) `(##sys#peek-and-free-c-string ,body '0)]
     [(nonnull-c-string*) `(##sys#peek-and-free-nonnull-c-string ,body '0)]
+    [(symbol) `(##sys#intern-symbol (##sys#peek-c-string ,body '0))]
     [else
      (match type
-       [('instance cname sname)
+       [((or 'instance 'instance-ref) cname sname)
 	`(##tinyclos#make-instance-from-pointer ,body ,sname) ]
        [('nonnull-instance cname sname)
 	`(make ,sname 'this ,body) ]
@@ -1064,7 +1151,7 @@
 
 (define (chop-separator str)
   (let ([len (sub1 (string-length str))])
-    (if (and (> len 0) (char=? (string-ref str len) pathname-directory-separator))
+    (if (and (> len 0) (char=? (string-ref str len) ##sys#pathname-directory-separator))
 	(substring str 0 len)
 	str) ) )
 
@@ -1079,8 +1166,8 @@
 ;;; Print version/usage information:
 
 (define (print-version . b)
-  (when (:optional b #f) (printf "~A~%" banner))
-  (printf "~A~%~A" build-information copyright) )
+  (when (:optional b #f) (printf "~A" banner))
+  (printf "~A~%~A" (chicken-version #t) copyright) )
 
 (define (print-usage)
   (print-version)
@@ -1109,24 +1196,17 @@ Usage: chicken FILENAME OPTION ...
   Language options:
 
     -feature SYMBOL             register feature identifier
-    -no-feature SYMBOL          unregister feature identifier
     -ffi-define SYMBOL          define preprocessor macro for ``easy'' FFI parser
     -ffi-include-path PATH      set include path for ``easy'' FFI parser
-    -ffi                        compile C/C++ code and generate Scheme bindings
-    -r5rs                       equivalent to `-hygienic -strict'
+    -ffi-no-include             don't parse include files when encountered by the FFI parser
+    -ffi                        parse and compile C/C++ code and generate Scheme bindings
+    -ffi-parse                  parse C/C++ code without including it
 
   Syntax related options:
 
     -case-insensitive           don't preserve case of read symbols
-    -hygienic -syntax           use syntax-case macro package
     -keyword-style STYLE        allow alternative keyword syntax (none, prefix or suffix)
-    -strict                     disable non-standard macros and symbolsyntax
-    -strict-srfi-0              disable non-standard macros except `cond-expand'
-    -strict-reader              disable non-standard read syntax
-    -strict-letrec              enable fully R5RS compliant `letrec'
-    -run-time-macros            low-level macros are made available at run-time
-    -hygienic-at-run-time -syntax-at-run-time
-                                enable hygienic macros at run-time
+    -run-time-macros            macros are made available at run-time
 
   Translation options:
 
@@ -1140,19 +1220,24 @@ Usage: chicken FILENAME OPTION ...
     -debug-level NUMBER         set level of available debugging information
     -no-trace                   disable tracing information
     -profile                    executable emits profiling information 
+    -profile-name FILENAME      name of the generated profile information file
+    -accumulate-profile         executable emits profiling information in append mode
+    -no-lambda-info             omit additional procedure-information
 
   Optimization options:
 
     -optimize-level NUMBER      enable certain sets of optimization options
     -optimize-leaf-routines     enable leaf routine optimization
     -lambda-lift                enable lambda-lifting
-    -usual-integrations         assume standard procedures are not redefined
+    -no-usual-integrations      standard procedures may be redefined
     -unsafe                     disable safety checks
     -block                      enable block-compilation
     -disable-interrupts         disable interrupts in compiled code
     -fixnum-arithmetic          assume all numbers are fixnums
     -benchmark-mode             fixnum mode, no interrupts and opt.-level 3
     -disable-stack-overflow-checks  disables detection of stack-overflows.
+    -inline                     enable inlining
+    -inline-limit               set inlining threshold
 
   Configuration options:
 
@@ -1170,6 +1255,8 @@ Usage: chicken FILENAME OPTION ...
     -prologue FILENAME          include file before main source file
     -epilogue FILENAME          include file after main source file
     -dynamic                    compile as dynamically loadable code
+    -require-extension NAME     require extension NAME in compiled code
+    -extension                  compile as extension (dynamic or static)
 
   Obscure options:
 
@@ -1178,6 +1265,9 @@ Usage: chicken FILENAME OPTION ...
     -disable-c-syntax-checks    disable syntax check of C code fragments
     -unsafe-libraries           marks the generated file as being linked
                                 with the unsafe runtime system
+    -raw                        do not generate implicit init- and exit code			       
+    -emit-external-prototypes-first  emit protoypes for callbacks before foreign
+                                declarations
 
 EOF
 ) )
@@ -1241,6 +1331,13 @@ EOF
      (printf "~S ~S~%" key val) )
    real-name-table) )
 
+(define (source-info->string info)
+  (match info
+    ((file ln name)
+     (let ((lns (->string ln)))
+       (conc file ": " lns (make-string (max 0 (- 4 (string-length lns))) #\space) " " name) ) )
+    (_ (and info (->string info))) ) )
+
 
 ;;; We need this for constant folding:
 
@@ -1258,6 +1355,13 @@ EOF
 	  [i2 (+ i 2)] )
       (printf "~%~A<~A ~S" ind class params)
       (for-each (cut loop i2 <>) subs)
+      (let ([len (##sys#size n)])
+	(when (fx> len 4)
+	  (printf "[~S" (##sys#slot n 4))
+	  (do ([i 5 (fx+ i 1)])
+	      ((fx>= i len))
+	    (printf " ~S" (##sys#slot n i)) )
+	  (write-char #\]) ) )
       (write-char #\>) ) )
   (newline) )
 
@@ -1272,7 +1376,6 @@ EOF
 	    (let ([_ (read-char port)]		; swallow #\>
 		  [decl #f]
 		  [parse #f]
-		  [swig #f]
 		  [exec #f] )
 	      (case (peek-char port)
 		[(#\!)
@@ -1288,7 +1391,6 @@ EOF
 		[(#\()
 		 (let ([head (read port)])
 		   (match head
-		     [('swig module args ...) (set! swig (cdr head))]
 		     [_ (for-each
 			 (lambda (t)
 			   (case t
@@ -1300,20 +1402,7 @@ EOF
 		[else (set! decl #t)] )
 	      (let ([text (scan-sharp-greater-string port)])
 		(check-c-syntax text)
-		(when swig (set! no-c-syntax-checks #t))
 		`(begin
-		   ,@(if swig
-			 (let* ([mod (car swig)]
-				[modname (->string mod)]
-				[init (string-append modname "_swig_init")] )
-			   `((declare 
-			       (foreign-declare
-				,(sprintf "C_extern void ~A(int,C_word,C_word) C_noret;" init) ) 
-			       (custom-declare
-				(swig ,mod ,(string-append modname "_swig.i") ,@(cdr swig))
-				,text) )
-			     ((##core#primitive ,init)) ) )
-			 '() )
 		   ,@(if decl
 			 `((declare (foreign-declare ,text)))
 			 '() )
@@ -1325,7 +1414,7 @@ EOF
 			   `((declare 
 			       (foreign-declare
 				,(sprintf "static C_word ~A() { ~A; return C_SCHEME_UNDEFINED; }\n" tmp text) ) )
-			     (##core#inline ,tmp) ) )
+			     (##core#inline ,(symbol->string tmp)) ) )
 			 '() ) ) ) )
 	    (old-hook char port) ) ) ) ) )
 
@@ -1336,7 +1425,6 @@ EOF
 	(cond [(eof-object? c) (quit "unexpected end of `#> ... <#' sequence")]
 	      [(char=? c #\newline)
 	       (newline out)
-	       (set! ##sys#read-line-counter (add1 ##sys#read-line-counter)) 
 	       (loop) ]
 	      [(char=? c #\<)
 	       (let ([c (read-char port)])
@@ -1371,43 +1459,16 @@ EOF
 	    (lambda ()
 	      (close-output-port out)
 	      (old) ) ) )
-	(unless csc-control-file
-	  (set! csc-control-file (open-output-file (pathname-replace-extension source-filename "csc")))
-	  (display "#%csc\n" csc-control-file) 
-	  (set! compiler-cleanup-hook
-	    (let ([old compiler-cleanup-hook])
-	      (lambda ()
-		(close-output-port csc-control-file)
-		(old) ) ) ) )
-	(fprintf csc-control-file "~S~%" (cons* tag name fname args)) ) )
+	(emit-control-file-item (cons* tag name fname args)) ) )
     (for-each (cute display <> (cdr a)) strings) ) )
 
-
-;;; Compute real names for identifiers exported from modules:
-
-(define (compute-module-exports mexps exps)
-  (define (real-names syms)
-    (append-map
-     (lambda (sym)
-       (let ([sym2 (macroexpand sym)])
-	 (if (eq? sym2 sym)
-	     '()
-	     `((##core#set! ,sym ,sym2)) ) ) )
-     syms) )
-  (concatenate
-   (cons
-    (if exps 
-	(real-names exps)
-	'() )
-    (map (match-lambda
-	   [(mod . exps) 
-	    (macroexpand `(import ,mod))
-	    (real-names exps) ] )
-	 mexps) ) ) ) 
-
-
-;;; User-defined optimizers:
-
-(define (register-optimizer sym proc)
-  (set! optimizers-used #t)
-  (##sys#hash-table-set! user-level-optimizers sym proc) )
+(define (emit-control-file-item item)
+  (unless csc-control-file
+    (set! csc-control-file (open-output-file (pathname-replace-extension source-filename "csc")))
+    (display "#%csc\n" csc-control-file) 
+    (set! compiler-cleanup-hook
+      (let ([old compiler-cleanup-hook])
+	(lambda ()
+	  (close-output-port csc-control-file)
+	  (old) ) ) ) )
+  (fprintf csc-control-file "~S~%" item) )
