@@ -1,6 +1,6 @@
 ;;;; tcp.scm - Networking stuff
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -40,7 +40,7 @@
   (fixnum-arithmetic)
   (no-bound-checks)
   (export tcp-close tcp-listen tcp-connect tcp-accept tcp-accept-ready? ##sys#tcp-port->fileno tcp-listener? tcp-addresses
-	  tcp-abandon-port tcp-listener-port)
+	  tcp-abandon-port tcp-listener-port tcp-listener-fileno tcp-port-numbers)
   (bound-to-procedure
    ##net#socket ##net#bind ##net#connect ##net#listen ##net#accept 
    ##net#close ##net#recv ##net#send ##net#select ##net#select-write ##net#gethostaddr ##net#io-ports ##sys#update-errno
@@ -49,10 +49,19 @@
   (foreign-declare #<<EOF
 #include <errno.h>
 #ifdef _WIN32
+# if _MSC_VER > 1300
 # include <winsock2.h>
+# include <ws2tcpip.h>
+# else
+# include <winsock.h>
+# endif
+# define socklen_t       int
 static WSADATA wsa;
 # define fcntl(a, b, c)  0
 # define EWOULDBLOCK     0
+# define EINPROGRESS     0
+# define typecorrect_getsockopt(socket, level, optname, optval, optlen)	\
+    getsockopt(socket, level, optname, (char *)optval, optlen)
 #else
 # include <fcntl.h>
 # include <sys/types.h>
@@ -61,10 +70,19 @@ static WSADATA wsa;
 # include <netinet/in.h>
 # include <unistd.h>
 # include <netdb.h>
-# define SD_RECEIVE      0
-# define SD_SEND         1
+# include <signal.h>
 # define closesocket     close
 # define INVALID_SOCKET  -1
+# define typecorrect_getsockopt getsockopt
+#endif
+
+#ifndef SD_RECEIVE
+# define SD_RECEIVE      0
+# define SD_SEND         1
+#endif
+
+#ifdef ECOS
+#include <sys/sockio.h>
 #endif
 
 static char addr_buffer[ 20 ];
@@ -76,7 +94,7 @@ EOF
 (cond-expand
  [unsafe
   (eval-when (compile)
-    (define-macro (##sys#check-structure x y) '(##core#undefined))
+    (define-macro (##sys#check-structure x y . _) '(##core#undefined))
     (define-macro (##sys#check-range x y z) '(##core#undefined))
     (define-macro (##sys#check-pair x) '(##core#undefined))
     (define-macro (##sys#check-list x) '(##core#undefined))
@@ -90,6 +108,7 @@ EOF
  [else] )
 
 (define-foreign-variable errno int "errno")
+(define-foreign-variable strerror c-string "strerror(errno)")
 
 (define-foreign-type sockaddr* (pointer "struct sockaddr"))
 (define-foreign-type sockaddr_in* (pointer "struct sockaddr_in"))
@@ -107,14 +126,14 @@ EOF
 (define-foreign-variable _einprogress int "EINPROGRESS")
 
 (define ##net#socket (foreign-lambda int "socket" int int int))
-(define ##net#bind (foreign-lambda int "bind" int pointer int))
+(define ##net#bind (foreign-lambda int "bind" int scheme-pointer int))
 (define ##net#listen (foreign-lambda int "listen" int int))
 (define ##net#accept (foreign-lambda int "accept" int c-pointer c-pointer))
 (define ##net#close (foreign-lambda int "closesocket" int))
-(define ##net#send (foreign-lambda int "send" int pointer int int))
-(define ##net#recv (foreign-lambda int "recv" int pointer int int))
+(define ##net#send (foreign-lambda int "send" int scheme-pointer int int))
+(define ##net#recv (foreign-lambda int "recv" int scheme-pointer int int))
 (define ##net#shutdown (foreign-lambda int "shutdown" int int))
-(define ##net#connect (foreign-lambda int "connect" int pointer int))
+(define ##net#connect (foreign-lambda int "connect" int scheme-pointer int))
 
 (define ##net#make-nonblocking
   (foreign-lambda* bool ([int fd])
@@ -127,7 +146,7 @@ EOF
     "struct sockaddr_in sa;"
     "unsigned char *ptr;"
     "int len = sizeof(struct sockaddr_in);"
-    "if(getsockname(s, (struct sockaddr *)&sa, &len) != 0) return(NULL);"
+    "if(getsockname(s, (struct sockaddr *)&sa, (socklen_t *)&len) != 0) return(NULL);"
     "ptr = (unsigned char *)&sa.sin_addr;"
     "sprintf(addr_buffer, \"%d.%d.%d.%d\", ptr[ 0 ], ptr[ 1 ], ptr[ 2 ], ptr[ 3 ]);"
     "return(addr_buffer);") )
@@ -135,17 +154,23 @@ EOF
 (define ##net#getsockport
   (foreign-lambda* int ([int s])
     "struct sockaddr_in sa;"
-    "unsigned char *ptr;"
     "int len = sizeof(struct sockaddr_in);"
-    "if(getsockname(s, (struct sockaddr *)&sa, &len) != 0) return(-1);"
+    "if(getsockname(s, (struct sockaddr *)&sa, (socklen_t *)(&len)) != 0) return(-1);"
     "else return(ntohs(sa.sin_port));") )
 
+(define ##net#getpeerport
+ (foreign-lambda* int ([int s])
+   "struct sockaddr_in sa;"
+   "int len = sizeof(struct sockaddr_in);"
+   "if(getpeername(s, (struct sockaddr *)&sa, (socklen_t *)(&len)) != 0) return(-1);"
+   "else return(ntohs(sa.sin_port));") )
+
 (define ##net#getpeername 
   (foreign-lambda* c-string ([int s])
     "struct sockaddr_in sa;"
     "unsigned char *ptr;"
-    "int len = sizeof(struct sockaddr_in);"
-    "if(getpeername(s, (struct sockaddr *)&sa, &len) != 0) return(NULL);"
+    "unsigned int len = sizeof(struct sockaddr_in);"
+    "if(getpeername(s, (struct sockaddr *)&sa, ((unsigned int *)&len)) != 0) return(NULL);"
     "ptr = (unsigned char *)&sa.sin_addr;"
     "sprintf(addr_buffer, \"%d.%d.%d.%d\", ptr[ 0 ], ptr[ 1 ], ptr[ 2 ], ptr[ 3 ]);"
     "return(addr_buffer);") )
@@ -155,6 +180,7 @@ EOF
 #ifdef _WIN32
      return(WSAStartup(MAKEWORD(1, 1), &wsa) == 0);
 #else
+     signal(SIGPIPE, SIG_IGN);
      return(1);
 #endif
 EOF
@@ -167,30 +193,34 @@ EOF
   (foreign-lambda* int ((c-string serv) (c-string proto))
     "struct servent *se;
      if((se = getservbyname(serv, proto)) == NULL) return(0);
-     else return(se->s_port);") )     
+     else return(ntohs(se->s_port));") )     
 
 (define ##net#select
   (foreign-lambda* int ((int fd))
     "fd_set in;
      struct timeval tm;
+     int rv;
      FD_ZERO(&in);
      FD_SET(fd, &in);
      tm.tv_sec = tm.tv_usec = 0;
-     if(select(fd + 1, &in, NULL, NULL, &tm) == -1) return(-1);
-     else return(FD_ISSET(fd, &in) ? 1 : 0);") )
+     rv = select(fd + 1, &in, NULL, NULL, &tm);
+     if(rv > 0) { rv = FD_ISSET(fd, &in) ? 1 : 0; }
+     return(rv);") )
 
 (define ##net#select-write
   (foreign-lambda* int ((int fd))
     "fd_set out;
      struct timeval tm;
+     int rv;
      FD_ZERO(&out);
      FD_SET(fd, &out);
      tm.tv_sec = tm.tv_usec = 0;
-     if(select(fd + 1, NULL, &out, NULL, &tm) == -1) return(-1);
-     else return(FD_ISSET(fd, &out) ? 1 : 0);") )
+     rv = select(fd + 1, NULL, &out, NULL, &tm);
+     if(rv > 0) { rv = FD_ISSET(fd, &out) ? 1 : 0; }
+     return(rv);") )
 
 (define ##net#gethostaddr
-  (foreign-lambda* bool ((pointer saddr) (c-string host) (unsigned-short port))
+  (foreign-lambda* bool ((scheme-pointer saddr) (c-string host) (unsigned-short port))
     "struct hostent *he = gethostbyname(host);"
     "struct sockaddr_in *addr = (struct sockaddr_in *)saddr;"
     "if(he == NULL) return(0);"
@@ -222,12 +252,14 @@ EOF
 			    [p (##net#getservbyname s proto)] )
 		       (when (eq? 0 p)
 			 (##sys#update-errno)
-			 (##sys#signal-hook #:network-error 'tcp-connect "can not compute port from service" s) )
+			 (##sys#signal-hook
+			  #:network-error 'tcp-connect (##sys#string-append "can not compute port from service - " strerror)
+			  s) )
 		       p) )
 		    (loop (fx+ i 1)) ) ) ) ) ) ) ) )
 
 (define ##net#fresh-addr
-  (foreign-lambda* void ((pointer saddr) (unsigned-short port))
+  (foreign-lambda* void ((scheme-pointer saddr) (unsigned-short port))
     "struct sockaddr_in *addr = (struct sockaddr_in *)saddr;"
     "memset(addr, 0, sizeof(struct sockaddr_in));"
     "addr->sin_family = AF_INET;"
@@ -246,16 +278,16 @@ EOF
                       return(setsockopt(socket, SOL_SOCKET, SO_REUSEADDR, (const char *)&yes, sizeof(int)));") 
 		   s) )
       (##sys#update-errno)
-      (##sys#signal-hook #:network-error 'tcp-listen "error while setting up socket" s) )
+      (##sys#signal-hook #:network-error 'tcp-listen (##sys#string-append "error while setting up socket - " strerror) s) )
     (let ([addr (make-string _sockaddr_in_size)])
       (if host
 	  (unless (##net#gethostaddr addr host port)
-	    (##sys#signal-hook #:network-error 'tcp-listen "getting listener host IP failed" host port) )
+	    (##sys#signal-hook #:network-error 'tcp-listen "getting listener host IP failed - " host port) )
 	  (##net#fresh-addr addr port) )
       (let ([b (##net#bind s addr _sockaddr_in_size)])
 	(when (eq? -1 b)
 	  (##sys#update-errno)
-	  (##sys#signal-hook #:network-error 'tcp-listen "can not bind to socket" s port) )
+	  (##sys#signal-hook #:network-error 'tcp-listen (##sys#string-append "can not bind to socket - " strerror) s port) )
 	(values s addr) ) ) ) )
 
 (define-constant default-backlog 10)
@@ -267,7 +299,7 @@ EOF
       (let ([l (##net#listen s w)])
 	(when (eq? -1 l)
 	  (##sys#update-errno)
-	  (##sys#signal-hook #:network-error 'tcp-listen "can not listen on socket" s port) )
+	  (##sys#signal-hook #:network-error 'tcp-listen (##sys#string-append "can not listen on socket - " strerror) s port) )
 	(##sys#make-structure 'tcp-listener s) ) ) ) )
 
 (define (tcp-listener? x) 
@@ -279,20 +311,20 @@ EOF
   (let ([s (##sys#slot tcpl 1)])
     (when (fx= -1 (##net#close s))
       (##sys#update-errno)
-      (##sys#signal-hook #:network-error 'tcp-close "can not close TCP socket" tcpl) ) ) )
+      (##sys#signal-hook #:network-error 'tcp-close (##sys#string-append "can not close TCP socket - " strerror) tcpl) ) ) )
 
-(define-constant buffer-size 1024)
+(define-constant +input-buffer-size+ 1024)
+(define-constant +output-chunk-size+ 8192)
 
 (define ##net#io-ports
   (let ([make-input-port make-input-port]
 	[make-output-port make-output-port] 
-	[make-string make-string] 
-	[substring substring] )
+	[make-string make-string] )
     (lambda (fd)
       (unless (##net#make-nonblocking fd)
 	(##sys#update-errno)
-	(##sys#signal-hook #:network-error "can not create TCP ports") )
-      (let* ([buf (make-string buffer-size)]
+	(##sys#signal-hook #:network-error (##sys#string-append "can not create TCP ports - " strerror)) )
+      (let* ([buf (make-string +input-buffer-size+)]
 	     [data (vector fd #f #f)]
 	     [buflen 0]
 	     [bufindex 0]
@@ -303,7 +335,7 @@ EOF
 	       (lambda ()
 		 (when (fx>= bufindex buflen)
 		   (let ([n (let loop ()
-			      (let ([n (##net#recv fd buf buffer-size 0)])
+			      (let ([n (##net#recv fd buf +input-buffer-size+ 0)])
 				(if (eq? -1 n)
 				    (if (eq? errno _ewouldblock) 
 					(begin
@@ -312,12 +344,15 @@ EOF
 					  (loop) )
 					(begin
 					  (##sys#update-errno)
-					  (##sys#signal-hook #:network-error "can not read from socket" fd) ) )
+					  (##sys#signal-hook 
+					   #:network-error
+					   (##sys#string-append "can not read from socket - " strerror) 
+					   fd) ) )
 				    n) ) ) ] )
 		     (set! buflen n)
 		     (set! bufindex 0) ) )
 		 (if (fx>= bufindex buflen)
-		     (end-of-file)
+		     #!eof
 		     (let ([c (##core#inline "C_subchar" buf bufindex)])
 		       (set! bufindex (fx+ bufindex 1))
 		       c) ) )
@@ -326,7 +361,8 @@ EOF
 		     (let ([f (##net#select fd)])
 		       (when (eq? f -1)
 			 (##sys#update-errno)
-			 (##sys#signal-hook #:network-error "can not check socket for input" fd) )
+			 (##sys#signal-hook #:network-error (##sys#string-append "can not check socket for input - " strerror) 
+					    fd) )
 		       (eq? f 1) ) ) )
 	       (lambda ()
 		 (unless iclosed
@@ -334,55 +370,82 @@ EOF
 		   (unless (##sys#slot data 1) (##net#shutdown fd _sd_receive))
 		   (when (and oclosed (eq? -1 (##net#close fd)))
 		     (##sys#update-errno)
-		     (##sys#signal-hook #:network-error "can not close socket input port" fd) ) ) ) ) ]
+		     (##sys#signal-hook #:network-error (##sys#string-append "can not close socket input port - " strerror)
+					fd) ) ) ) ) ]
 	     [out
 	      (make-output-port
 	       (lambda (s) 
-		 (let ([len (##sys#size s)])
-		   (let loop ()
-		     (let ([n (##net#send fd s len 0)])
-		       (cond [(eq? -1 n)
-			      (if (eq? errno _ewouldblock)
-				  (begin
-				    ;(##sys#thread-block-for-i/o! ##sys#current-thread fd #f)
-				    (yield) 
-				    (loop) )
-				  (begin
-				    (##sys#update-errno)
-				    (##sys#signal-hook #:network-error "can not write to socket" fd s) ) ) ]
-			     [(fx< n len)
-			      (set! s (substring s n len))
-			      (set! len (##sys#size s))
-			      (loop) ] ) ) ) ) )
+		 (let ((len0 (##sys#size s)))
+		   (when (fx> len0 0)
+		     (let loop ((len len0) 
+				(s s) )
+		       (let* ((count (fxmin +output-chunk-size+ len))
+			      (n (##net#send fd s count 0)) )
+			 (cond [(eq? -1 n)
+				(if (eq? errno _ewouldblock)
+				    (begin
+				      (##sys#thread-block-for-i/o! ##sys#current-thread fd #f)
+				      (yield) 
+				      (loop len s) )
+				    (begin
+				      (##sys#update-errno)
+				      (##sys#signal-hook 
+				       #:network-error (##sys#string-append "can not write to socket - " strerror) fd) ) ) ]
+			       [(fx< n len)
+				(loop (fx- len n) (##sys#substring s n (##sys#size s))) ] ) ) ) ) ) )
 	       (lambda ()
 		 (unless oclosed
 		   (set! oclosed #t)
 		   (unless (##sys#slot data 2) (##net#shutdown fd _sd_send))
 		   (when (and iclosed (eq? -1 (##net#close fd)))
 		     (##sys#update-errno)
-		     (##sys#signal-hook #:network-error "can not close socket output port" fd) ) ) ) ) ] )
+		     (##sys#signal-hook
+		      #:network-error (##sys#string-append "can not close socket output port - " strerror) fd) ) ) ) ) ] )
+	(##sys#setslot in 3 "(tcp)")
+	(##sys#setslot out 3 "(tcp)")
+	(##sys#setslot in 7 'socket)
+	(##sys#setslot out 7 'socket)
 	(##sys#setslot (##sys#port-data in) 0 data)
 	(##sys#setslot (##sys#port-data out) 0 data)
 	(values in out) ) ) ) )
 
 (define (tcp-accept tcpl)
   (##sys#check-structure tcpl 'tcp-listener)
-  (##sys#thread-block-for-i/o! ##sys#current-thread (##sys#slot tcpl 1) #t)
-  (yield)
-  (let ([fd (##net#accept (##sys#slot tcpl 1) #f #f)])
-    (when (eq? -1 fd)
-      (##sys#update-errno)
-      (##sys#signal-hook #:network-error 'tcp-accept "could not accept from listener" tcpl) )
-    (##net#io-ports fd) ) )
+  (let ([fd (##sys#slot tcpl 1)])
+    (let loop ()
+      (if (eq? 1 (##net#select fd))
+	  (let ([fd (##net#accept fd #f #f)])
+	    (when (eq? -1 fd)
+	      (##sys#update-errno)
+	      (##sys#signal-hook 
+	       #:network-error 'tcp-accept (##sys#string-append "could not accept from listener - " strerror) 
+	       tcpl) )
+	    (##net#io-ports fd) )
+	  (begin
+	    (##sys#thread-block-for-i/o! ##sys#current-thread fd #t)
+	    (yield)
+	    (loop) ) ) ) ) )
 
 (define (tcp-accept-ready? tcpl)
-  (##sys#check-structure tcpl 'tcp-listener)
+  (##sys#check-structure tcpl 'tcp-listener 'tcp-accept-ready?)
   (let ([f (##net#select (##sys#slot tcpl 1))])
     (when (eq? -1 f)
       (##sys#update-errno)
-      (##sys#signal-hook #:network-error 'tcp-accept-ready? "can not check socket for input" tcpl) )
+      (##sys#signal-hook 
+       #:network-error 'tcp-accept-ready? (##sys#string-append "can not check socket for input - " strerror) 
+       tcpl) )
     (eq? 1 f) ) )
 
+(define get-socket-error
+  (foreign-lambda* int ((int socket))
+		   "int err, optlen;"
+		   "optlen = sizeof(err);"
+		   "if (typecorrect_getsockopt(socket, SOL_SOCKET, SO_ERROR, &err, (socklen_t *)&optlen) == -1)"
+		   "return(-1);"
+		   "return(err);"))
+
+(define general-strerror (foreign-lambda c-string "strerror" int))
+
 (define (tcp-connect host . more)
   (let ([port (:optional more #f)])
     (##sys#check-string host)
@@ -395,15 +458,17 @@ EOF
       (define (fail)
 	(##net#close s)
 	(##sys#update-errno)
-	(##sys#signal-hook #:network-error 'tcp-connect "can not connect to socket" host port) )
+	(##sys#signal-hook 
+	 #:network-error 'tcp-connect (##sys#string-append "can not connect to socket - " strerror) 
+	 host port) )
       (when (eq? -1 s)
 	(##sys#update-errno)
-	(##sys#signal-hook #:network-error 'tcp-connect "can not create socket" host port) )
+	(##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "can not create socket - " strerror) host port) )
       (unless (##net#gethostaddr addr host port)
 	(##sys#signal-hook #:network-error 'tcp-connect "can not find host address" host) )
       (unless (##net#make-nonblocking s)
 	(##sys#update-errno)
-	(##sys#signal-hook #:network-error 'tcp-connect "fcntl() failed") )
+	(##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "fcntl() failed - " strerror)) )
       (when (eq? -1 (##net#connect s addr _sockaddr_in_size))
 	(if (eq? errno _einprogress)
 	    (let loop ()
@@ -414,6 +479,11 @@ EOF
 		  (yield)
 		  (loop) ) ) )
 	    (fail) ) )
+      (let ((err (get-socket-error s)))
+	(cond ((= err -1) 
+	       (##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "getsockopt() failed - " strerror)))
+	      ((> err 0) 
+	       (##sys#signal-hook #:network-error 'tcp-connect (##sys#string-append "can not create socket - " (general-strerror err))))))
       (##net#io-ports s) ) ) )
 
 (define (##sys#tcp-port->fileno p)
@@ -424,22 +494,32 @@ EOF
   (let ([d (##sys#port-data p)])
     (if d
 	(##sys#slot d 0)
-	(##sys#signal-hook #:type-error "bad argument type - not a TCP port" p) ) ) )
+	(##sys#signal-hook #:type-error "bad argument type - not a TCP port - " p) ) ) )
 
 (define (tcp-addresses p)
   (let ([fd (##sys#tcp-port->fileno p)])
     (values 
      (or (##net#getsockname fd)
-	 (##sys#signal-hook #:network-error 'tcp-addresses "can not compute local address" p) )
+	 (##sys#signal-hook #:network-error 'tcp-addresses (##sys#string-append "can not compute local address - " strerror) p) )
      (or (##net#getpeername fd)
-	 (##sys#signal-hook #:network-error 'tcp-addresses "can not compute remote address" p) ) ) ) )
+	 (##sys#signal-hook #:network-error 'tcp-addresses (##sys#string-append "can not compute remote address - " strerror) p) ) ) ) )
+
+(define (tcp-port-numbers p)
+ (let ([fd (##sys#tcp-port->fileno p)])
+   (values
+    (or (##net#getsockport fd)
+        (##sys#signal-hook #:network-error 'tcp-port-numbers (##sys#string-append "can not compute local port - " strerror) p) )
+    (or (##net#getpeerport fd)
+        (##sys#signal-hook #:network-error 'tcp-port-numbers (##sys#string-append "can not compute remote port - " strerror) p) ) ) ) )
 
 (define (tcp-listener-port tcpl)
-  (##sys#check-structure tcpl 'tcp-listener)
+  (##sys#check-structure tcpl 'tcp-listener 'tcp-listener-port)
   (let* ([fd (##sys#slot tcpl 1)]
 	 [port (##net#getsockport fd)] )
     (when (eq? -1 port)
-      (##sys#signal-hook #:network-error 'tcp-listener-port "can not obtain listener port" tcpl fd) )
+      (##sys#signal-hook
+       #:network-error 'tcp-listener-port (##sys#string-append "can not obtain listener port - " strerror) 
+       tcpl fd) )
     port) )
 
 (define (tcp-abandon-port p)
@@ -447,3 +527,7 @@ EOF
    (##sys#tcp-port-data p)
    (if (##sys#slot p 1) 2 1)
    #t) ) 
+
+(define (tcp-listener-fileno l)
+  (##sys#check-structure l 'tcp-listener 'tcp-listener-fileno)
+  (##sys#slot l 1) )
