@@ -37,8 +37,10 @@
 
 (include "parameters")
 
+(cond-expand [(not chicken-compile-shared) (declare (unit tinyclos))]
+	     [else] )
+
 (declare
-  (unit tinyclos)
   (uses extras)
   (usual-integrations)
   (fixnum) 
@@ -234,7 +236,10 @@
             y = C_block_item(y, 1);
           }
         }
-        return(C_block_item(v, i + 1));
+        if (x == C_SCHEME_END_OF_LIST && y == C_SCHEME_END_OF_LIST)
+          return(C_block_item(v, i + 1));
+        else
+          goto mismatch;
       }
       else if(free_index == -1) free_index = i;
     mismatch:
@@ -438,7 +443,7 @@
 (define hash-arg-list
   (foreign-lambda* unsigned-int ((scheme-object args) (scheme-object svector)) "
     C_word tag, h, x;
-    int n, i, j;
+    int n, i, j, len = 0;
     for(i = 0; args != C_SCHEME_END_OF_LIST; args = C_block_item(args, 1)) {
       x = C_block_item(args, 0);
       if(C_immediatep(x)) {
@@ -481,8 +486,9 @@
         default: i += 255;
         }
       }
+      ++len;
     }
-    return(i & (C_METHOD_CACHE_SIZE - 1));") )
+    return((i + len) & (C_METHOD_CACHE_SIZE - 1));") )
 
 
 ;
@@ -513,15 +519,17 @@
 	[(port? x) (if (input-port? x) <input-port> <output-port>)]
 	[(##core#inline "C_pointerp" x) <pointer>]
 	[(##core#inline "C_taggedpointerp" x) <tagged-pointer>]
+	[(##core#inline "C_swigpointerp" x) <swig-pointer>]
 	[(##core#inline "C_locativep" x) <locative>]
 	[(##sys#bytevector? x) <byte-vector>]
 	[(##sys#generic-structure? x)
 	 (case (##sys#slot x 0)
-	   [(dummy-environment) <environment>]
+	   [(environment) <environment>]
 	   [(array) <array>]
 	   [(hash-table) <hash-table>]
 	   [(queue) <queue>]
 	   [(condition) <condition>]
+	   [(condition-variable) <condition-variable>]
 	   [(char-set) <char-set>]
 	   [(time) <time>]
 	   [(lock) <lock>]
@@ -536,6 +544,11 @@
 	   [(f32vector) <f32vector>]
 	   [(f64vector) <f64vector>]
 	   [(tcp-listener) <tcp-listener>]
+	   [(thread) <thread>]
+	   [(mutex) <mutex>]
+	   [(continuation) <continuation>]
+	   [(read-table) <read-table>]
+	   [(regexp) <regexp>]
 	   [else <structure>] ) ]
 	[else (##sys#error 'class-of "can not compute class of primitive object" x)] ) )
 
@@ -665,7 +678,7 @@
 	   (setter (##sys#slot info 1)) )
       (setter object new-value))))
 
-(define slot-ref ##tinyclos#slot-ref)
+(define slot-ref (getter-with-setter ##tinyclos#slot-ref ##sys#tinyclos#slot-set!))
 (define slot-set! ##tinyclos#slot-set!)
 
 (define lookup-slot-info
@@ -867,13 +880,27 @@
     (##tinyclos#slot-set!
      generic
      'methods
-     (cons method
-	   (filter-in
-	    (lambda (m) 
-	      (let ([ms1 (method-specializers m)]
-		    [ms2 (method-specializers method)] )
-		(not (every2 (lambda (x y) (eq? x y)) ms1 ms2) ) ) )
-	    (##tinyclos#slot-ref generic 'methods))))
+     (let* ([ms1 (method-specializers method)]
+	    [l1 (length ms1)] )
+       (let filter-in-method ([methods (##tinyclos#slot-ref generic 'methods)])
+	 (if (null? methods)
+	     (list method)
+	     (let* ([mm (##sys#slot methods 0)]
+		    [ms2 (method-specializers mm)]
+		    [l2 (length ms2)])
+	       (cond ((> l1 l2)
+		      (cons mm (filter-in-method (##sys#slot methods 1))))
+		     ((< l1 l2)
+		      (cons method methods))
+		     (else
+		      (let check-method ([ms1 ms1]
+					 [ms2 ms2])
+			(cond ((and (null? ms1) (null? ms2))
+			       (cons method (##sys#slot methods 1))) ;; skip the method already in the generic
+			      ((eq? (##sys#slot ms1 0) (##sys#slot ms2 0))
+			       (check-method (##sys#slot ms1 1) (##sys#slot ms2 1)))
+			      (else
+			       (cons mm (filter-in-method (##sys#slot methods 1)))))))))))))
     (if (memq generic generic-invocation-generics)
 	(set! method-cache-tag (vector))
 	(%entity-cache-set! generic #f) )
@@ -924,11 +951,13 @@
 				(memq (car args) generic-invocation-generics))
 			   (let ([proc 
 				  (method-procedure
+				    ; select the first method of one argument
 				   (let lp ([lis (generic-methods generic)])
-				     (let ([tail (##sys#slot lis 1)])
-				       (if (null? tail)
-					   (##sys#slot lis 0)
-					   (lp tail)) ) ) ) ] )
+				     (if (null? lis)
+				       (##sys#error "Unable to find original compute-apply-generic")
+				       (if (= (length (method-specializers (##sys#slot lis 0))) 1)
+					 (##sys#slot lis 0)
+					 (lp (##sys#slot lis 1)))))) ] )
 			     (lambda (args) (apply proc #f args)) )
 			   (let ([x (compute-apply-methods generic)]
 				 [y ((compute-methods generic) args)] )
@@ -945,9 +974,13 @@
       (lambda (args)
 	(let ([applicable
 	       (filter-in (lambda (method)
-			    (every2 applicable?
-				   (method-specializers method)
-				   args))
+                            (let check-applicable ([list1 (method-specializers method)]
+                                                   [list2 args])
+                              (cond ((null? list1) #t)
+                                    ((null? list2) #f)
+                                    (else
+                                      (and (applicable? (##sys#slot list1 0) (##sys#slot list2 0))
+                                           (check-applicable (##sys#slot list1 1) (##sys#slot list2 1)))))))
 			  (generic-methods generic) ) ] )
 	  (if (or (null? applicable) (null? (##sys#slot applicable 1))) 
 	      applicable
@@ -974,8 +1007,10 @@
 	   [else
 	    (cond ((and (null? specls1) (null? specls2))
 		   (##sys#error "two methods are equally specific" generic))
-		  ((or (null? specls1) (null? specls2))
-		   (##sys#error "two methods have different number of specializers" generic))
+		  ;((or (null? specls1) (null? specls2))
+		  ; (##sys#error "two methods have different number of specializers" generic))
+                  ((null? specls1) #f)
+                  ((null? specls2) #t)
 		  ((null? args)
 		   (##sys#error "fewer arguments than specializers" generic))
 		  (else
@@ -1067,7 +1102,7 @@
 			       (call-with-values (lambda () (compute-getter-and-setter class slot allocator)) cons) ) )
 		       slots) ] )
 	    (##tinyclos#slot-set! class 'nfields nfields)
-	    (##tinyclos#slot-set! class 'field-initializers field-initializers)
+	    (##tinyclos#slot-set! class 'field-initializers (reverse field-initializers))
 	    (##tinyclos#slot-set! class 'getters-n-setters getters-n-setters))))) ) )
 
 (add-method initialize
@@ -1124,7 +1159,6 @@
       (lambda (call-next-method class)
 	(compute-std-cpl class class-direct-supers))))
 
-
 (add-method compute-slots
   (make-method (list <class>)
     (lambda (call-next-method class)
@@ -1181,21 +1215,16 @@
                 'direct-slots  '()
                 'name          "primitive"))
 
-#|
-(define make-primitive-class
-  (lambda (name . class)
-    (make (if (null? class) <primitive-class> (car class))
-      'name name
-      'direct-supers (list <top>)
-      'direct-slots  '())))
-|#
-
 (define (make-primitive-class name . sclasses)
   (make <primitive-class>
         'direct-supers (if (null? sclasses) (list <primitive>) sclasses)
         'direct-slots  '()
         'name          name))
 
+(add-method initialize
+  (make-method (list <primitive>)
+    (lambda (call-next-method object initargs) object)))
+
 (define <void>      (make-primitive-class "void"))
 (define <null>      (make-primitive-class "null"))
 (define <boolean>   (make-primitive-class "boolean"))
@@ -1215,7 +1244,7 @@
 (define <structure>      (make-primitive-class "structure"))
 (define <procedure> (make-primitive-class "procedure" <procedure-class>))
 (define <end-of-file> (make-primitive-class "end-of-file"))
-(define <environment> (make-primitive-class "environment" <structure>))	; (Benedikt insisted on this)
+(define <environment> (make-primitive-class "environment" <structure>))
 (define <hash-table> (make-primitive-class "hash-table" <structure>))
 (define <promise> (make-primitive-class "promise" <structure>))
 (define <queue> (make-primitive-class "queue" <structure>))
@@ -1235,8 +1264,14 @@
 (define <f64vector> (make-primitive-class "f64vector" <byte-vector>))
 (define <pointer> (make-primitive-class "pointer"))
 (define <tagged-pointer> (make-primitive-class "tagged-pointer" <pointer>))
+(define <swig-pointer> (make-primitive-class "swig-pointer" <pointer>))
 (define <locative> (make-primitive-class "locative"))
 (define <tcp-listener> (make-primitive-class "tcp-listener" <structure>))
+(define <thread> (make-primitive-class "thread" <structure>))
+(define <mutex> (make-primitive-class "mutex" <structure>))
+(define <regexp> (make-primitive-class "regexp" <structure>))
+(define <condition-variable> (make-primitive-class "condition-variable" <structure>))
+(define <read-table> (make-primitive-class "read-table" <structure>))
 (define <c++-object> (make <class> 'name "c++-object" 'direct-supers (list <object>) 'direct-slots '(this)))
 
 (set! method-caching-enabled #t)
@@ -1247,6 +1282,7 @@
 (define initialize-slots
   (let ([not-there (gensym)])
     (lambda (object initargs)
+      (##sys#check-list initargs 'initialize-slots)
       (for-each
        (lambda (slot)
 	 (let* ([name (car slot)]
@@ -1260,28 +1296,29 @@
 
 (add-method print-object
   (make-method (list <object>)
-    (lambda (call-next-method x port)
-      (fprintf port "#<~A>" (class-name (class-of x))) ) ) )
+    (lambda (call-next-method x . port)
+      (fprintf (:optional port ##sys#standard-output) "#<~A>" (class-name (class-of x))) ) ) )
 
 (add-method print-object
   (make-method (list <primitive>)
-    (lambda (call-next-method x port)
-      (write x port) ) ) )
+    (lambda (call-next-method x . port)
+      (write x (:optional port ##sys#standard-output)) ) ) )
 
 (add-method print-object
   (make-method (list <class>)
-    (lambda (call-next-method x port)
-      (fprintf port "#<class ~A>" (class-name x)) ) ) )
+    (lambda (call-next-method x . port)
+      (fprintf (:optional port ##sys#standard-output) "#<class ~A>" (class-name x)) ) ) )
 
 (add-method print-object
   (make-method (list <generic>)
-    (lambda (call-next-method x port)
-      (fprintf port "#<generic ~A>" (%entity-name x)) ) ) )
+    (lambda (call-next-method x . port)
+      (fprintf (:optional port ##sys#standard-output) "#<generic ~A>" (%entity-name x)) ) ) )
 
 (add-method describe-object
   (make-method (list <object>) 
-    (lambda (call-next-method x port)
-      (let ([class (class-of x)])
+    (lambda (call-next-method x . port)
+      (let ([class (class-of x)]
+	    [port (:optional port ##sys#standard-output)] )
 	(fprintf port "instance of class ~A:~%" (class-name class))
 	(for-each 
 	 (lambda (s)
@@ -1291,13 +1328,13 @@
 
 (add-method describe-object
   (make-method (list <class>) 
-    (lambda (call-next-method x port)
-      (fprintf port "class ~A~%" (class-name x)) ) ) )
+    (lambda (call-next-method x . port)
+      (fprintf (:optional port ##sys#standard-output) "class ~A~%" (class-name x)) ) ) )
 
 (add-method describe-object
   (make-method (list <generic>) 
-    (lambda (call-next-method x port)
-      (fprintf port "generic ~A~%" (%entity-name x)) ) ) )
+    (lambda (call-next-method x . port)
+      (fprintf (:optional port ##sys#standard-output) "generic ~A~%" (%entity-name x)) ) ) )
 
 (define ensure-generic
   (let ([make-generic make-generic])
@@ -1320,7 +1357,18 @@
 (define (subclass? x y)
   (if (memq y (compute-cpl x)) #t #f) )
 
+(define (instance-of? x class)
+  (let ([cl (class-of x)])
+    (or (eq? cl class)
+	(subclass? cl class) ) ) )
+
 (define (make-instance-from-pointer ptr cls)
   (and ptr 
        (not (##sys#null-pointer? ptr))
        (make cls 'this ptr) ) )
+
+(add-method initialize
+  (make-method (list <c++-object>)
+    (lambda (call-next-method obj initargs)
+      (when (and (pair? initargs) (eq? 'this (##sys#slot initargs 0)))
+	(##tinyclos#slot-set! obj 'this (cadr initargs)) ) ) ) )
