@@ -1,6 +1,6 @@
 ;;;; utils.scm - Utilities for scripting and file stuff
 ;
-; Copyright (c) 2000-2004, Felix L. Winkelmann
+; Copyright (c) 2000-2005, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -38,6 +38,7 @@
   (uses regex extras)
   (usual-integrations)
   (fixnum)
+  (hide chop-pds)
   (disable-interrupts) )
 
 
@@ -88,42 +89,47 @@
 ;;; Pathname operations:
 
 (define absolute-pathname?
-  (let ([rx (string-append "([A-Za-z]:)?[\\/\\" (string pathname-directory-separator) "].+")]
+  (let ([rx (string-append "([A-Za-z]:)?[\\/\\" (string ##sys#pathname-directory-separator) "].*")]
 	[string-match string-match] )
     (lambda (pn)
       (##sys#check-string pn 'absolute-pathname?)
       (pair? (string-match rx pn)) ) ) )
 
+(define (chop-pds str)
+  (and str
+       (let ([len (##sys#size str)])
+	 (if (and (fx> len 0) 
+		  (eq? ##sys#pathname-directory-separator (##core#inline "C_subchar" str (fx- len 1))) )
+	     (##sys#substring str 0 (fx- len 1))
+	     str) ) ) )
+
 (let ([string-append string-append]
       [absolute-pathname? absolute-pathname?]
-      [pds (string pathname-directory-separator)] 
-      [pes (string pathname-extension-separator)] )
+      [pds (string ##sys#pathname-directory-separator)] )
   (define (conc-dirs dirs)
     (##sys#check-list dirs 'make-pathname)
     (let loop ([strs dirs])
       (if (null? strs)
 	  ""
-	  (string-append (car strs) pds (loop (cdr strs))) ) ) )
+	  (string-append (chop-pds (car strs)) pds (loop (cdr strs))) ) ) )
   (define (_make-pathname dir file . ext)
     (let ([dirs (cond [(or (not dir) (null? dir)) ""]
-		      [(string? dir) dir]
+		      [(string? dir) (conc-dirs (list dir))]
 		      [else (conc-dirs dir)] ) ]
 	  [file (or file "")]
 	  [ext (if (pair? ext) (or (car ext) "") "")] )
+      (##sys#check-string file 'make-pathname)
       (##sys#check-string ext 'make-pathname)
       (string-append
        dirs
-       (if (and-let* ([dlen (##sys#size dirs)]
-		      [(fx> dlen 0)]
-		      [c0 (##core#inline "C_subchar" dirs (fx- dlen 1))] 
-		      [(not (char=? pathname-directory-separator c0))] )
-	     (not (char=? #\/ c0)) )
-	   pds
-	   "")
-       file
+       (if (and dir
+		(fx> (##sys#size file) 0) 
+		(eq? ##sys#pathname-directory-separator (##core#inline "C_subchar" file 0)) )
+	   (##sys#substring file 1 (##sys#size file))
+	   file)
        (if (and (fx> (##sys#size ext) 0)
-		(not (char=? (##core#inline "C_subchar" ext 0) pathname-extension-separator)) )
-	   pes
+		(not (char=? (##core#inline "C_subchar" ext 0) #\.)) )
+	   "."
 	   "")
        ext) ) )
   (set! make-pathname _make-pathname)
@@ -142,20 +148,21 @@
        ext) ) ) )
 
 (define decompose-pathname
-  (let* ([pes (string pathname-extension-separator)]
-	 [set (##sys#string-append "\\/\\" (string pathname-directory-separator))]
-	 [rx1 (string-append "^(.*[" set "])?([^" set "]+)(\\" pes "([^" set pes "]+))$")]
-	 [rx2 (string-append "^(.*[" set "])?((\\" pes ")?[^" set "]+)$")] 
+  (let* ([set (##sys#string-append "\\/\\" (string ##sys#pathname-directory-separator))]
+	 [rx1 (string-append "^(.*[" set "])?([^" set "]+)(\\.([^" set ".]+))$")]
+	 [rx2 (string-append "^(.*[" set "])?((\\.)?[^" set "]+)$")] 
 	 [string-match string-match] )
     (lambda (pn)
       (##sys#check-string pn 'decompose-pathname)
-      (let ([m (string-search rx1 pn)])
-	(if m
-	    (values (cadr m) (caddr m) (car (cddddr m)))
-	    (let ([m (string-search rx2 pn)])
-	      (if m
-		  (values (cadr m) (caddr m) #f)
-		  (values pn #f #f) ) ) ) ) ) ) )
+      (if (eq? (##sys#size pn) 0)
+	  (values #f #f #f) 
+	  (let ([m (string-search rx1 pn)])
+	    (if m
+		(values (chop-pds (cadr m)) (caddr m) (car (cddddr m)))
+		(let ([m (string-search rx2 pn)])
+		  (if m
+		      (values (chop-pds (cadr m)) (caddr m) #f)
+		      (values pn #f #f) ) ) ) ) ) ) ) )
 
 (let ([decompose-pathname decompose-pathname])
   (set! pathname-directory 
@@ -228,9 +235,7 @@
   (define (file-iterator file thunk)
     (if (string=? file "-")
         (for-each-line thunk)
-        (let ((f (open-input-file file)))
-          (for-each-line thunk f)
-          (close-input-port f))))
+        (with-input-from-file file (cut for-each-line thunk) ) ) )
   (let ((args (command-line-arguments)))
     (if (null? args)
         ;; If no arguments, take from stdin, 
@@ -246,3 +251,43 @@
     (if (port? file)
 	(read-string #f file)
 	(with-input-from-file file (cut read-string #f)) ) ) )
+
+
+;;; Handy little things:
+
+(define (shift! lst #!optional default)
+  (if (null? lst)
+      default
+      (begin
+	(##sys#check-pair lst 'shift!)
+	(let ([x (##sys#slot lst 0)]
+	      [d (##sys#slot lst 1)] )
+	  (##sys#check-pair d 'shift!)
+	  (##sys#setslot lst 1 (##sys#slot d 1))
+	  (##sys#setslot lst 0 (##sys#slot d 0))
+	  x) ) ) )
+
+(define (unshift! x lst)
+  (##sys#check-pair lst 'unshift!)
+  (##sys#setslot lst 1 (cons (##sys#slot lst 0) (##sys#slot lst 1)))
+  (##sys#setslot lst 0 x)
+  lst)
+
+
+;;;; Port-mapping (found in Gauche):
+
+(define (port-for-each fn thunk)
+  (let loop ()
+    (let ((x (thunk)))
+      (unless (eq? x #!eof)
+	(fn x)
+	(loop) ) ) ) )
+
+(define port-map
+  (let ((reverse reverse))
+    (lambda (fn thunk)
+      (let loop ((xs '()))
+	(let ((x (thunk)))
+	  (if (eq? x #!eof)
+	      (reverse xs)
+	      (loop (cons x xs))))))))
